;
;  ***********************************************************************
;
;   MENUET Copyright 2000-2001 Ville Mikael Turjanmaa
;
;   See file COPYING for details with these additional details:
;     - All code written in 32 bit x86 assembly language
;     - No external code (eg. bios) at process execution time
;
;   Programming:
;     Ville Mikael Turjanmaa, villemt@itu.jyu.fi
;      * main os code
;     Jan-Michael Brummer, BUZZ2@gmx.de
;      * bugfixes in mouse & display drivers
;      * code for cd-player
;     Felix Kaiser, info@felix-kaiser.de
;      * AMD K6-II compatible IRQ's
;      * APM management
;     Paolo Minazzi, paolo.minazzi@libero.it
;      * Sound Blaster
;      * Fat 32 write
;     quickcode@mail.ru
;      * 320x200 palette & convert
;      * Vesa 1.2 bankswitch for S3 cards
;     Alexey, kgaz@crosswinds.net
;      * Voodoo compatible graphics
;   Translating:
;     Yury Smagin, smagin_yury@hotmail.com
;      * to russia
;     Johannes Weiser, johannes.weiser@chello.at
;      * to german
;
;   Information based upon:
;     - Public domain
;       mouse PS/2 port enable
;     - John Fine
;       A20, TSS, GDT, IDT
;
;   Compile with NASM 0.98+
;
;  ***********************************************************************
;
;
;
;   MEMORY MAP
;
;
;   BOOT:
;
;   0:9000     byte   bits per pixel
;   0:9001     word   scanline length
;   0:9008     word   vesa video mode
;   0:900A     word   X res
;   0:900C     word   Y res
;   0:9010     byte   mouse port
;   0:9014     dword  Vesa 1.2 pm bank switch
;   0:9018     dword  Vesa 2.0 LFB address
;
;
;   RUNNING:
;
;   2000  ->   2FFF   window list
;
;         0000 dword  x start
;         0004 dword  y start
;         0008 dword  x size
;         000C dword  y size
;         0010 dword  color of work area
;         0014 dword  color of grab bar
;         0018 dword  color of frames
;
;   3000  ->   3FFF   task list
;
;         00   dword  process count
;         04   dword  no of processes
;
;         20   dword  application event mask
;         24   dword  process number
;         2a   word   0, state 3=zombie, 4=terminate
;         2e   byte   window number on screen
;         30   dword  exact position in memory
;         34   dword  counter sum
;         38   dword  time stamp counter add
;         3c   dword  cpu usage in cpu timer tics
;
;   4000  ->   AFFF   basic text font
;   BF00  ->   BFFF   bytes, 1 if process running/memory in use
;   C000  ->   C3FF   window stack C000 no of windows - all in words
;   C402  ->   C7FF   window position in stack
;   D000  ->   DFFF   Additional app info
;
;         00   dword  draw limit - x start
;         04   dword  draw limit - y start
;         08   dword  draw limit - x end
;         0C   dword  draw limit - y end
;         1C   dword  app int vector
;
;   E000  byte        multitasking started
;   E004  dword       save for boot mem 0x24
;   E010  dword       pm bank switch address
;   E014  dword       window multiplier
;   E020  dword       putpixel address
;   E024  dword       getpixel address
;   E030  dword       Vesa 1.2 pm bank switch address
;   F000  dword       hd partition 1 start
;   F004  dword       root start at partition 1
;   F008  dword       root start at hd
;   F100  byte        1=only os draw on screen
;   F200  dword       mousepicture -pointer
;   F300  dword       x & y temp for windowmove
;   F400  byte        no of keys in buffer
;   F401  byte        'buffer'
;   F402  ->   F4FF   reserved for keys
;   F500  byte        no of buttons in buffer
;   F501  byte        'buffer'
;   F502  ->   F5FF   reserved for buttons
;   F600  dword       tsc / second
;   F604  byte        mouse port: 1 ps2, 2 com1, 3 com2
;   FB00  ->   FB0F   mouse memory 00 chunk count - FB0A-B x - FB0C-D y
;   FB10  ->   FB17   mouse color mem
;   FB21              x move
;   FB22              y move
;   FB28              high bits temp
;   FB30              color temp
;   FB40  byte        buttons down
;   FB4A  ->   FB4D   FB4A-B x-under - FB4C-D y-under
;   FBF0  byte        multiplied window
;   FBF1  byte        bits per pixel
;   FBF3  byte        RDTSC, 0=not present, 1=present
;   FC00  ->   FCFE   com1/ps2 buffer
;   FCFF              com1/ps2 buffer count starting from FC00
;   FDF0  ->   FDF3   timer tick inc every 1/100 sec 4 bytes
;   FE00  dword       screen x size
;   FE04  dword       screen y size
;   FE08  dword       screen y multiplier
;   FE0C  dword       screen mode
;   FE10  dword       entries in hd cache
;   FE80  dword       address of LFB in physical
;   FE84  dword       address of applications memory start in physical
;   FE88  dword       address of button list
;   FF00  byte        1 = system shutdown request
;   FF01  dword       if >1 then activate process
;   FFF0  byte        1 = redraw background request from app
;   FFF1  byte        1 = diskette int occur
;   FFF2              write and read bank in screen
;   FFF4  byte        0 if first mouse draw & do not return picture under
;   FFF5  byte        1 do not draw pointer
;   FFFF  byte        do not change task
;
;  10000  ->  1FFFF   32 bit os
;
;  20000  ->  2FFFF   32 bit stack
;
;  30000  ->  3FFFF   System Management Mode Save
;
;  40000  ->  4FFFF   TSS's of processes
;
;  50000  ->  500FF   TSS of scheduler
;
;  51000  ->  54000   esp0,esp1,esp2
;
;  5F000  ->  5FFFF   schedulers stack
;
;  60000  ->  7FFFF   paging tables
;
;  80000  ->  8FFFF   additional app info, in 256 byte steps
;
;           00        name of app running
;           80 dword  address of random shaped window area
;           84 byte   shape area scale
;           88 dword  used GDT of TSS
;
;  90000  ->  9FFFF   tmp
;
;  A0000  ->  AFFFF   screen access area
;
;  B0000  ->  FFFFF   bios rest in peace -area
;
; 100000  -> 27FFFF   diskette image
; 280000  -> 281000   diskette fat
;
; 2A0000  -> 2B00ff   wav device data
;
; 2C0000  -> 2C3fff   button info
;
;         0000 word   number of buttons
;
;         first button entry at 0x10
;
;        +0000 word   process number
;        +0002 word   button id number
;        +0004 word   x start
;        +0006 word   x size
;        +0008 word   y start
;        +000A word   y size
;
; 2D0000  -> 2DFFFF   reserved port area
;
;     0000   dword    no of port areas reserved
;
;     0010   dword    process id
;            dword    start port
;            dword    end port
;            dword    0
;
;
; 2E0000  -> 2EFFFF   irq data area
;
; 2F0000  -> 2FFFFF   low memory save
;
; 300000  -> 3FFFFF   background image, max 1 M
;
; 400000  -> 5FFFFF   display info
;
; 600000  -> 7FFFFF   hd cache
;
; 800000  -> BFFFFF   mapped to LFB
;
; C00000 -> 3FFFFFF   for applications
;


;
; Optimize checkmouse
;

; os code


%include "cfgsys.inc"

; these are needed even if !CFG_FAT32 for UNSUPPORTED
SET_FAT32_DATA      equ 0xd0000000
READ_HD_FAT32_FILE  equ 0xd0000001
WRITE_HD_FAT32_FILE equ 0xd0000002

; these are needed even if !CFG_SOUND_SB16 for UNSUPPORTED
SB16_load_music     equ 0xc0000000
SB16_play_music     equ 0xc0000001



%ifdef CFG_SOUND_SB16
DMA             equ 1                   ;SB - your DMA 8 bit [0,1,3]
DMAPage         equ 0x2A                ;first page of 64K in phisical RAM

Rate            equ 44100               ;freq in Hz
DataLength      equ 60000               ;data length to play
SB16Buffer      equ 0x2A0000
SB16_Status     equ SB16Buffer+65536    ;byte = 0  driver=not free
					;     = 1  driver=free

%endif ; CFG_SOUND_SB16



SEGMENT START USE16


	  jmp     startofcode

	  
version   db      'MENUET 0.63 Beta',13,10,0  ; string position fixed

	  dd      endofcode

linef     db      13,10,0

okt       db      " ... OK",0
%ifdef CFG_VIDEO_TEST_VESA
novesa    db      "Display: EGA/CGA",13,10,0
vervesa   db      "Display bios: Vesa x.x ",13,10,0
%endif
%ifdef CFG_LEAVE_UNUSED_CODE
modena    db      "Fatal - VBE 0x112+ required.",0
%endif
%ifdef CFG_TEST_386
not386    db      "Fatal - CPU 386+ required.",0
%endif
btns      db      "Fatal - 24 or 32 bpp.",0
fatalsel  db      "Fatal - Graphics mode not supported by hardware.",0
badsect   db      "Fatal - Bad sector. Replace floppy. ",0

bt24      db      "   Bits per pixel: 24",13,10,0
bt32      db      "   Bits per pixel: 32",13,10,0

%ifdef CFG_VIDMODE_ASK
gr_modes  
%ifdef CFG_VIDEO_VESA12
	  db      " Define set_bank function for Vesa1.2",13,10          
%endif
%ifdef CFG_VIDEO_VESA20
%ifdef CFG_VIDMODE_VESA20_640x480x24
	  db      " [1] Vesa2.0  640 x  480 x 16.7M  LFB",13,10
%endif
%ifdef CFG_VIDMODE_VESA20_800x600x24
	  db      " [2] Vesa2.0  800 x  600 x 16.7M  LFB",13,10
%endif
%ifdef CFG_VIDMODE_VESA20_1024x768x24
	  db      " [3] Vesa2.0 1024 x  768 x 16.7M  LFB",13,10
%endif
%ifdef CFG_VIDMODE_VESA20_1280x1024x24
	  db      " [4] Vesa2.0 1280 x 1024 x 16.7M  LFB",13,10
%endif
%endif ; CFG_VIDEO_VESA20
%ifdef CFG_VIDEO_VESA12
%ifdef CFG_VIDMODE_VESA12_640x480x24
	  db      " [5] Vesa1.2  640 x  480 x 16.7M  Bnk",13,10
%endif
%ifdef CFG_VIDMODE_VESA12_800x600x24
	  db      " [6] Vesa1.2  800 x  600 x 16.7M  Bnk",13,10
%endif
%ifdef CFG_VIDMODE_VESA12_1024x768x24
	  db      " [7] Vesa1.2 1024 x  768 x 16.7M  Bnk",13,10
%endif
%ifdef CFG_VIDMODE_VESA12_1280x1024x24
	  db      " [8] Vesa1.2 1280 x 1024 x 16.7M  Bnk",13,10
%endif
%endif ; CFG_VIDEO_VESA12
%ifdef CFG_VIDMODE_320x200
	  db      " [9] EGA/CGA  320 x  200 x 8 bit",13,10
%endif
	  db      0          

askmode   db      "Select mode ? ",0
%endif ; CFG_VIDMODE_ASK

%ifdef CFG_MOUSE_ASK
askmouse  db      "Mouse hardware ports:",13,10
%ifdef CFG_MOUSE_PS2
	  db      " [1] PS/2 (USB),"
%endif
%ifdef CFG_MOUSE_COM1
	  db "  [2] Com1,  "
%endif
%ifdef CFG_MOUSE_COM2
	  db "[3] Com2 "
%endif
	  db 13,10
	  db      " Select port ? ",0
%endif ; CFG_MOUSE_ASK

diskload  db      "Loading diskette: 00 %",8,8,8,8,0
backspace db      8,0
pros      db      "00",8,8,0

keyin     db      "x",0

d80x25
	  db       'ͻ'
	  db       '                                                                              '
	  db       '                                                                              '
	  db       'ĺ'
	  db       '                                                                              '
%ifdef CFG_IDE
	  db       '                                           Primary Master   :                 '
	  db       '                                           Primary Slave    :                 '
	  db       '                                           Secondary Master :                 '
	  db       '                                           Secondary Slave  :                 '
%else
	  db       '                                                                              '
	  db       '                                                                              '
	  db       '                                                                              '
	  db       '                                                                              '
%endif
%ifdef CFG_SOUND_SB16
	  db       '                                           Sound Blaster    :                 '
%else
	  db       '                                                                              '
%endif
	  db       '                                                                              '
	  db       '                                                                              '
	  db       '                                                                              '
	  db       '                                                                              '
	  db       '                                                                              '
	  db       '                                                                              '
	  db       '                                                                              '
	  db       '                                                                              '
	  db       '                                                                              '
	  db       '                                                                              '
	  db       '                                                                              '
	  db       '                                                                              '
	  db       '                                                                              '
	  db       ' MenuetOS comes with ABSOLUTELY NO WARRANTY; See file COPYING for details     '
	  db       'ͼ'


; display text

print:
	pusha
	mov     ax,186+0x0E00
	xor     bh,bh
	int     0x10
	mov     ax,32+0x0E00
	xor     bh,bh
	int     0x10
	popa
printplain:
	pusha
	mov     dx,0x1000
	mov     es,dx
	mov     ds,dx
      lod2:
	cld
	lodsb
	cmp     al,0
	jz      rprint
	mov     ah,$0e
	xor     bh,bh
	int     0x10
	jmp     lod2
      rprint:
	popa
	ret

; start of code


startofcode:


; draw frames as white on blue

	mov  ax,0x0003
	mov  bx,0x0000
	mov  dx,0x0000
	int  0x10

	push word 0xb800
	pop  es
	mov  di,0
	push word 0x1000
	pop  ds
	mov  si,d80x25
	mov  cx,80*25
	mov  ah,1*16+15
      dfl1:
	cld
	lodsb
	stosw
	loop dfl1
	mov  dx,0x1000
	mov  es,dx


; say hi to user

	mov     si,linef
	call    printplain
	call    printplain
	mov     si,version
	call    print
	mov     si,linef
	call    printplain

%ifdef CFG_TEST_386

; test for 386+

	pushf
	pop     ax
	mov     dx,ax
	xor     ax,0x4000
	push    ax
	popf
	pushf
	pop     ax
	and     ax,0x4000
	and     dx,0x4000
	cmp     ax,dx
	jnz     cpugood
	mov     si,not386
	call    print
	jmp     $
      cpugood:

%endif ; CFG_TEST_386

%ifdef CFG_VIDEO_TEST_VESA
; test for vesa 2.0

	mov     ax,0x4f00
	push    word 0x0000
	pop     es
	mov     di,0xa000
	int     0x10

	cmp     ax,0x004f
	je      vesaok2
	push    word 0x1000
	pop     es
	mov     si,novesa
	call    print
	mov     ax,16
	jmp     vesaerr

      vesaok2:

	mov     ax,[es:di+4]
	mov     dx,ax
	add     ax,48*256+48
	push    word 0x1000
	pop     es
	mov     [es:vervesa+19],ah
	mov     [es:vervesa+21],al
	mov     si,vervesa
	call    print

      vesaerr:
%endif ; CFG_VIDEO_TEST_VESA

%ifdef CFG_VIDMODE_ASK
; Ask user graphics mode


	mov     si,gr_modes
	call    printplain

	push    word 0x1000
	pop     es

	mov     si,askmode
	call    print

	xor     eax,eax

      gml0:
	in      al,0x60

	cmp     eax,2
	jb      sgml1
	cmp     eax,9
	jg      sgml1
%ifdef CFGSYS_VIDMODE_SOME_UNSUPPORTED
; FIXME: there is a better way to do this
%ifdef CFG_VIDMODE_VESA20_640x480x24
	cmp     eax,2
	je vidmode_supported
%endif
%ifdef CFG_VIDMODE_VESA20_800x600x24
	cmp     eax,3
	je vidmode_supported
%endif
%ifdef CFG_VIDMODE_VESA20_1024x768x24
	cmp     eax,4
	je vidmode_supported
%endif
%ifdef CFG_VIDMODE_VESA20_1280x1024x24
	cmp     eax,5
	je vidmode_supported
%endif
%ifdef CFG_VIDMODE_VESA12_640x480x24
	cmp     eax,6
	je vidmode_supported
%endif
%ifdef CFG_VIDMODE_VESA12_800x600x24
	cmp     eax,7
	je vidmode_supported
%endif
%ifdef CFG_VIDMODE_VESA12_1024x768x24
	cmp     eax,8
	je vidmode_supported
%endif
%ifdef CFG_VIDMODE_VESA12_1280x1024x24
	cmp     eax,9
	je vidmode_supported
%endif
	jmp gml0
	vidmode_supported:
%endif ; CFG_VIDMODE_SOME_UNSUPPORTED
	mov     si,ax
	sub     si,2
	shl     si,4
	add     si,gr_table
	mov     bx,[es:si+0]
	mov     cx,[es:si+4]
	mov     dx,[es:si+8]
	jmp     gml10
      sgml1:
%ifdef CFG_VIDMODE_320x200
	cmp     al,10
	jnz     gml00
	mov     bx,0x13
	mov     cx,640
	mov     dx,480
	push    word 0x0
	pop     es
	mov     [es:0x9000],byte 32
	push    word 0x1000
	pop     es
	jmp     gml10
%endif ; CFG_VIDMODE_320x200
      gml00:

	jmp     gml0

%else ; ! CFG_VIDMODE_ASK
	xor eax,eax
	mov al,CFG_VIDMODE_ASSUME
%ifdef CFG_VIDMODE_320x200
	cmp al,8
	jnz other_vidmode
	mov     bx,0x13
	mov     cx,640
	mov     dx,480
	push    word 0x0
	pop     es
	mov     [es:0x9000],byte 32
	push    word 0x1000
	pop     es
	jmp vidmode_cont
other_vid_mode:	
%endif


	mov     si,ax
	shl     si,4
	add     si,gr_table
	mov     bx,[es:si+0]
	mov     cx,[es:si+4]
	mov     dx,[es:si+8]
	jmp     vidmode_cont

%endif ; CFG_VIDMODE_ASK
     gr_table:

	dd      0x112+0100000000000000b ,  640 ,  480 , 0
	dd      0x115+0100000000000000b ,  800 ,  600 , 0
	dd      0x118+0100000000000000b , 1024 ,  768 , 0
	dd      0x11B+0100000000000000b , 1280 , 1024 , 0
	dd      0x112                   ,  640 ,  480 , 0
	dd      0x115                   ,  800 ,  600 , 0
	dd      0x118                   , 1024 ,  768 , 0
	dd      0x11B                   , 1280 , 1024 , 0

%ifdef CFG_VIDMODE_ASK
      gml10:

	add     al,47
	mov     [keyin],al
	mov     si,keyin
	call    printplain
%else
      vidmode_cont:
%endif ; CFG_VIDMODE_ASK

	push    word 0x0000
	pop     es
	mov     [es:0x9008],bx
	mov     [es:0x900A],cx
	mov     [es:0x900C],dx
	push    word 0x1000
	pop     es

      aml1:
	in      al,0x60
	cmp     al,128
	jb      aml1

	mov     ax,32
	cmp     bx,0x13
	je      nov


%ifdef CFG_VIDEO_VESA20
; Find Vesa 2.0 LFB & BPP


	mov     ax,0x4f01
	mov     cx,bx
	and     cx,0xfff
	push    word 0x0000
	pop     es
	mov     di,0xa000
	int     0x10

	; LFB

	mov     ecx,[es:di+0x28]
	mov     [es:0x9018],ecx
	push    ecx

	; BPP

	movzx   ax,byte [es:di+0x19]
	mov     [es:0x9000],ax

	; ---- vbe voodoo
	%define BytesPerScanLine 0x10
	push ax
	mov ax, [es:di+BytesPerScanLine]
	mov [es:0x9001],ax
	pop ax
	; -----
%endif ; CFG_VIDEO_VESA20
      nov:

	cmp     ax,24             ; 24 ?
	jnz     nbpp24
	mov     si,bt24
	jmp     bppl
      nbpp24:

	cmp     ax,32             ; 32 ?
	jnz     nbpp32
	mov     si,bt32
	jmp     bppl
      nbpp32:

	mov     si,btns
	call    print
	jmp     $

      bppl:
	call    printplain
	mov     si,linef
	call    print

%ifdef CFG_VIDEO_VESA12
; Find Vesa 1.2 pm bank switch address

	mov     ax,0x4f0A
	mov     bx,0x0
	int     0x10
	xor     eax,eax
	xor     ebx,ebx
	mov     ax,es
	shl     eax,4
	mov     bx,di
	add     eax,ebx
	xor     ebx,ebx
	mov     bx,[es:di]
	add     eax,ebx
	push    word 0x0
	pop     es
	mov     [es:0x9014],eax
%endif ; CFG_VIDEO_VESA12


%ifdef CFG_MOUSE_ASK
; get mouse port

	mov     si,askmouse
	call    print


      getmouseport:

	in      al,0x60
%ifdef CFG_MOUSE_PS2
	cmp     al,2
	je      mselected
%endif
%ifdef CFG_MOUSE_COM1
	cmp     al,3
	je      mselected
%endif
%ifdef CFG_MOUSE_COM2
	cmp     al,4
	je      mselected
%endif

	jmp     getmouseport
      mselected:
%else
	mov al,CFG_MOUSE_ASSUME
	inc al
%endif ; CFG_MOUSE_ASK

	push    word 0x0000
	pop     es
	sub     al,1
	mov     [es:0x9010],al
	push    word 0x1000
	pop     es


%ifdef CFG_MOUSE_ASK
	add     al,48
	mov     [keyin],al

	mov     si,keyin
	call    printplain

      waitmsel:
	in      al,0x60
	cmp     al,128
	jbe     waitmsel

	mov     si,linef
	call    printplain
	call    printplain
%endif ; CFG_MOUSE_ASK


; read diskette to memory


	mov     si,diskload
	call    print

	mov     ax,0x0000               ; reset drive
	mov     dx,0x0000                
	int     0x13

	mov     cx,0x0001               ; startcyl,startsector
	mov     dx,0x0000               ; starthead,drive
	push    word 80*2               ; read no of sect

      reads:
	pusha
	xor     si,si
      newread:
	push    word 0x0
	pop     es
	mov     bx,0xa000               ; es:bx -> data area
	mov     ax,0x0200+18            ; read, no of sectors to read
	int     0x13

	cmp     ah,0
	jz      goodread
    
	add     si,1
	cmp     si,10
	jnz     newread

	mov     si,linef
	call    printplain
	mov     si,badsect
	call    print

	jmp     $

      goodread:

	; move -> 1mb

	mov     si,movedesc
	push    word 0x1000
	pop     es
	mov     cx,256*18
	mov     ah,0x87
	int     0x15
	mov     eax,[es:movedesc+0x18+2]
	add     eax,512*18
	mov     [es:movedesc+0x18+2],eax

	popa

	inc     dh
	cmp     dh,2
	jnz     bb2
	mov     dh,0
	inc     ch

	pusha                        ; print prosentage

	push    word 0x1000
	pop     es

	xor     eax,eax  ; 5
	mov     al,ch
	shr     eax,2
	and     eax,1
	mov     ebx,5
	mul     bx
	add     al,48
	mov     [es:pros+1],al

	xor     eax,eax  ; 10
	mov     al,ch  
	shr     eax,3
	add     al,48
	mov     [es:pros],al

	mov     si,pros
	call    printplain

	popa

      bb2:

	pop     ax
	dec     ax
	push    ax
	cmp     ax,0
	jnz     rs

	jmp     readdone
      rs:
	jmp     reads

      movedesc

	db      0x00,0x00,0x0,0x00,0x00,0x00,0x0,0x0
	db      0x00,0x00,0x0,0x00,0x00,0x00,0x0,0x0
	db      0xff,0xff,0x0,0xa0,0x00,0x93,0x0,0x0
	db      0xff,0xff,0x0,0x00,0x10,0x93,0x0,0x0
	db      0x00,0x00,0x0,0x00,0x00,0x00,0x0,0x0
	db      0x00,0x00,0x0,0x00,0x00,0x00,0x0,0x0
	db      0x00,0x00,0x0,0x00,0x00,0x00,0x0,0x0
	db      0x00,0x00,0x0,0x00,0x00,0x00,0x0,0x0

      readdone:

	pop     ax

	mov     dx,0x3f2             ; floppy motor off
	mov     al,0
	out     dx,al

	mov     si,backspace
	call    printplain
	call    printplain
	mov     si,okt
	call    printplain



; Page table

	map_mem EQU 63                  ; amount of memory to map

	mov     bx,0x6000
	mov     es,bx                   ; [es:di] = 6000:0
	xor     edi,edi                 
	mov     ecx,256*map_mem         ; Map (mapmem) M
	mov     eax,7
	cld

      pt2:

	cmp     ecx,256*(map_mem-8)                       ;  8 M map to LFB
	jnz     pt3
	pop     eax
	add     eax,7
      pt3:

	cmp     ecx,256*(map_mem-12)                      ;   12 M back to linear = physical
	jnz     pt4
	mov     eax,12*0x100000 + 7
      pt4:

	stosd

	add     eax,4096

	loop    pt2



; 4 Meg Page Directory (comment page table)


	;mov     edi, 0x68000
	;mov     eax, 0x00000000 +110000111b    ;  0 M    - global page
	;stosd
	;mov     eax, 0x00400000 + 10000111b    ;  4 M
	;stosd
	;mov     eax, 0x00800000 + 10000111b    ;  8 M
	;stosd
	;mov     eax, 0x00C00000 + 10000111b    ; 12 M
	;stosd
	;mov     eax, 0x01000000 + 10000111b    ; 16 M
	;stosd
	;mov     eax, 0x01400000 + 10000111b    ; 20 M
	;stosd
	;mov     eax, 0x01800000 + 10000111b    ; 24 M
	;stosd
	;pop     eax
	;add     eax, 110000111b                ; 28 M    - global page              
	;stosd
	;mov     eax,    0x68000 + 10000111b    ; Self pointer
	;stosd                            
	;mov     eax,CR4
	;and     eax,0xffffffff-32       ; set PAE to 0      \
	;or      eax,00010000b           ; set PSE to 1      / -> 4 M  pages
	;or      eax,10000000b           ; PGE -> global paging allowed
	;mov     CR4,eax



; 4 KByte Page Directory

       mov     bx , 0x7000
       mov     es , bx                 ; [es:di] = 7000:0
       xor     edi, edi                 

       mov     eax, 0x60007            ; 0 M
       stosd                           ;
       mov     eax, 0x61007            ; 4 M
       stosd                           ;
       mov     eax, 0x62007            ; 8 M
       stosd                           ;
       mov     eax, 0x63007            ; 12 M
       stosd                           ;
       mov     eax, 0x64007            ; 16 M
       stosd                           ;
       mov     eax, 0x65007            ; 20 M
       stosd                           ;
       mov     eax, 0x66007            ; 24 M
       stosd                           ;
       mov     eax, 0x67007            ; 28 M
       stosd                           ;
       mov     eax, 0x68007            ; 32 M
       stosd                           ;
       mov     eax, 0x69007            ; 36 M
       stosd                           ;
       mov     eax, 0x6a007            ; 40 M
       stosd                           ;
       mov     eax, 0x6b007            ; 44 M
       stosd                           ;
       mov     eax, 0x6c007            ; 48 M
       stosd                           ;
       mov     eax, 0x6d007            ; 52 M
       stosd                           ;
       mov     eax, 0x6e007            ; 56 M
       stosd                           ;
       mov     eax, 0x6f007            ; 60 M
       stosd                           ;

       mov     eax, 0x70007            ; Self pointer
       stosd                           ;

       mov     eax, 0x70000 +8+16      ; Set up page directory and enable caches
       mov     CR3, eax   



; Set graphics


	push    word 0x0000
	pop     es
	mov     bx,[es:0x9008]

	mov     ax,bx                ; vga 320x200
	cmp     bx,0x13
	je      setgr
	mov     ax,0x4f02            ; vesa x.x
      setgr:

	int     0x10
	cmp     ah,0
	jz      gmok

	mov     si,linef
	call    printplain
	mov     si,fatalsel
	call    print
	jmp     $

      gmok:



; Enable paging ?

	mov     ecx,0x00000001
	and     ebx, 65535
	cmp     ebx, 00100000000000000b ; lfb -> paging
	jb      nopaging
	or      ecx, 0x80000000
      nopaging:
     

; Enable 32 bit

	cli                             ; disable all irqs
	cld

	mov     al,255                  ; mask all irqs
	out     0xa1,al
	out     0x21,al

    .5: in      al, 0x64                ; Enable A20
	test    al, 2
	jnz     .5
	mov     al, 0xD1
	out     0x64, al
    .6: in      al, 0x64
	test    al, 2
	jnz     .6
	mov     al, 0xDF
	out     0x60, al

	lgdt    [cs:gdts]               ; Load GDT

	mov     eax, CR0                ; Turn on paging // protected mode
	or      eax, ecx
					; and     eax, 10011111b*65536*256 + 0xffffff
	mov     CR0, eax

	jmp     short $+2

	mov     ax,os_data              ; Selector for os
	mov     ds,ax
	mov     es,ax
	mov     fs,ax
	mov     gs,ax
	mov     ss,ax

	mov     esp,dword 0x2F000       ; Set stack

	db      0x66                    ; make this as 32 bit instruction
	db      0xea                    ; Jump Instruction
	dd      b32                     ; Target Address
	dw      os_code                 ; Segment Selector


BITS 32  

b32:

; zero 0x280000-32 M

	xor   eax,eax
	mov   edi,0x280000
	mov   ecx,(0x2000000-0x280000) / 4
	cld
	rep   stosd


; save & zero low memory

	mov   esi,0x0000
	mov   edi,0x2F0000
	mov   ecx,0x10000 / 4
	cld
	rep   movsd

	xor   eax,eax
	mov   edi,0
	mov   ecx,0x10000 / 4
	cld
	rep   stosd


; save real mode variables


	movzx eax,byte [0x2f0000+0x9010]  ; mouse port
	mov   [0xF604],al

	mov   al,[0x2f0000+0x9000]        ; bpp
	mov   [0xFBF1],al

	movzx eax,word [0x2f0000+0x900A]  ; X max
	sub   eax,1
	mov   [0xfe00],eax
	movzx eax,word [0x2f0000+0x900C]  ; Y max
	sub   eax,1
	mov   [0xfe04],eax

	movzx eax,word [0x2f0000+0x9008]  ; screen mode
	mov   [0xFE0C],eax

%ifdef CFG_VIDEO_VESA12
	mov   eax,[0x2f0000+0x9014]       ; Vesa 1.2 bnk sw add
	mov   [0xE030],eax
%endif
					  ; BytesPerScanLine 
	mov   [0xfe08],word 640*4         ; for 320x200 
	cmp   [0xFE0C],word 0x13
	je    srmvl1                      
	mov   ax,[0x2f0000+0x9001]        ; for others 
	mov   [0xfe08],ax
      srmvl1:


; Graphics addresses

%ifdef CFGSYS_VIDEO_VESA_12AND20

	cmp     [0xfe0c],word 0100000000000000b
	jge     setvesa20
	cmp     [0xfe0c],word 0x13
	je      v20ga32

%endif

%ifdef CFG_VIDEO_VESA12
	mov     [0xe020],dword Vesa12_putpixel24
	mov     [0xe024],dword Vesa12_getpixel24
	cmp     [0xfbf1],byte 24
	jz      ga24
	mov     [0xe020],dword Vesa12_putpixel32
	mov     [0xe024],dword Vesa12_getpixel32
      ga24:
	jmp     v20ga24
%endif

%ifdef CFG_VIDEO_VESA20
      setvesa20:
	mov     [0xe020],dword Vesa20_putpixel24
	mov     [0xe024],dword Vesa20_getpixel24
	cmp     [0xfbf1],byte 24
	jz      v20ga24
      v20ga32:
	mov     [0xe020],dword Vesa20_putpixel32
	mov     [0xe024],dword Vesa20_getpixel32
%endif

%ifdef CFG_VIDEO_VESA20 CFG_VIDEO_VESA12
      v20ga24:
%endif


; tss's in gdt table


	mov    ecx,0x40000
	mov    edi,0
	mov    edx,64

      setgdtl2:

	mov    [edi+gdts+ tss0 +0], word 128                           ;  limit 0:15
	mov    [edi+gdts+ tss0 +2], cx                                 ;  base  0:15
	mov    eax,ecx
	shr    eax,16
	mov    [edi+gdts+ tss0 +4], al                                 ;  base  23:16
	mov    [edi+gdts+ tss0 +7], ah                                 ;  base  31:24
	mov    [edi+gdts+ tss0 +5], word 01010000b *256 +10001001b 

	add    ecx,128
	add    edi,8

	dec    edx
	jnz    setgdtl2




; idt


	mov    edi,0
	mov    edx,256

      setidtl1:

	mov    esi,edi
	shr    esi,1
	add    esi,sys_int
	mov    ecx,[esi] 

	mov    [edi+idts+ 8 +0], cx                    
	mov    [edi+idts+ 8 +2], word os_code
	mov    [edi+idts+ 8 +4], word 0x8f00            ; D_TRAP
	shr    ecx,16
	mov    [edi+idts+ 8 +6], cx

	add    edi,8

	dec    edx
	jnz    setidtl1



; load idt

	lidt    [cs:idts]         ; Load IDT

   

; calculate fat chain for diskette

	call  calculatefatchain


; load char.mt

	mov   eax,char
	mov   ebx,0
	mov   ecx,16800
	mov   edx,0x4000
	call  fileread


%ifdef CFG_TEST_RDTSC
; RDTSC availability - bit 4 in edx


	pushfd                  ; get current flags
	pop     eax
	mov     ecx,eax
	xor     eax,0x00200000  ; attempt to toggle ID bit
	push    eax
	popfd
	pushfd                  ; get new EFLAGS
	pop     eax
	push    ecx             ; restore original flags
	popfd
	and     eax,0x00200000  ; if we couldn't toggle ID,
	and     ecx,0x00200000  ; then this is i486
	cmp     eax,ecx
	jz      nopentium

	; It's Pentium or later. Use CPUID to get processor family.

	xor     eax,eax         ; get processor info
	inc     al
	cpuid
	and     edx,00010000b
	cmp     edx,0
	jz      nordtsc

	mov     [0xfbf3],byte 1

      nordtsc:
      nopentium:
%elifdef CFG_RDTSC
	mov     [0xfbf3],byte 1
%endif ; CFG_TEST_RDTSC




; Detect devices

	call  detect_devices


; timer set to 100/sec

	mov al,0x34              ; set to 100Hz
	out 0x43,al

	mov al,0x9b              ; lsb    1193180 / 1193
	out 0x40,al

	mov al,0x2e              ; msb
	out 0x40,al


; reprogram all irqs to ints 0x20-0x2f

	call  rerouteirqs



; set mouse

	call  setmouse


; set preliminary windowing stack & position in stack

	call  setwindowdefaults


; set background defaults

	call calculatebackground


; reserve system irq's & ports

	call reserve_irqs_ports


; set readports by irq handlers

	call setirqreadports


; set up OS task


	; name for OS/IDLE process

	mov   [0x80000+256+0],dword 'OS/I'
	mov   [0x80000+256+4],dword 'DLE '


	; task list


	mov eax,1
	mov [0x3004],eax    ; number of processes
	mov [0x3000],eax    ; process count

	mov [0x3020+0xE],byte  1   ; on screen number
	mov [0x3020+0x4],dword 1   ; process id number


	; set default flags & stacks

	mov     eax,0x10202     ; sti & resume
	mov     [l.eflags],eax

	mov     ax,os_code
	mov     [l.ss0],ax
	mov     [l.ss1],ax
	mov     [l.ss2],ax

	mov     [l.esp0],dword 0x52000
	mov     [l.esp1],dword 0x53000
	mov     [l.esp2],dword 0x54000

	; osloop - TSS

	mov     eax,CR3
	mov     [l.cr3],eax

	mov     [l.eip],dword osloop

	mov     [l.esp],dword 0x2fff0

	mov     ax,os_code
	mov     [l.cs],ax

	mov     ax,os_data
	mov     [l.ss],ax
	mov     [l.ds],ax
	mov     [l.es],ax
	mov     [l.fs],ax
	mov     [l.gs],ax

	; move tss to 0x40000+128

	mov     eax,tss_sceleton
	mov     ebx,0x40000+128
	mov     ecx,120
	call    memmove

	mov     ax,tss0
	ltr     ax



; TSS for scheduler


	mov     eax,CR3
	mov     [l.cr3],eax

	mov     [l.eip],dword irq0

	mov     [l.esp],dword 0x5fff0

	mov     ax,os_code
	mov     [l.cs],ax

	mov     ax,os_data
	mov     [l.ss],ax
	mov     [l.ds],ax
	mov     [l.es],ax
	mov     [l.fs],ax
	mov     [l.gs],ax

	; move tss to 0x50000

	mov     eax,tss_sceleton
	mov     ebx,0x50000
	mov     ecx,120
	call    memmove



; enable all irq's


	cli
	mov   al,0
	out   0x21,al
	out   0xA1,al


; read tsc / sec

	call   _rdtsc
	mov    ecx,eax

	mov    esi,250               ; wait 1/4 a second
	call   delay_ms

	call   _rdtsc
	sub    eax,ecx
	shl    eax,2

	mov    [0xf600],eax          ; save tsc / sec


; set variables

	call   set_variables

%ifdef CFG_VIDMODE_320x200
; palette for 320x200

	call   palette320x200
%endif

; start multitasking

	mov   [0xe000],byte 1        ; multitasking enabled

	mov   eax,init_task    ; load initial task
	call  start_application_fl   ; enables ints


	jmp   $                      ;  wait here for timer to take control


	; Fly :)



; os-loop

osloop:

	call   checkmouse
	call   checkbuttons
	call   checkwindows
	call   checkmisc
	call   checkEgaCga
	call   checkidle

	jmp    osloop




; command data

char
	db    'CHAR    MT '
init_task
	db    CFG_INIT_TASK


; SUBS




setwindowdefaults:

	pusha

	mov   ax,0
	mov   ecx,0xc000

      swdl:

	add   ax,1
	add   ecx,2

	mov   [ecx+0x000],ax               ; process no

	mov   [ecx+0x400],ax               ; positions in stack
					   ; the more high, the more surface

	cmp   ecx,dword 0xc400-2
	jnz   swdl

	popa

	ret



reserve_irqs_ports:
	
	pusha

	mov  [irq_owner+4*0],byte 1    ; timer
	mov  [irq_owner+4*1],byte 1    ; keyboard
%ifdef CFG_SOUND_SB16
	mov  [irq_owner+4*5],byte 1    ; sound blaster
%endif
	mov  [irq_owner+4*6],byte 1    ; floppy diskette

	mov  [irq_owner+4*13],byte 1   ; math co-pros
	mov  [irq_owner+4*14],byte 1   ; ide I
	mov  [irq_owner+4*15],byte 1   ; ide II

	movzx eax,byte [0xf604]        ; mouse irq
	sub   eax,1
	add   eax,mouseirqtable
	movzx eax,byte [eax]
	shl   eax,2
	mov   [irq_owner+eax],byte 1


				       ; RESERVE PORTS

	mov   edi,1                    ; 0x00-0xff
	mov   [0x2d0000],edi
	shl   edi,4
	add   edi,0x2d0000
	mov   [edi],dword 1
	mov   [edi+4],dword 0x0
	mov   [edi+8],dword 0xff

%ifdef CFG_MOUSE_COM1
	cmp   [0xf604],byte 2          ; com1 mouse ?  -> 0x3f0-0x3ff
	jne   ripl1
	mov   edi,[0x2d0000]
	add   edi,1              
	mov   [0x2d0000],edi
	shl   edi,4
	add   edi,0x2d0000
	mov   [edi],dword 1
	mov   [edi+4],dword 0x3f0
	mov   [edi+8],dword 0x3ff
      ripl1:
%endif ; CFG_MOUSE_COM1

%ifdef CFG_MOUSE_COM2
	cmp   [0xf604],byte 3          ; com2 mouse ?  -> 0x2f0-0x2ff
	jne   ripl2
	mov   edi,[0x2d0000]
	add   edi,1              
	mov   [0x2d0000],edi
	shl   edi,4
	add   edi,0x2d0000
	mov   [edi],dword 1
	mov   [edi+4],dword 0x2f0
	mov   [edi+8],dword 0x2ff
      ripl2:
%endif ; CFG_MOUSE_COM2

	popa

	ret

; Note this is a table, so all entries must be here, regardless of config.
mouseirqtable   db  12    ; ps2
		db  4     ; com1
		db  3     ; com2



setirqreadports:

	mov   [irq12read+0],dword 0x60 + 0x01000000  ; read port 0x60 , byte
	mov   [irq12read+4],dword 0                  ; end of port list

	mov   [irq04read+0],dword 0x3f8 + 0x01000000 ; read port 0x3f8 , byte
	mov   [irq04read+4],dword 0                  ; end of port list

	mov   [irq03read+0],dword 0x2f8 + 0x01000000 ; read port 0x2f8 , byte
	mov   [irq03read+4],dword 0                  ; end of port list

	ret




set_variables:


	mov   ecx,0x100                       ; clear port 0x60
      clearport60:
	in    al,0x60
	loop  clearport60

	mov   [0xfcff],byte 0                 ; mouse buffer
	mov   [0xf400],byte 0                 ; keyboard buffer
	mov   [0xf500],byte 0                 ; button buffer

	mov   [0xfb0a],dword 100*65536+100    ; mouse x/y
%ifdef CFG_SOUND_SB16
	mov   byte [SB16_Status],0            ; Minazzi Paolo
%endif
	mov   [0x400000-12],dword 1           ; tiled background

	mov   [0xfe80],dword 0x100000*8       ; LFB address
	mov   [0xfe84],dword 0x100000*12      ; apps mem base address
	mov   [0xfe88],dword 0x2C0000         ; address of button list

	ret



checkidle:

	pusha

	call change_task

	mov  eax,[idlemem]
	mov  ebx,[0xfdf0]
	cmp  eax,ebx
	jnz  idle_exit

	call _rdtsc
	mov  ecx,eax

      idle_loop:

	mov  eax,[0xfdf0]
	cmp  ebx,eax
	jz   idle_loop

	mov  [idlemem],eax

	call _rdtsc
	sub  eax,ecx
	mov  ebx,[idleuse]
	add  ebx,eax
	mov  [idleuse],ebx

	popa

	ret

      idle_exit:

	mov  ebx,[0xfdf0]
	mov  [idlemem],ebx

	call change_task

	popa

	ret


idlemem        dd 0x0
idleuse        dd 0x0
idleusesec     dd 0x0

process_number dd 0x1

novesachecksum dd 0x0

checkEgaCga:

	cmp    [0xfe0c],dword 0x13
	je     cnvl
	ret
      cnvl:

	pusha

	mov    ecx,[0xfb0a]
	cmp    ecx,[novesachecksum]
	jne    novesal
	popa
	ret

      novesal:
	mov    [novesachecksum],ecx

	mov    ecx,0
	movzx  eax,word[0xfb0c]    
	cmp    eax,100
	jge    m13l3
	mov    eax,100 
      m13l3:
	cmp    eax,480-100
	jbe    m13l4
	mov    eax,480-100
      m13l4:
	sub    eax,100
	mov    ebx,640*4
	mul    ebx
	add    ecx,eax
      
	movzx  eax,word[0xfb0a]    
	cmp    eax,160
	jge    m13l1
	mov    eax,160
      m13l1:
	cmp    eax,640-160
	jbe    m13l2
	mov    eax,640-160
      m13l2:
	sub    eax,160
	mov    ebx,4
	mul    ebx
	add    ecx,eax
      

	mov    esi,[0xfe80]
	add    esi,ecx

	mov    edi,0xa0000
	mov    edx,200
	mov    ecx,320 
	cld

     m13pix:        
	lodsd   
	
	push   eax
	mov    ebx,eax
	
	and    eax,(128+64+32)         ; blue
	shr    eax,5

	and    ebx,(128+64+32)*256     ; green
	shr    ebx,8+2
	add    eax,ebx

	pop    ebx
	and    ebx,(128+64)*256*256    ; red
	shr    ebx,8+8
	add    eax,ebx

	stosb
	loop   m13pix
	mov    ecx,320
	add    esi,4*(640-320)
	dec    edx
	jnz    m13pix

	popa

	ret

%ifdef CFG_VIDMODE_320x200
palette320x200:


	mov    dx,0x3c8
	mov    al,0
	out    dx,al

	mov    ecx,256
	mov    dx,0x3c9
	xor    eax,eax

     palnew: 

       mov al,0 
       test ah,64 
       jz pallbl1 
       add al,21 
     pallbl1: 
       test ah,128 
       jz pallbl2 
       add al,42 
     pallbl2: 
       out dx,al 
       mov al,0 
       test ah,8 
       jz pallbl3 
       add al,8 
     pallbl3: 
       test ah,16 
       jz pallbl4 
       add al,15 
     pallbl4: 
       test ah,32 
       jz pallbl5 
       add al,40 
     pallbl5: 
       out dx,al 
       mov al,0 
       test ah,1 
       jz pallbl6 
       add al,8 
     pallbl6: 
       test ah,2 
       jz pallbl7 
       add al,15 
     pallbl7:
       test ah,4 
       jz pallbl8 
       add al,40 
     pallbl8: 
       out dx,al 
       add ah,1 
       loop palnew 

       ret
%endif ; CFG_VIDMODE_320x200


; find free memory area

find_free_mem:

	push   eax
	push   ebx
	push   ecx
	push   edx
	push   edi

	mov    esi,0x100000*16

	add    edi,esi
	dec    edi 

	mov    eax,2 

	mov    edx,[0x3004]
	cmp    edx,1
	jne    fgdt
	jmp    findf4

     fgdt:

	mov    edx,2

      findgdtl1:

	mov    ecx,edx                               ; get gdt of tss
	shl    ecx,8
	add    ecx,0x80000+0x88
	mov    ecx,[ecx]
	shl    ecx,3   

	; eax run base -> ebx limit

	mov    al,[ecx+gdts+ app_code +4]            ;  base  23:16
	mov    ah,[ecx+gdts+ app_code +7]            ;  base  31:24
	shl    eax,16
	mov    ax,[ecx+gdts+ app_code +2]            ;  base  0:15

	movzx  ebx,word [ecx+gdts+ app_code +0]
	shl    ebx,12
	add    ebx,eax

	cmp    eax,edi
	jg     findfl3

	cmp    ebx,esi
	jb     findfl3 

	add    esi,4096
	add    edi,4096

	cmp    edi,63*0x100000
	jbe    fgdt

	mov    esi,0               ; not enough memory found
	jmp    find_free_ret


      findfl3:

	inc    edx
	cmp    edx,[0x3004]
	jbe    findgdtl1

     findfl1:

	; find free gdt place

	mov    eax,2

     findf31:

	mov    edx,2
	add    eax,1

     findf3:

	mov    ecx,edx              ; get gdt of tss
	shl    ecx,8
	add    ecx,0x80000+0x88
	mov    ecx,[ecx]
	cmp    eax,ecx
	je     findf31

	add    edx,1
	cmp    edx,[0x3004]
	jbe    findf3

     findf4:                        ; set gdt

	mov    [gdt_place],eax

	mov    ecx,eax
	shl    ecx,3

	inc    edi                  ; memory requirement in 4096 bytes
	sub    edi,esi
	add    edi,4095
	shr    edi,12
	dec    edi


	; code

	mov    eax,esi 
	mov    ebx,edi

	mov    [ecx+gdts+ app_code +2], ax                                 ;  base  0:15
	shr    eax,16
	mov    [ecx+gdts+ app_code +4], al                                 ;  base  23:16
	mov    [ecx+gdts+ app_code +7], ah                                 ;  base  31:24

	mov    [ecx+gdts+ app_code +0], bx                                 ;  limit 

	mov    [ecx+gdts+ app_code +5], word 11010000b *256 +10011010b 


	; data

	mov    eax,esi 

	mov    [ecx+gdts+ app_data +2], ax                                 ;  base  0:15
	shr    eax,16
	mov    [ecx+gdts+ app_data +4], al                                 ;  base  23:16
	mov    [ecx+gdts+ app_data +7], ah                                 ;  base  31:24

	mov    [ecx+gdts+ app_data +0], bx                                 ;  limit

	mov    [ecx+gdts+ app_data +5], word 11010000b *256 +10010010b 


      find_free_ret:

	pop    edi
	pop    edx
	pop    ecx
	pop    ebx
	pop    eax

	ret



gdt_place dd 0x0


get_app_params:

    pusha

%ifdef CFG_HEADER_00

    cmp  [0x90000+6],word '00'
    jne  no_first_header

    mov  eax,[0x90000+12]
    mov  [app_start],eax
    mov  eax,[0x90000+16]
    mov  [app_i_end],eax
    mov  eax,[0x90000+20]
    mov  [app_mem],eax
    shr  eax,1
    sub  eax,0x10
    mov  [app_esp],eax    
    mov  eax,[0x90000+24]
    mov  [app_i_param],eax
    mov  [app_i_icon],dword 0

    popa

    ret

  no_first_header:

%endif ; CFG_HEADER_00

%ifdef CFG_HEADER_01

    cmp  [0x90000+6],word '01'
    jne  no_second_header

    mov  eax,[0x90000+12]
    mov  [app_start],eax
    mov  eax,[0x90000+16]
    mov  [app_i_end],eax
    mov  eax,[0x90000+20]
    mov  [app_mem],eax
    mov  eax,[0x90000+24]
    mov  [app_esp],eax    
    mov  eax,[0x90000+28]
    mov  [app_i_param],eax
    mov  eax,[0x90000+32]
    mov  [app_i_icon],eax

    popa

    ret

  no_second_header:

%endif ; CFG_HEADER_01

    popa

    ret



app_start    dd  0x0
app_i_end    dd  0x0
app_mem      dd  0x0
app_esp      dd  0x0
app_i_param  dd  0x0
app_i_icon   dd  0x0

app_mem_pos  dd  0x0

start_application_fl:


    push  ebx
    push  eax

    mov   ebx,1
    mov   ecx,20000
    mov   edx,0x90000
    mov   esi,0
    call  fileread

  file_in_place:

    cli

    cmp   [0x90000+0],dword 'MENU'       ; valid menuet app ?
    jnz   noflrun
    cmp   [0x90000+4],word  'ET'
    jnz   noflrun

    call  get_app_params

    mov   edi,[app_mem]

    call  find_free_mem

    cmp   esi,0
    jne   start_app_c

  noflrun:

    pop   eax
    pop   ebx

    sti
    ret


  start_app_c:

    mov   [app_mem_pos],esi

    mov   edi,[app_mem_pos]           ; clear run location
    mov   ecx,[app_mem]  
    mov   eax,0x0
    cld
    rep   stosb

    mov   eax,0x90000               ; move process to its run location
    mov   ebx,[app_mem_pos]
    mov   ecx,[app_i_end]
    call  memmove

    mov   esi,[esp+4]               ; 'command line' parameter
    cmp   esi,0
    jz    no_app_params
    cmp   [app_i_param],dword 0
    jz    no_app_params
    mov   eax,[app_i_param]
    add   eax,[app_mem_pos]
    mov   edi,eax
    mov   ecx,256
    cld
  app_new_param:
    cmp   [esi],byte 0
    jz    no_app_params
    movsb
    loop  app_new_param
  no_app_params:

    mov   ebx,[0x3004]              ; increase number of processes
    inc   ebx
    mov   [0x3004],ebx

    mov   edx,ebx                   ; process number for screen window
    shl   edx,1
    add   edx,0xc000
    mov   [edx],bx
    add   edx,0x0400
    mov   [edx],bx

    mov   edi,ebx                   ; clear 0x80000 (256 bytes)
    shl   edi,8
    add   edi,0x80000
    mov   ecx,256 / 4
    mov   eax,0
    cld
    rep   stosd

    shl   ebx,5                     ; * 32 +0x3000
    add   ebx,0x3000

    mov   edi,ebx                   ; clear 0x3000+ & 0x2000+ & 0x80000+ (256 bytes)
    mov   ecx,32 /4
    xor   eax,eax
    cld
    rep   stosd
    mov   edi,ebx
    sub   edi,0x1000
    mov   ecx,32 / 4
    cld
    rep   stosd

    push  ebx
    mov   eax,1                     ; screen id
  nscc2:
    add   eax,1
    mov   ecx,[0x3004]
    mov   edx,0x3020+0xe
    cld
  nscc:
    movzx ebx,byte [edx]
    cmp   ebx,eax
    jz    nscc2
    add   edx,0x20
    loop  nscc
    pop   ebx
    mov   [ebx+0xe],al

    mov   [ebx],dword 1+2+4         ; report events: windowdraw, key, button

    mov   eax,[process_number]      ; process id number
    add   eax,1
    mov   [process_number],eax
    mov   [ebx+4],eax

    mov   ecx,ebx                   ; set draw limits
    add   ecx,0xd000-0x3000
    mov   [ecx+0],dword 0           
    mov   [ecx+4],dword 0
    mov   eax,[0xfe00]
    mov   [ecx+8],eax
    mov   eax,[0xfe04]
    mov   [ecx+12],eax

    mov   eax,[app_mem_pos]           ; position in memory
    mov   [ebx+0x10],eax


    ; TSS

    mov   eax,CR3
    mov   [l.cr3],eax

    mov   eax,[app_start]
    mov   [l.eip],eax

    mov   eax,[app_esp]
    mov   [l.esp],eax


    mov   edx,[0x3004]              ; gdt's
    shl   edx,8
    add   edx,0x80000+0x88
    mov   ebx,[gdt_place]
    mov   [edx],ebx

    shl   ebx,3

    mov   ax,app_code
    add   ax,bx   
    mov   [l.cs],ax

    mov   ax,app_data
    add   ax,bx   
    mov   [l.ss],ax
    mov   [l.ds],ax
    mov   [l.es],ax
    mov   [l.fs],ax
    mov   [l.gs],ax

    mov   eax,tss_sceleton          ; move tss to 0x40000+
    mov   ebx,[0x3004]
    shl   ebx,7
    add   ebx,0x40000
    mov   ecx,120
    call  memmove


    ; *


    pop   eax
    pop   ebx

    mov   ebx,[0x3004]              ; save name of the process
    shl   ebx,8
    add   ebx,0x80000
    mov   ecx,11
    call  memmove

    mov   [0xf400],byte 0           ; empty keyboard buffer
    mov   [0xf500],byte 0           ; empty button buffer

    sti

    ret


%ifdef CFG_HD
start_application_hd:


   
    push  ebx
    push  eax

;  eax,startf           ; location
;  ebx,12*3             ; [loclen]
;  ecx,data_area+20000

;  eax pointer to file
;  ebx file lenght
;  ecx start 512 byte block number
;  edx number of blocks to read
;  esi pointer to return/work area (atleast 20 000 bytes)

    mov   esi,ecx
    mov   ecx,1
    mov   edx,1
    pusha
    call  read_hd_file
    popa

    mov   ecx,1

  hdnw:                                 ; read rest of the file

    push  ecx
    push  esi
    mov   eax,[esp+8]
    mov   ebx,[esp+12]
    mov   edx,1
    call  read_hd_file
    pop   esi
    pop   ecx

    cmp   eax,0
    jz    hdnw2

    cli
    mov   eax,12
    jmp   stnshd

  hdnw2:

    push  ecx
    mov   eax,esi
    add   eax,1024
    mov   ebx,ecx
    sub   ebx,1
    shl   ebx,9
    add   ebx,0x90000
    mov   ecx,512
    call  memmove
    pop   ecx

    add   ecx,1

    jmp   hdnw


  nohdrun:

    pop   eax
    pop   ebx
    ret


  stnshd:

    jmp   file_in_place             ; use start_application_fl
				    ; for the parameters

%endif ; CFG_HD

sys_outport:

    mov   ecx,[0x2d0000]
    cmp   ecx,0 
    jne   sopl8
    
    mov   eax,1
    ret

  sopl8:
    mov   edx,[0x3010]
    mov   edx,[edx+0x4]
    and   ebx,65535
    cld
  sopl1:

    mov   esi,ecx
    shl   esi,4
    add   esi,0x2d0000

    cmp   edx,[esi+0]
    jne   sopl2
    cmp   ebx,[esi+4]
    jb    sopl2
    cmp   ebx,[esi+8]
    jg    sopl2

    jmp   sopl3

  sopl2:

    dec   ecx
    jnz   sopl1

    mov   eax,1

    ret

  sopl3:

    mov   dx,bx
    out   dx,al

    mov   eax,0

    ret


checkscreenpixel:

    cld
    mov   esi,[0x3004]
    inc   esi

  sciloop:

    cmp   esi,2
    jg    scic3

    mov   ecx,1          ; os pixel
    ret

  scic3:

    cld

    dec   esi

    mov   ecx,esi
    shl   ecx,1
    add   ecx,0xc400
    movzx edx,word [ecx]

    mov   edi,edx
    shl   edi,5
    add   edi,0x2000

    mov   ecx,[edi+0]
    mov   edx,[edi+4]

    cmp   ecx,eax
    jbe   sci1
    jmp   sciloop
  sci1:
    cmp   edx,ebx
    jbe   sci2
    jmp   sciloop
  sci2:

    add   ecx,[edi+8]
    add   edx,[edi+12]

    cmp   eax,ecx
    jbe   sci3
    jmp   sciloop
  sci3:
    cmp   ebx,edx
    jbe   sci4
    jmp   sciloop
  sci4:

    mov   ecx,esi
    shl   ecx,1
    movzx ecx,word [ecx+0xc400]       ; process of pixel

    ; check that the process has a rectangle window

    mov   edx,ecx
    shl   edx,8
    add   edx,0x80000+0x80
%ifdef CFG_RANDOM_WIN
    cmp   [edx],dword 0
    jne   rand_shaped
%endif
    jmp   rect_shaped

%ifdef CFG_RANDOM_WIN
  rand_shaped:

    pusha

    sub   eax,[edi+0]    
    sub   ebx,[edi+4]    

    push  ecx

    mov   ecx,[edx+4]
    shr   eax,cl
    shr   ebx,cl
    mov   esi,[edi+8]
    add   esi,1
    shr   esi,cl

    imul  ebx,esi
    add   eax,ebx
    add   eax,[edx]

    pop   ecx

    mov   edx,ecx
    shl   edx,5
    add   eax,[edx+0x3000+0x10]

    cmp   [eax],byte 1
    je    rand_window_pixel

    popa

    jmp   sciloop


  rand_window_pixel:

    popa
%endif ; CFG_RANDOM_WIN

  rect_shaped:


    shl   ecx,5
    add   ecx,0x3000
    movzx ecx,byte [ecx+0xe]   ; screen id of process

    ret


calculatescreen:


;  eax  x start
;  ebx  y start
;  ecx  x end
;  edx  y end

     pusha

     push eax

   csp1:

     push ecx
     push edx
     push eax
     push ebx

     call checkscreenpixel

     mov  eax,[esp+4]
     mov  ebx,[esp+0]

     cld
     mov  esi,eax
     mov  eax,[0xfe00]
     inc  eax
     mul  ebx
     add  eax,esi
     add  eax,0x400000
     mov  [eax],cl

     pop  ebx
     pop  eax
     pop  edx
     pop  ecx

     cld

     inc  eax
     cmp  eax,ecx
     jbe  csp1

     mov  eax,[esp]

     inc  ebx
     cmp  ebx,edx
     jbe  csp1

     pop  eax

     popa

     ret


setscreen:

;  eax  x start
;  ebx  y start
;  ecx  x end
;  edx  y end


     pusha

     push esi
     push eax

   csp11:

     push eax
     push ebx
     push ecx
     push edx

     mov  esi,eax
     mov  eax,[0xfe00]
     add  eax,1
     mul  ebx
     add  eax,esi
     add  eax,0x400000

     mov  cl,[esp+20]
     mov  [eax],cl

     pop  edx
     pop  ecx
     pop  ebx
     pop  eax

     add  eax,1
     cmp  eax,ecx
     jbe  csp11

     pop  eax
     push eax

     add  ebx,1
     cmp  ebx,edx
     jbe  csp11

     add  esp,8

     popa

     ret



%ifdef CFG_SOUND_SB16
sys_sb16:

     cmp  [sb16],word 0
     jnz  sb16l1

     mov  eax,1
     ret


   sb16l1:

     cmp  eax,1    ; set volume - main
     jnz  sb16l2

     cld

     mov  dx,[sb16]
     add  dx,4
     mov  al,$22
     out  dx,al

     mov  esi,1
     call delay_ms
     mov  eax,ebx
     add  edx,1
     out  dx,al

     mov  eax,0
     ret

   sb16l2:


     cmp  eax,2    ; set volume - cd
     jnz  sb16l3

     cld

     mov  dx,[sb16]
     add  dx,4
     mov  al,$28
     out  dx,al

     mov  esi,1
     call delay_ms
     mov  eax,ebx
     add  edx,1
     out  dx,al

     mov  eax,0
     ret


   sb16l3:


     mov  eax,2  ; invalid command
     ret



sys_sb16II:


     cmp  [sb16],word 0
     jnz  IIsb16l1

     mov  eax,1
     ret


   IIsb16l1:

     cmp  eax,1    ; set volume - main
     jnz  IIsb16l2

     cld

     ; L

     mov  dx,[sb16]
     add  dx,4
     mov  al,$30
     out  dx,al

     mov  eax,ebx
     add  edx,1
     out  dx,al

     ; R

     mov  dx,[sb16]
     add  dx,4
     mov  al,$31
     out  dx,al

     mov  eax,ebx
     add  edx,1
     out  dx,al

     mov  eax,0
     ret

   IIsb16l2:


     cmp  eax,2    ; set volume - cd
     jnz  IIsb16l3

     cld

     ; L

     mov  dx,[sb16]
     add  dx,4
     mov  al,$36
     out  dx,al

     mov  eax,ebx
     add  edx,1
     out  dx,al

     ; R

     mov  dx,[sb16]
     add  dx,4
     mov  al,$37
     out  dx,al

     mov  eax,ebx
     add  edx,1
     out  dx,al


     mov  eax,0
     ret


   IIsb16l3:


     mov  eax,2  ; invalid command
     ret

%endif ; CFG_SOUND_SB16

%ifdef CFG_SOUND_WSS
sys_wss:

     cmp  [wss],word 0
     jnz  wssl1

     mov  eax,1
     ret

   wssl1:

     cmp  eax,1    ; set volume - main
     jnz  wssl2

     mov  eax,0
     ret

   wssl2:


     cmp  eax,2    ; set volume - cd
     jnz  wssl3

     cld

     mov  dx,[wss] ; L
     add  dx,4
     mov  al,$2
     out  dx,al
     mov  esi,1
     call delay_ms
     mov  eax,ebx
     add  edx,1
     out  dx,al

     mov  dx,[wss] ; R
     add  dx,4
     mov  al,$3
     out  dx,al
     mov  esi,1
     call delay_ms
     mov  eax,ebx
     add  edx,1
     out  dx,al

     mov  eax,0
     ret


   wssl3:


     mov  eax,2  ; invalid command
     ret
%endif ; CFG_SOUND_WSS


display_number:

; eax = print type, al=0 -> ebx is number 
;                   al=1 -> ebx is pointer
;                   ah=0 -> display decimal 
;                   ah=1 -> display hexadecimal
;                   ah=2 -> display binary
;                   eax bits 16-21 = number of digits to display (0-32)
;                   eax bits 22-31 = reserved
;                 
; ebx = number or pointer
; ecx = x shl 16 + y
; edx = color


     cmp   eax,0xffff            ; length > 0 ?
     jge   cont_displ
     ret
   cont_displ:


     pusha

     cmp   al,1                  ; ecx is a pointer ?  
     jne   displnl1
     
     mov   edi,[0x3010]
     mov   edi,[edi+0x10]
     mov   ebx,[edi+ebx]

   displnl1:

     sub   esp,64

     cmp   ah,0                  ; DESIMAL
     jne   no_display_desnum

     shr   eax,16
     and   eax,0x2f

     push  eax
     
     mov   edi,[0x3010]
     mov   edi,[edi+0x10]
     add   edi,esp
     add   edi,4+64

     mov   ecx,eax
     mov   eax,ebx
     mov   ebx,10
   d_desnum:
     xor   edx,edx
     div   ebx
     add   dl,48
     mov   [edi],dl
     dec   edi
     loop  d_desnum

     pop   eax

     call  draw_num_text   

     add   esp,64
     popa 
     ret


   no_display_desnum:
    
     
     cmp   ah,0x01               ; HEXADECIMAL
     je    yes_display_hexnum
     jmp   no_display_hexnum
   yes_display_hexnum:

     shr   eax,16
     and   eax,0x2f

     push  eax
     
     mov   edi,[0x3010]
     mov   edi,[edi+0x10]
     add   edi,esp
     add   edi,4+64

     mov   ecx,eax
     mov   eax,ebx
     mov   ebx,16
   d_hexnum:
     xor   edx,edx
     div   ebx
     add   edx,hexletters
     mov   dl,[edx]
     mov   [edi],dl
     dec   edi
     loop  d_hexnum

     pop   eax

     call  draw_num_text   
     add   esp,64
     popa 
     ret

   no_display_hexnum:
    

     cmp   ah,0x02               ; BINARY
     je    yes_display_binnum
     jmp   no_display_binnum
   yes_display_binnum:

     shr   eax,16
     and   eax,0x2f

     push  eax
     
     mov   edi,[0x3010]
     mov   edi,[edi+0x10]
     add   edi,esp
     add   edi,4+64

     mov   ecx,eax
     mov   eax,ebx
     mov   ebx,2
   d_binnum:
     xor   edx,edx
     div   ebx
     add   dl,48
     mov   [edi],dl
     dec   edi
     loop  d_binnum

     pop   eax
   
     call  draw_num_text   
     add   esp,64
     popa 
     ret

   no_display_binnum:
     

     add   esp,64
     popa 
     ret




draw_num_text:

  
	     ; dtext
	     ;
	     ; eax x & y
	     ; ebx color
	     ; ecx start of text
	     ; edx length
	     ; edi 1 force
     
     mov   edx,eax

     mov   ecx,65
     sub   ecx,eax
     add   ecx,esp
     mov   edi,[0x3010]
     add   ecx,[edi+0x10]
     add   ecx,4

     mov   eax,[esp+64+32-8+4]
     mov   ebx,[esp+64+32-12+4]

     push  edx                       ; add window start x & y
     push  ebx
     mov   edx,[0x3010]
     mov   ebx,[edx-0x1000]
     shl   ebx,16
     add   ebx,[edx-0x1000+4]
     add   eax,ebx
     pop   ebx
     pop   edx

     mov   edi,0
     call  dtext

     ret
     



sys_setup:

; 1=roland mpu midi base , base io address
; 2=keyboard   1, base kaybap 2, shift keymap, 9 country 1eng 2fi 3ger 4rus
; 3=cd base    1, pri.master 2, pri slave 3 sec master, 4 sec slave
; 4=sb16 base , base io address
; 5=system language, 1eng 2fi 3ger 4rus
; 6=wss base , base io address
; 7=hd base    1, pri.master 2, pri slave 3 sec master, 4 sec slave
; 8=fat32 partition in hd

     cmp  eax,1
     jnz  syse0

     cmp  ebx,0x100
     jb   syse0
     mov  edx,65535
     cmp  edx,ebx
     jb   syse0

     cld
     mov  [mididp],bx
     inc  bx
     mov  [midisp],bx

     mov  eax,0
     ret

   syse0:

     cmp  eax,2
     jnz  syse1

     cmp  ebx,1
     jnz  kbnobase

     mov  edi,[0x3010]
     add  edi,0x10
     add  ecx,[edi]

     mov  eax,ecx
     mov  ebx,keymap
     mov  ecx,128
     call memmove
     mov  eax,0
     ret

   kbnobase:

     cmp  ebx,2
     jnz  kbnoshift

     mov  edi,[0x3010]
     add  edi,0x10
     add  ecx,[edi]

     mov  eax,ecx
     mov  ebx,keymap_shift
     mov  ecx,128
     call memmove

     mov  eax,0
     ret

   kbnoshift:

     cmp  ebx,9
     jnz  kbnocountry

     mov  [keyboard],cx

     mov  eax,0
     ret


   kbnocountry:

     mov  eax,1
     ret

   syse1:

     cmp  eax,3
     jnz  syse2

     cmp  ebx,1
     jnz  noprma
     mov  [cdbase],word 0x1f0
     mov  [cdid],word 0xa0
   noprma:
     cmp  ebx,2
     jnz  noprsl
     mov  [cdbase],word 0x1f0
     mov  [cdid],word 0xb0
   noprsl:
     cmp  ebx,3
     jnz  nosema
     mov  [cdbase],word 0x170
     mov  [cdid],word 0xa0
   nosema:
     cmp  ebx,4
     jnz  nosesl
     mov  [cdbase],word 0x170
     mov  [cdid],word 0xb0
   nosesl:

     mov  eax,0
     ret

   syse2:

     cmp  eax,4
     jnz  syse3

     cmp  ebx,0x100
     jb   syse3
     mov  edx,65535
     cmp  edx,ebx
     jb   syse3

     mov  [sb16],bx

     mov  eax,0
     ret

   syse3:

     cmp  eax,5
     jnz  syse4

     mov  [syslang],ebx

     mov  eax,0
     ret

   syse4:

     cmp  eax,6
     jnz  syse5

     cmp  ebx,0x100
     jb   syse3
     mov  edx,65535
     cmp  edx,ebx
     jb   syse3

     mov  [wss],ebx

     mov  eax,0
     ret

   syse5:

     cmp  eax,7
     jz   syse7no
     jmp  syse7

   syse7no:

     cmp  ebx,1
     jnz  noprmahd
     mov  [hdbase],word 0x1f0
     mov  [hdid],word 0x0
     mov  [hdpos],dword 1
   noprmahd:
     cmp  ebx,2
     jnz  noprslhd
     mov  [hdbase],word 0x1f0
     mov  [hdid],word 0x10
     mov  [hdpos],dword 2
   noprslhd:
     cmp  ebx,3
     jnz  nosemahd
     mov  [hdbase],word 0x170
     mov  [hdid],word 0x0
     mov  [hdpos],dword 3
   nosemahd:
     cmp  ebx,4
     jnz  noseslhd
     mov  [hdbase],word 0x170
     mov  [hdid],word 0x10
     mov  [hdpos],dword 4
   noseslhd:

     mov  eax,0
     ret

   syse7:

     cmp  eax,8
     jnz  syse8

     mov  [fat32part],ebx

     mov  eax,0
     ret

   syse8:


     mov  eax,1
     ret


sys_getsetup:

; 1=roland mpu midi base , base io address
; 2=keyboard   1, base kaybap 2, shift keymap, 9 country 1eng 2fi 3ger 4rus
; 3=cd base    1, pri.master 2, pri slave 3 sec master, 4 sec slave
; 4=sb16 base , base io address
; 5=system language, 1eng 2fi 3ger 4rus
; 6=wss base
; 7=hd base    1, pri.master 2, pri slave 3 sec master, 4 sec slave
; 8=fat32 partition in hd

     cmp  eax,1
     jnz  gsyse0

     mov  eax,0
     ret

   gsyse0:

     cmp  eax,2
     jnz  gsyse1

     cmp  ebx,9
     jnz  gkbnobase

     xor  eax,eax
     mov  ax,[keyboard]

     ret

   gkbnobase:

   gsyse1:

     cmp  eax,3
     jnz  gsyse2

     mov  eax,0
     ret

   gsyse2:

     cmp  eax,4
     jnz  gsyse3

     mov  eax,0
     ret

   gsyse3:

     cmp  eax,5
     jnz  gsyse4

     mov  eax,[syslang]

     ret

   gsyse4:

     mov  eax,1
     ret



read_screen_pixel:

%ifdef CFGSYS_VIDEO_VESA_12AND20

     cmp   [0xfe0c],word 0100000000000000b
     jge   rspv20
     cmp   [0xfe0c],word 0x13
     je    rspv20

%endif

%ifdef CFG_VIDEO_VESA12
     call  vesa12_read_screen_pixel
     ret
%endif

%ifdef CFGSYS_VIDEO_VESA_12AND20

   rspv20:

%endif

%ifdef CFG_VIDEO_VESA20
     call  vesa20_read_screen_pixel
     ret
%endif


%ifdef CFG_VIDEO_VESA20
vesa20_read_screen_pixel:


     and   eax,0x3FFFFF

     cmp   [0xfbf1],byte 24      ; 24 or 32 bpp ?
     jz    rsp24

     mov   ebx,eax

     shl   ebx,2
     add   ebx,[0xfe80]

     mov   eax,[ebx]
     and   eax,0x00ffffff

     ret


  rsp24:

     xor   edx,edx
     mov   ebx,3
     mul   ebx

     mov   ebx,eax
     add   ebx,[0xfe80]

     mov   eax,[ebx]
     and   eax,0x00ffffff

     ret
%endif ; CFG_VIDEO_VESA20


%ifdef CFG_VIDEO_VESA12
vesa12_read_screen_pixel:


     and   eax,0x3FFFFF

     cmp   [0xfbf1],byte 24      ; 24 or 32 bpp ?
     jz    v12rsp24

     mov   edi,eax
     shl   edi,2

     mov   eax,edi
     
     shr   eax,16
     call  set_bank
     and   edi,65535
     add   edi,0xa0000

     mov   eax,[edi]
     and   eax,0x00ffffff

     ret


  v12rsp24:

     xor   edx,edx
     mov   ebx,3
     mul   ebx

     mov   edi,eax
     
     shr   eax,16
     call  set_bank
     and   edi,65535
     add   edi,0xa0000

     mov   eax,[edi]
     and   eax,0x00ffffff

     ret
%endif ; CFG_VIDEO_VESA12


%ifdef CFG_READSTRING
read_string:

    ; eax  read_area
    ; ebx  color of letter
    ; ecx  color of background
    ; edx  number of letters to read
    ; esi  [x start]*65536 + [y_start]

    ret
%endif ; CFG_READSTRING

display_settings:

;    eax = 0         ; DISPLAY redraw
;          ebx = 0   ; all
;
;    eax = 1         ; BUTTON type
;          ebx = 0   ; flat
;          ebx = 1   ; 3D 



     pusha

     cmp  eax,0       ; redraw display
     jne  dspl0
     cmp  ebx,0
     jne  dspl0
     cmp  [buttontypechanged],dword 1
     jne  dspl00
     mov  [buttontypechanged],dword 0
     mov  [dlx],dword 0
     mov  [dly],dword 0
     mov  eax,[0xfe00]
     mov  [dlxe],eax
     mov  eax,[0xfe04]
     mov  [dlye],eax
     mov  esi,1
     call redrawscreen
   dspl00:
     popa
     ret
   dspl0:
	 
     cmp  eax,1       ; button type
     jne  dspl1
     and  ebx,1
     cmp  ebx,[buttontype]
     je   dspl9
     mov  [buttontype],ebx
     mov  [buttontypechanged],dword 1
    dspl9:
     popa
     ret
   dspl1:

     popa

     ret




readmousepos:


; eax=0 screen relative
; eax=1 window relative
; eax=2 buttons pressed


    cmp  eax,0
    jnz  nosr
    mov  eax,[0xfb0a]
    shl  eax,16
    mov  ax,[0xfb0c]
    ret
  nosr:

    cmp  eax,1
    jnz  nowr
    mov  eax,[0xfb0a]
    shl  eax,16
    mov  ax,[0xfb0c]
    mov  esi,[0x3010]
    sub  esi,0x1000
    mov  bx,[esi]
    shl  ebx,16
    mov  bx,[esi+4]
    sub  eax,ebx
    ret
  nowr:

    cmp   eax,2
    jnz   nomb
    movzx eax,byte [0xfb40]
    ret
  nomb:

    ret



%ifdef CFG_SOUND_MIDI
is_input:

   push edx

   mov  dx,[midisp]
   in   al,dx
   and  al,0x80

   pop  edx
   ret


is_output:

   push edx

   mov  dx,[midisp]
   in   al,dx
   and  al,0x40

   pop  edx
   ret


get_mpu_in:

   push edx

   mov  dx,[mididp]
   in   al,dx

   pop  edx
   ret


put_mpu_out:

   push edx

   mov  dx,[mididp]
   out  dx,al

   pop  edx
   ret


setuart:

 su1:
   call is_output
   cmp  al,0
   jnz  su1

   mov  dx,[midisp]
   mov  al,0xff
   out  dx,al

 su2:
   mov  dx,[midisp]
   mov  al,0xff
   out  dx,al

   call is_input
   cmp  al,0
   jnz  su2

   call get_mpu_in
   cmp  al,0xfe
   jnz  su2

 su3:
   call is_output
   cmp  al,0
   jnz  su3

   mov  dx,[midisp]
   mov  al,0x3f
   out  dx,al



   ret

sys_midi:

     cmp  [mididp],word 0
     jnz  sm0

     mov  eax,1
     ret

   sm0:

     cmp  eax,1
     jnz  smn1

   sm1:
     call setuart
     mov  eax,0

     ret

   smn1:

     cmp  eax,2
     jnz  smn2

   sm10:
     call get_mpu_in
     call is_output
     cmp  al,0
     jnz  sm10

     mov  al,bl
     call put_mpu_out
     mov  eax,0

     ret


   smn2:

     cmp  eax,3
     jnz  smn3


     mov  eax,0
     ret

   smn3:

     mov  eax,0
     ret
%endif ; CFG_SOUND_MIDI



detect_devices:

    ret



sys_end:

     mov   eax,[0x3010]
     add   eax,0xa
     mov   [eax],byte 4  ; terminate this program
   waitterm:             ; wait here for termination
     mov   eax,5
     call  delay_hs
     jmp   waitterm


sys_system:

     cmp  eax,1
     jnz  nosystemboot

     mov  [0xFF00],byte 1
     mov  eax,0

     ret

   nosystemboot:


     cmp  eax,2
     jnz  noprocessterminate

     mov  edx,3
     cmp  ebx,edx
     jb   noprocessterminate
     mov  edx,[0x3004]
     cmp  ebx,edx
     jg   noprocessterminate

     mov  eax,[0x3004]
     shl  ebx,5
     add  ebx,0x3000+0xa
     mov  [ebx],byte 3  ; mark as zombie, terminate this program

   wterm:
     call change_task
     cmp  eax,[0x3004]
     jz   wterm

     ret

   noprocessterminate:


     cmp  eax,3
     jnz  nowindowactivate

     cmp  ebx,2
     jb   nowindowactivate
     cmp  ebx,[0x3004]
     jg   nowindowactivate

     ; edi = position at 0x2000+

     mov  edi,ebx

     shl  ebx,1
     add  ebx,0xc000
     mov  esi,[ebx]
     and  esi,65535

     xor  edx,edx
     mov  dx,[0x3004]
     cmp  esi,edx
     jz   nowindowactivate

     mov  [0xff01],edi
     mov  eax,0

     ret

   nowindowactivate:


     cmp  eax,4
     jnz  nogetidletime

     mov  eax,[idleusesec]

     ret

   nogetidletime:


     cmp  eax,5
     jnz  nogettscsec

     mov  eax,[0xf600]

     ret

   nogettscsec:

     ret


%ifdef CFG_CDAUDIO
sys_cd_audio:

     cmp  [cdbase],word 0
     jnz  cdcon

     mov  eax,1
     ret

   cdcon:

     ; eax=1 cdplay at ebx 0x00FFSSMM         
     ; eax=2 get tracklist size of ecx to [ebx]
     ; eax=3 stop/pause playing

     cmp  eax,1
     jnz  nocdp

     call sys_cdplay
     ret

   nocdp:

     cmp eax,2
     jnz nocdtl

     mov edi,[0x3010]
     add edi,0x10
     add ebx,[edi]

     call sys_cdtracklist

     ret

   nocdtl:

     cmp eax,3
     jnz nocdpause

     call sys_cdpause

     ret

   nocdpause:

     mov eax,0xffffff01

     ret



sys_cd_atapi_command:


     pusha

     mov  dx,[cdbase]
     add  dx,6
     mov  ax,[cdid]
     out  dx,al

     mov  esi,10
     call delay_ms

     mov  dx,[cdbase]
     add  dx,7
     in   al,dx
     and  al,$80
     cmp  al,0
     jnz  res
     jmp  cdl6

   res:

     mov dx,[cdbase]
     add dx,7
     mov al,$8
     out dx,al

     mov dx,[cdbase]
     add dx,0x206
     mov al,0xe
     out dx,al

     mov  esi,1
     call delay_ms

     mov dx,[cdbase]
     add dx,0x206
     mov al,$8
     out dx,al

     mov  esi,30
     call delay_ms

     xor  cx,cx

   cdl5:

     inc  cx
     cmp  cx,200
     jz   cdl6

     mov  dx,[cdbase]
     add  dx,7
     in   al,dx
     and  al,$88
     cmp  al,$00
     jz   cdl5

     mov  esi,10
     call delay_ms

     jmp  cdl5

   cdl6:

     mov dx,[cdbase]
     add dx,4
     mov al,0
     out dx,al

     mov dx,[cdbase]
     add dx,5

     mov al,0
     out dx,al

     mov dx,[cdbase]
     add dx,7

     mov al,0xec
     out dx,al

     mov  esi,5
     call delay_ms

     mov dx,[cdbase]
     add dx,1
     mov al,0
     out dx,al

     add dx,1
     mov al,0
     out dx,al

     add dx,1
     mov al,0
     out dx,al

     add dx,1
     mov al,0
     out dx,al

     add dx,1
     mov al,128
     out dx,al

     add dx,2
     mov al,0xa0
     out dx,al

     xor  cx,cx
     mov  dx,[cdbase]
     add  dx,7

   cdl1:
     inc  cx
     cmp  cx,100
     jz   cdl2

     in   al,dx
     and  ax,$88
     cmp  al,$08
     jz   cdl2

     mov  esi,2
     call delay_ms

     jmp  cdl1

   cdl2:

     popa

     ret





sys_cdplay:

     mov  ax,5
     push ax

     push ebx

   cdplay:

     call sys_cd_atapi_command

     cli

     mov  dx,[cdbase]

     mov  ax,$0047
     out  dx,ax

     mov  al,1
     mov  ah,[esp+0] ; min ??
     out  dx,ax
     mov  ax,[esp+1] ; fr sec
     out  dx,ax
     mov  ax,256+99
     out  dx,ax
     mov  ax,$0001
     out  dx,ax
     mov  ax,$0000
     out  dx,ax

     mov  esi,10
     call delay_ms

     sti

     add  dx,7
     in   al,dx

     test al,1
     jz   cdplayok

     cld
     mov  ax,[esp+4]
     dec  ax
     mov  [esp+4],ax
     cmp  ax,0
     jz   cdplayfail

     jmp  cdplay

   cdplayfail:
   cdplayok:

     pop  ebx
     pop  ax

     mov  eax,0

     ret





sys_cdtracklist:

     ;  get tracklist

     push ebx

   tcdplay:

     call sys_cd_atapi_command

     mov  dx,[cdbase]

     mov  ax,$43+2*256
     out  dx,ax
     mov  ax,$0
     out  dx,ax
     mov  ax,$0
     out  dx,ax
     mov  ax,$0
     out  dx,ax
     mov  ax,200
     out  dx,ax
     mov  ax,$0
     out  dx,ax

     in   al,dx

     mov  cx,1000
     mov  dx,[cdbase]
     add  dx,7
     cld

   cdtrnwewait:
     mov  esi,10
     call delay_ms
     in   al,dx
     and  al,128
     cmp  al,0
     jz   cdtrl1

     loop cdtrnwewait

   cdtrl1:

     ; read the result

     mov  ecx,[esp+0]

     mov  dx,[cdbase]

   cdtrread:
     add  dx,7
     in   al,dx
     and  al,8
     cmp  al,8
     jnz  cdtrdone

     sub  dx,7
     in   ax,dx
     mov  [ecx],ax

     add  ecx,2

     jmp  cdtrread

   cdtrdone:

     pop  ecx
     mov  eax,0
     ret



sys_cdpause:

     call sys_cd_atapi_command

     mov  dx,[cdbase]
     mov  ax,$004B
     out  dx,ax
     mov  ax,0
     out  dx,ax
     mov  ax,0
     out  dx,ax
     mov  ax,0
     out  dx,ax
     mov  ax,0
     out  dx,ax
     mov  ax,0
     out  dx,ax

     mov  esi,10
     call delay_ms

     add  dx,7
     in   al,dx

     mov  eax,0
     ret

%endif ; CFG_CDAUDIO


sys_cachetodiskette:


    pusha

    popa

    ret


bgrchanged  dd  0x0


sys_background:

    cmp   eax,1
    jnz   nosb1
    mov   [0x400000-8],ebx
    mov   [0x400000-4],ecx
    mov   [bgrchanged],dword 1
    ret
  nosb1:
    cmp   eax,2
    jnz   nosb2
    mov   edx,0x100000-16
    cmp   edx,ebx
    jbe   nosb2
    mov   edx,[ebx]
    and   edx,255*256*256*256
    and   ecx,255*256*256+255*256+255
    add   edx,ecx
    mov   [ebx+0x300000],edx
    mov   [bgrchanged],dword 1
    ret
  nosb2:
    cmp   eax,3
    jnz   nosb3
    cmp   [bgrchanged],dword 0
    je    nosb31
    mov   [bgrchanged],dword 0
    mov   [0xfff0],byte 1
   nosb31:
    ret
  nosb3:
    cmp   eax,4
    jnz   nosb4
    cmp   ebx,[0x400000-12]
    je    nosb41
    mov   [0x400000-12],ebx
    mov   [bgrchanged],dword 1
   nosb41:
    ret
  nosb4:
    cmp   eax,5
    jnz   nosb5
    mov   edi,[0x3010]
    add   ebx,[edi+0x10]
    mov   esi,ebx
    mov   edi,ecx
    add   ecx,edx
    cmp   ecx,0x100000-16
    jbe   nsb52
    ret
   nsb52:
    add   edi,0x300000
    mov   ecx,edx
    cmp   ecx,0x100000-16
    jbe   nsb51
    ret
   nsb51:
    mov   [bgrchanged],dword 1
    cld
    rep   movsb
    ret
  nosb5:

    ret



sys_getbackground:


    cmp   eax,1
    jnz   nogb1
    mov   eax,[0x400000-8]
    shl   eax,16
    mov   ax,[0x400000-4]
    ret
  nogb1:

    cmp   eax,2
    jnz   nogb2
    mov   edx,0x100000-16
    cmp   edx,ebx
    jbe   nogb2
    mov   eax,[ebx+0x300000]
    and   eax,255*256*256+255*256+255
    ret
  nogb2:

    cmp   eax,4
    jnz   nogb4
    mov   eax,[0x400000-12]
    ret
  nogb4:

    ret




sys_getkey:

    movzx ebx,word [0x3000]     ; top of windowing stack ?
    shl   ebx,1
    add   ebx,0xc000
    movzx ecx,word [ebx]
    mov   edx,[0x3004]
    cmp   ecx,edx
    je    sysgkl1
    mov   eax,1 
    ret

  sysgkl1:

    mov  al,[0xf400]
    cmp  al,0
    jnz  gkc1 
    mov  eax,1  ; no key in buffer
    ret
  gkc1:
    xor  eax,eax
    mov  ah,[0xf401]
    mov  [0xf400],byte 0
    mov  al,0
    ret


sys_getbutton:

    xor  eax,eax
    xor  ebx,ebx
    xor  ecx,ecx
    xor  edx,edx

    mov  bx,[0x3000]     ; top of windowing stack ?
    shl  ebx,1
    add  ebx,0xc000
    mov  cx,[ebx]
    mov  dx,[0x3004]

    cmp  cx,dx
    jz   gbot
    mov  eax,1
    ret
  gbot:
    mov  al,[0xf500]
    mov  bl,0
    cmp  al,bl
    jnz  gbc1
    mov  eax,1           ; no button in buffer
    ret
  gbc1:
    xor  eax,eax
    mov  ah,[0xf501]
    mov  [0xf500],byte 0
    mov  al,0
    ret



sys_cpuusage:

;  returns
;
;  10 name
;
;  22 dword - start in mem
;  26 dword - used mem
;

    push ebx
    push ebx
    push eax

    ; ** return memory usage

    xor  edx,edx
    mov  eax,0x20
    mul  ebx
    add  eax,0x3000+0x1c
    mov  ebx,eax

    pop  eax

    mov  ecx,[ebx]
    mov  [eax],ecx

    mov  ebx,[esp]

    shl  ebx,1
    add  ebx,0xc000
    mov  cx,[ebx]
    mov  [eax+4],cx

    mov  ebx,[esp]

    shl  ebx,1
    add  ebx,0xc400
    mov  cx,[ebx]
    mov  [eax+6],cx

    pop  ebx
    push eax

    mov  eax,ebx
    shl  eax,8
    add  eax,0x80000

    pop  ebx
    add  ebx,10

    mov  ecx,11
    call memmove

    ; ** return memory usage

    mov    eax,0
    mov    edx,0x100000*16-4096

    pop    ecx                                   ; get gdt of tss
    cmp    ecx,1
    je     os_mem

    shl    ecx,8
    add    ecx,0x80000+0x88
    mov    ecx,[ecx]
    shl    ecx,3   

    ; eax run base -> edx used memory

    mov    al,[ecx+gdts+ app_code +4]                                  ;  base  23:16
    mov    ah,[ecx+gdts+ app_code +7]                                  ;  base  31:24
    shl    eax,16
    mov    ax,[ecx+gdts+ app_code +2]                                  ;  base  0:15

    movzx  edx,word [ecx+gdts+ app_code +0]
    shl    edx,12

  os_mem:
    add    ebx,12
    mov    [ebx+0],eax
    mov    [ebx+4],edx
  
   

    ; ** return number of processes

    mov  eax,[0x3004] 

    ret



; check if image is allowed to be drawn


checkimage:

	push  eax
	push  ebx
	push  ecx
	push  edx

	xor   edx,edx
	xor   ecx,ecx
	mov   cx,[0x3000]
	shl   ecx,1
	add   ecx,0xc000
	mov   dx,[ecx]
	mov   ax,[0x3004]
	cmp   dx,ax
	jz    imok

	jmp   imcheckinside

      imok:                  ; first in stack

	pop   edx
	pop   ecx
	pop   ebx
	pop   eax

	mov   ecx,0
	ret

      imcheckinside:

	mov   esi,edx        ; window of image -position in windowing stack

	pop   edx
	pop   ecx
	pop   ebx
	pop   eax

	call  cilimit

	ret




; check image limits

cilimit:

	push  eax
	push  ebx

	xor  eax,eax
	xor  ebx,ebx

	mov al,[0xe000]
	mov bl,byte 1
	cmp eax,ebx
	jz  cilc

	pop ebx
	pop eax

	mov ecx,0
	ret

      cilc:

	push  ecx
	push  edx

	mov   edi,[0x3010]
	mov   ecx,0xd000-0x3000
	add   edi,ecx

	mov   eax,[esp+12]
	mov   ebx,[esp+08]
	mov   ecx,[esp+04]
	mov   edx,[esp+00]

	mov   ecx,[edi+12]  ;ecx = area y end   edx = image y end
	cmp   edx,ecx
	jbe   cici1
	jmp   cicino
      cici1:
	mov   ecx,[edi+4]  ;ecx = area y start   ebx = image y start
	cmp   ecx,ebx
	jbe   cici2
	jmp   cicino
      cici2:

	mov   eax,[esp+12]
	mov   ebx,[esp+08]
	mov   ecx,[esp+04]
	mov   edx,[esp+00]

	mov   edx,[edi+8]
	cmp   ecx,edx
	jbe   cici3
	jmp   cicino
      cici3:
	mov   edx,[edi+0]
	cmp   edx,eax
	jbe   cici4
	jmp   cicino
      cici4:

	pop   edx
	pop   ecx
	pop   ebx
	pop   eax

	mov   ecx,0     ;inside of draw limits

	ret

      cicino:

	pop   edx
	pop   ecx
	pop   ebx
	pop   eax

	mov   ecx,1     ;outside of draw limits

	ret

sys_putimage:

%ifdef CFGSYS_VIDEO_VESA_12AND20

     cmp   [0xfe0c],word 0100000000000000b
     jge   piv20
     cmp   [0xfe0c],word 0x13
     je    piv20

%endif

%ifdef CFG_VIDEO_VESA12
     call  vesa12_putimage
     ret
%endif

%ifdef CFGSYS_VIDEO_VESA_12AND20

   piv20:

%endif

%ifdef CFG_VIDEO_VESA20
     call  vesa20_putimage
     ret
%endif



%ifdef CFG_VIDEO_VESA20
vesa20_putimage:


	mov [novesachecksum],dword 0

	push ebp ;
	push esi ;
	push edi ;

	push eax ; 
	push ebx ; +8 [ptrImage]
	push ecx ; +4 [BH]
	push edx ; +0 [xy]

	movzx eax,word [esp+2]   ; eax:=x
	movzx ebx,word [esp+0]   ; ebx:=y
	mov ecx,[0x3010]         ; 
	add eax,[ecx-0x1000]     ; eax+=Xwin
	add ebx,[ecx-0x1000+4]   ; ebx+=Ywin

	xor edx, edx
	mov ecx, eax             ; ecx:=(x+Xwin)
	mov eax, [0xfe08]        ; BytesPerScanLine
	mul ebx                  ; *(y+Ywin)
	mov edi, eax             ; edi:=BytesPerScanLine*(y+Ywin)

	xor edx, edx
	mov eax, [0xfe00]        ; ScreenXSize
	inc eax                  ; +1
	mul ebx                  ; *(y+Ywin)
	mov ebp, eax             ; 
	add ebp, ecx             ; +(x+Win)
	add ebp, 0x400000        ; ebp:=(y+Ywin)*(ScreenXSize+1)+(x+Xwin)+AddrBuffer

	add edi,ecx ; +x
	add edi,ecx ; +x
	add edi,ecx ; +x

	cmp [0xfbf1],byte 24 ; 24 or 32 bpp ? - x start
	jz pi24
	add edi,ecx ; +x
      pi24:

	add edi,[0xfe80]          ; +AddrLFB

	;entry point in LFB >> EDI:=(y+Ywin)*BytesPerScanLine+X*BytesPerPixel+AddrLFB 

	; x size
 
	movzx eax,word [esp+6] ; eax:=B
	mov ecx,eax ; ecx:=B
	add ecx,eax ; +B
	add ecx,eax ; +B

	cmp [0xfbf1],byte 24 ; 24 or 32 bpp ? - x size
	jz pi243 
	add ecx,eax ; +B
      pi243:
	; ecx:=B*PytesPerPixel
	mov esi,[esp+8] ; esi:=AddrImg

	movzx ebx,word [esp+4] ; ebx:=H

	; check limits while draw ?

	push eax ; B
	push ecx ; B*PytesPerPixel

	mov eax,[0x3010]

	mov ecx,[eax+0xd000-0x3000+0]
	cmp ecx,0
	jnz dbcblimitlset2

	mov ecx,[eax+0xd000-0x3000+4]
	cmp ecx,0
	jnz dbcblimitlset2

	mov ecx,[eax+0xd000-0x3000+8]
	cmp ecx,[0xfe00] ; ecx <> Screen X size
	jnz dbcblimitlset2

	mov ecx,[eax+0xd000-0x3000+12]
	cmp ecx,[0xfe04] ; ecx <> Screen Y size
	jnz dbcblimitlset2

	pop ecx ; B*BytesPerScanLine
	pop eax ; B

	push dword 0 

	jmp dbcblimitlno2

      dbcblimitlset2:

	pop ecx ; B*BytesPerScanLine
	pop eax ; B

	push dword 1

      dbcblimitlno2:


	cmp [0xfbf1],byte 24 ; 24 or 32 bpp ?
	jz pi24bit
	jmp pi32bit



pi24bit:

	cld
	mov ecx,eax

      newpi:

	push edi
	push esi
	push ecx
	push ebx

	; edi:=(y+Ywin)*BytesPerScanLine+X*BytesPerPixel+AddrLFB 

	mov ebx,[0x3010]
	mov bl,[ebx+0xe]
	mov bh,[esp+4*4]

	np24:

	   cmp bl,[ds:ebp]  ; bl(No currentWin) <> Buffer_NoWin
	   jnz imp24no

	   mov edx,[esi]

	   cmp bh,0
	   jz  imp24yes

	   call voodoodbcplimit
	   jnz imp24no

	 imp24yes:

	   mov [edi],dx
	   shr edx,16
	   mov [edi+2],dl

	 imp24no:

	   add esi,3  ; ptrImage+=3
	   add edi,3  ; ptrScreen+=3
	   inc ebp    ; ptrBuffer++

	   dec ecx    ; B--
	   jnz np24   ; B <> 0


	pop ebx
	pop ecx
	pop esi
	pop edi


	add edi,[0xfe08] ; ptrScreen+=BytesPerScanLine

	xor eax,eax
	mov ax,[esp+4+2+4] ; eax:=B

	sub ebp, eax
	add ebp, [0xfe00] ;
	inc ebp ; ptrBuffer:=ptrBuffer-B+Screen_Xsize+1

	lea eax,[eax+eax*2] ;
	add esi,eax ; ptrImage:=ptrImage+B*3

	dec ebx ; H--
	jnz newpi ; until H=0

	add esp,7*4
	pop ebp

	mov eax,0

	ret


pi32bit:

	cld
	shr ecx,2

      newpi32:

	push edi
	push esi
	push ecx
	push ebx

	mov ebx,[0x3010]
	mov bl,[ebx+0xe]
	mov bh,[esp+4*4]

	cld

	np32:

	  cmp bl,[ds:ebp]
	  jnz imp32no

	  mov edx,[esi]

	  cmp bh,0
	  jz imp32yes

	  call voodoodbcplimit
	  jnz imp32no

	imp32yes:

	  mov [edi],dx
	  shr edx,16
	  mov [edi+2],dl

	imp32no:

	  add esi,3 ; esi+=3 ptrImage+=3
	  add edi,4 ; edi+=4 ptrScreen+=4
	  inc ebp

	  dec ecx ; B--
	  jnz np32


	 np32d:
 
 
      pop ebx
      pop ecx
      pop esi
      pop edi 


      add edi,[0xfe08] ; ptrScreen+=BytesPerScanLine

      xor eax,eax
      mov ax,[esp+4+2+4] ; B

      sub ebp, eax ; 
      add ebp, [0xfe00] ;
      inc ebp ; ptrBuffer:=ptrBuffer-B+Screen_Xsize+1

      lea eax,[eax+eax*2] ; 
      add esi,eax ; ptrImage:=ptrImage+B*3

      dec ebx ; H--
      jnz newpi32

      add esp,7*4
      pop ebp

      mov eax,0

      ret
%endif ; CFG_VIDEO_VESA20


%ifdef CFG_VIDEO_VESA12
vesa12_putimage:

;    mov  ebx,image
;    mov  ecx,320*65536+240
;    mov  edx,20*65536+20


    mov   [novesachecksum],dword 0

    push  esi
    push  edi

    push  eax
    push  ebx
    push  ecx
    push  edx

    movzx eax,word [esp+2]
    movzx ebx,word [esp+0]
    mov   ecx,[0x3010]
    add   eax,[ecx-0x1000]
    add   ebx,[ecx-0x1000+4]

    push  eax
    mov   eax,ebx         ; y
    mov   ebx,[0xfe08]
    mul   ebx
    pop   ecx

    add   eax,ecx         ; x
    add   eax,ecx
    add   eax,ecx

    cmp  [0xfbf1],byte 24     ; 24 or 32 bpp ? - x start
    jz   pi2412
    add  eax,ecx
  pi2412:

    add  eax,[0xfe80]
    mov  edi,eax


    ; x size

    movzx eax,word [esp+6]
    mov   ecx,eax
    add   ecx,eax
    add   ecx,eax

    cmp   [0xfbf1],byte 24     ; 24 or 32 bpp ? - x size
    jz    pi24312
    add   ecx,eax
  pi24312:

    mov  esi,[esp+8]

    movzx ebx,word [esp+4]

    ; check limits while draw ?

    push  eax
    push  ecx

    mov  eax,[0x3010]
	
    mov  ecx,[eax+0xd000-0x3000+0]
    cmp  ecx,0
    jnz  dbcblimitlset212

    mov  ecx,[eax+0xd000-0x3000+4]
    cmp  ecx,0
    jnz  dbcblimitlset212

    mov  ecx,[eax+0xd000-0x3000+8]
    cmp  ecx,[0xfe00]
    jnz  dbcblimitlset212

    mov  ecx,[eax+0xd000-0x3000+12]
    cmp  ecx,[0xfe04]
    jnz  dbcblimitlset212

    pop  ecx
    pop  eax

    push dword 0

    jmp  dbcblimitlno212

  dbcblimitlset212:

    pop  ecx
    pop  eax

    push dword 1
    
  dbcblimitlno212:

    cmp  [0xfbf1],byte 24     ; 24 or 32 bpp ?
    jz   pi24bit12
    jmp  pi32bit12


  pi24bit12:

    cld

    push  eax
    push  ebx
    push  edx

    xor   edx,edx
    mov   eax,ecx
    mov   ebx,3
    div   ebx
    mov   ecx,eax

    pop   edx
    pop   ebx
    pop   eax

  newpi12:

    push edi
    push esi
    push ecx
    push ebx

      xor  edx,edx
      mov  eax,edi
      sub  eax,[0xfe80]
      mov  ebx,3
      div  ebx
      add  eax,0x400000

      mov  ebx,[0x3010]
      mov  bl,[ebx+0xe]

      mov  bh,[esp+4*4]

      np2412:

	cmp  bl,[eax]
	jnz  imp24no12

	mov  edx,[esi]

	cmp  bh,0
	jz   imp24yes12

	call dbcplimit
	jnz  imp24no12

     imp24yes12:

	push eax
	push edi

	mov  eax,edi
	sub  eax,[0xfe80]
	shr  eax,16
	call set_bank
	and  edi,0xffff
	add  edi,0xa0000

	mov  [edi],edx
	shr  edx,2
	mov  [edi+2],dl

	sti

	pop  edi
	pop  eax

     imp24no12:

	inc  eax

	add  esi,3
	add  edi,3

	dec  ecx
	jnz  np2412

      np24d12:


     pop  ebx
     pop  ecx
     pop  esi
     pop  edi

    add  edi,[0xfe08]

    xor  eax,eax
    mov  ax,[esp+4+2+4]
    lea  eax,[eax+eax*2]
    add  esi,eax

    dec  ebx
    jz   nonewpi12

    jmp  newpi12

  nonewpi12:

    add  esp,7*4

    mov  eax,0

    ret


  pi32bit12:

    cld
    shr  ecx,2

   newpi3212:

    push edi
    push esi
    push ecx
    push ebx

      mov  eax,edi
      sub  eax,[0xfe80]
      shr  eax,2
      add  eax,0x400000

      mov   ebx,[0x3010]
      mov   bl,[ebx+0xe]

      mov   bh,[esp+4*4]

      np3212:

	cmp  bl,[eax]
	jnz  imp32no12

	mov  edx,[esi]

	cmp  bh,0
	jz   imp32yes12
 
	call dbcplimit
	jnz  imp32no12

      imp32yes12:

	push eax
	push edi
	mov  eax,edi
	sub  eax,[0xfe80]
	shr  eax,16
	call set_bank
	and  edi,0xffff
	add  edi,0xa0000
	
	mov  [edi],edx  
	
	sti

	pop  edi
	pop  eax

      imp32no12:

	inc  eax

	add  esi,3
	add  edi,4

	dec  ecx
	jnz  np3212

      np32d12:


     pop  ebx
     pop  ecx
     pop  esi
     pop  edi

    add   edi,[0xfe08]

    movzx eax,word [esp+4+2+4]
    lea   eax,[eax+eax*2]
    add   esi,eax

    dec   ebx
    jz    nonewpi3212

    jmp   newpi3212

  nonewpi3212:

    add   esp,7*4

    mov   eax,0

    ret
%endif ; CFG_VIDEO_VESA12




drawbuttonframes:

	push  esi
	push  edi
	push  eax
	push  ebx
	push  ecx
	push  edx

	shr   eax,16
	shr   ebx,16
	mov   edx,[0x3010]
	add   eax,[edx-0x1000]
	add   ebx,[edx-0x1000+4]
	mov   cx,ax
	mov   dx,bx
	shl   eax,16
	shl   ebx,16
	mov   ax,cx
	mov   bx,dx
	mov   cx,word [esp+12]
	add   ax,cx
	mov   esi,ebx
	mov   edi,0
	mov   ecx,[esp+0]
	add   ecx,0x202020
	add   eax,1
	call  draw_line
	movzx edx,word [esp+8]
	add   ebx,edx
	shl   edx,16
	add   ebx,edx
	mov   ecx,[esp+0]
	shr   ecx,1
	and   ecx,127*256*256+127*256+127
	call  draw_line

	mov   ebx,esi
	push  edx
	mov   edx,eax
	shr   edx,16
	mov   ax,dx
	mov   edx,ebx
	shr   edx,16
	mov   bx,dx
	mov   dx,[esp+8+4]
	add   bx,dx
	pop   edx
	mov   edi,0
	mov   ecx,[esp+0]
	add   ecx,0x202020
	add   ebx,1
	call  draw_line
	mov   esi,edx
	mov   dx,[esp+12]
	add   ax,dx
	shl   edx,16
	add   eax,edx
	add   ebx,1*65536
	mov   edx,esi
	mov   ecx,[esp+0]
	shr   ecx,1
	and   ecx,127*256*256+127*256+127
	call  draw_line

	pop   edx
	pop   ecx
	pop   ebx
	pop   eax
	pop   edi
	pop   esi

	ret

button_dececx:

	cmp   [buttontype],dword 1
	je    bdece
	ret
      bdece:

	push  eax

	mov   eax,0x01
	cmp   edi,20
	jg    bdl9
	mov   eax,0x02
      bdl9:

	test  ecx,0xff
	jz    bdl1
	sub   ecx,eax
      bdl1:

	shl   eax,8
	test  ecx,0xff00
	jz    bdl2
	sub   ecx,eax
      bdl2:
	

	shl   eax,8
	test  ecx,0xff0000
	jz    bdl3
	sub   ecx,eax
      bdl3:

	pop    eax

	ret


sys_button:

	push  esi
	push  edi
	push  eax
	push  ebx
	push  ecx
	push  edx


;        pusha
;        movzx ecx,ax
;        movzx edx,bx
;        shr   eax,16
;        shr   ebx,16
;        add   ecx,eax
;        add   edx,ebx
;        mov   edi,[esp+32]
;        call  drawbar
;        popa

	pusha
	push  ebx
	shr   eax,16
	shr   ebx,16
	mov   edx,[0x3010]
	mov   esi,[edx-0x1000]
	mov   edi,[edx-0x1000+4]
	add   eax,esi
	add   ebx,edi
	mov   cx,ax
	mov   dx,bx
	shl   eax,16
	shl   ebx,16
	mov   ax,cx
	mov   bx,dx
	xor   ecx,ecx
	mov   cx,[4+32+esp+12]
	add   eax,ecx
	mov   ecx,[4+32+esp+0]
	
	cmp   [buttontype],dword 0
	je    bdecel1
	add   ecx,0x141414
       bdecel1:

	movzx edi,word [esp]
       bnewline:
	call  button_dececx
	push  edi
	mov   edi,0
	call  draw_line
	pop   edi
	add   ebx,1*65536+1
	xor   edx,edx
	mov   dx,[esp]
	dec   dx
	mov   [esp],dx
	cmp   dx,0
	jnz   bnewline
	pop   ebx
	popa


	call  drawbuttonframes

	cld

	mov   edi,[0x3010]
	sub   edi,0x1000
	cmp   [edi+30],byte 1
	jnz   addbutt
	jmp   noaddbutt

      addbutt: 
 
	mov   edi,[0xfe88]
	movzx eax,word [edi]
	cmp   eax,1000
	jge   noaddbutt
	inc   eax
	mov   [edi],ax

	shl   eax,4
	add   eax,edi

	mov   bx,[0x3000]
	mov   [eax],bx

	add   eax,2         ; save button id number
	mov   bx,[esp+4]
	mov   [eax],bx

	add   eax,2         ; x start
	mov   bx,[esp+12+2]
	mov   [eax],bx

	add   eax,2         ; x size
	mov   bx,[esp+12+0]
	mov   [eax],bx

	add   eax,2         ; y start
	mov   bx,[esp+8+2]
	mov   [eax],bx

	add   eax,2         ; y size
	mov   bx,[esp+8+0]
	mov   [eax],bx

     noaddbutt:

	pop   edx
	pop   ecx
	pop   ebx
	pop   eax
	pop   edi
	pop   esi

	ret


sys_clock:


	push  eax

	mov   [esp+3],byte 0

	mov   al,00     ; seconds
	out   0x70,al
	in    al,0x71
	mov   [esp+2],al

	mov   al,02     ; minutes
	out   0x70,al
	in    al,0x71
	mov   [esp+1],al

	mov   al,04     ; hours
	out   0x70,al
	in    al,0x71
	mov   [esp+0],al

	pop   eax


	ret

sys_date:


	push  eax

	mov   [esp+3],byte 0

	mov   al,07     ; date
	out   0x70,al
	in    al,0x71
	mov   [esp+2],al

	mov   al,08     ; month
	out   0x70,al
	in    al,0x71
	mov   [esp+1],al

	mov   al,09     ; year
	out   0x70,al
	in    al,0x71
	mov   [esp+0],al

	pop   eax


	ret




; redraw status

sys_redrawstat:

    cmp   eax,2
    jnz   srl1

    mov   edx,[0x3010]      ; return whole screen draw area for this app
    mov   eax,0xd000-0x3000
    add   edx,eax
    mov   eax,dword 0
    mov   [edx+0],eax
    mov   eax,dword 0
    mov   [edx+4],eax
    mov   eax,[0xfe00]
    mov   [edx+8],eax
    mov   eax,[0xfe04]
    mov   [edx+12],eax

    mov   edi,[0x3010]
    sub   edi,0x1000
    mov   [edi+30],byte 1   ; no new position & buttons from app


  srl1:

    ret



sys_drawwindow_I:

sys_drawwindow_II:

    push  eax
    push  ebx
    push  ecx
    push  edx

    mov   eax,[0x3000]
    shl   eax,5
    mov   edi,eax
    add   edi,0x2000

    pop   edx
    pop   ecx
    pop   ebx
    pop   eax

    mov   [edi+16],ecx
    mov   [edi+20],edx
    mov   [edi+24],esi

    cmp   [edi+30],byte 1
    jz    newd

    mov   [edi+08],ax
    mov   [edi+12],bx
    shr   eax,16
    shr   ebx,16
    mov   [edi+00],ax
    mov   [edi+04],bx

    pusha                   ; wait for putimages to finish
    mov   eax,1
    call  delay_hs
    popa

    pusha
    xor   eax,eax
    xor   ebx,ebx
    xor   ecx,ecx
    xor   edx,edx
    mov   ax,[edi+00]
    mov   bx,[edi+04]
    mov   cx,[edi+08]
    mov   dx,[edi+12]
    add   cx,ax
    add   dx,bx
    call  calculatescreen
    popa

  newd:
    mov   [edi+31],byte 0   ; no redraw

    mov   edx,edi

    call  drawwindow_I

    ret



; EVENTS FOR APPS



sys_getevent:

     call   get_event_for_app

     ret



sys_waitforeventwithtimeout:

     mov   ebx,[0xfdf0]
     add   ebx,eax

   newwait_time:

     cmp   ebx,[0xfdf0]
     jg    swfet1

     mov   eax,0
     jmp   eventoccur_time

   swfet1:

     call  get_event_for_app
     cmp   eax,0
     jne   eventoccur_time

     call  change_task

     jmp   newwait_time

   eventoccur_time:

     ret




sys_waitforevent:


   newwait:

     call  get_event_for_app
     cmp   eax,0
     jne   eventoccur

     call  change_task

     jmp   newwait

   eventoccur:

     ret




get_event_for_app:


     pusha


     mov   edi,[0x3010]     ; window draw
     test  [edi],dword 1
     jz    no_eventoccur1
     mov   edi,[0x3010]   
     sub   edi,0x1000
     mov   al,[edi+31]
     cmp   al,0
     jz    no_eventoccur1
     popa
     mov   eax,1
     ret
   no_eventoccur1:


     mov   edi,[0x3010]     ; key in buffer
     test  [edi],dword 2
     jz    no_eventoccur2
     xor   edx,edx
     xor   ecx,ecx
     mov   cx,[0x3000]
     shl   ecx,1
     add   ecx,0xc000
     mov   dx,[ecx]
     mov   ax,[0x3004]
     cmp   dx,ax
     jnz   no_eventoccur2
     mov   ecx,2
     mov   al,[0xf400]
     cmp   al,0
     jz    no_eventoccur2
     popa
     mov   eax,2
     ret
   no_eventoccur2:


     mov   edi,[0x3010]     ; button
     test  [edi],dword 4
     jz    no_eventoccur3
     xor   edx,edx
     xor   ecx,ecx
     mov   cx,[0x3000]
     shl   ecx,1
     add   ecx,0xc000
     mov   dx,[ecx]
     mov   ax,[0x3004]
     cmp   dx,ax
     jnz   no_eventoccur3
     mov   ecx,3
     mov   al,[0xf500]
     cmp   al,0
     jz    no_eventoccur3
     popa
     mov   eax,3
     ret
   no_eventoccur3:


     mov   edi,[0x3010]     ; desktop background redraw
     test  [edi],dword 16
     jz    no_eventoccur5
     cmp   [0xfff0],byte 2 
     jnz   no_eventoccur5
     popa
     mov   eax,5
     ret
   no_eventoccur5:



     mov   edi,[0x3010]
     test  [edi],dword 0x00010000
     jz    no_eventoccur16
     mov   edi,16 -16
     shl   edi,2
     add   edi,irq_owner
     mov   edx,[edi]
     mov   eax,[0x3010]
     mov   eax,[eax+0x4]
     cmp   edx,eax
     jne   no_eventoccur16
     cmp   [0x2e0000],dword 0
     jz    no_eventoccur16
     popa
     mov   eax,16
     ret
   no_eventoccur16:

     mov   edi,[0x3010]
     test  [edi],dword 0x00020000
     jz    no_eventoccur17
     mov   edi,17 -16
     shl   edi,2
     add   edi,irq_owner
     mov   edx,[edi]
     mov   eax,[0x3010]
     mov   eax,[eax+0x4]
     cmp   edx,eax
     jne   no_eventoccur17
     cmp   [0x2e1000],dword 0
     jz    no_eventoccur17
     popa
     mov   eax,17
     ret
   no_eventoccur17:

     mov   edi,[0x3010]
     test  [edi],dword 0x00040000
     jz    no_eventoccur18
     mov   edi,18 -16
     shl   edi,2
     add   edi,irq_owner
     mov   edx,[edi]
     mov   eax,[0x3010]
     mov   eax,[eax+0x4]
     cmp   edx,eax
     jne   no_eventoccur18
     cmp   [0x2e2000],dword 0
     jz    no_eventoccur18
     popa
     mov   eax,18
     ret
   no_eventoccur18:

     mov   edi,[0x3010]
     test  [edi],dword 0x00080000
     jz    no_eventoccur19
     mov   edi,19 -16
     shl   edi,2
     add   edi,irq_owner
     mov   edx,[edi]
     mov   eax,[0x3010]
     mov   eax,[eax+0x4]
     cmp   edx,eax
     jne   no_eventoccur19
     cmp   [0x2e3000],dword 0
     jz    no_eventoccur19
     popa
     mov   eax,19
     ret
   no_eventoccur19:

     mov   edi,[0x3010]
     test  [edi],dword 0x00100000
     jz    no_eventoccur20
     mov   edi,20 -16
     shl   edi,2
     add   edi,irq_owner
     mov   edx,[edi]
     mov   eax,[0x3010]
     mov   eax,[eax+0x4]
     cmp   edx,eax
     jne   no_eventoccur20
     cmp   [0x2e4000],dword 0
     jz    no_eventoccur20
     popa
     mov   eax,20
     ret
   no_eventoccur20:

     mov   edi,[0x3010]
     test  [edi],dword 0x00200000
     jz    no_eventoccur21
     mov   edi,21 -16
     shl   edi,2
     add   edi,irq_owner
     mov   edx,[edi]
     mov   eax,[0x3010]
     mov   eax,[eax+0x4]
     cmp   edx,eax
     jne   no_eventoccur21
     cmp   [0x2e5000],dword 0
     jz    no_eventoccur21
     popa
     mov   eax,21
     ret
   no_eventoccur21:

     mov   edi,[0x3010]
     test  [edi],dword 0x00400000
     jz    no_eventoccur22
     mov   edi,22 -16
     shl   edi,2
     add   edi,irq_owner
     mov   edx,[edi]
     mov   eax,[0x3010]
     mov   eax,[eax+0x4]
     cmp   edx,eax
     jne   no_eventoccur22
     cmp   [0x2e6000],dword 0
     jz    no_eventoccur22
     popa
     mov   eax,22
     ret
   no_eventoccur22:

     mov   edi,[0x3010]
     test  [edi],dword 0x00800000
     jz    no_eventoccur23
     mov   edi,23 -16
     shl   edi,2
     add   edi,irq_owner
     mov   edx,[edi]
     mov   eax,[0x3010]
     mov   eax,[eax+0x4]
     cmp   edx,eax
     jne   no_eventoccur23
     cmp   [0x2e7000],dword 0
     jz    no_eventoccur23
     popa
     mov   eax,23
     ret
   no_eventoccur23:

     mov   edi,[0x3010]
     test  [edi],dword 0x01000000
     jz    no_eventoccur24
     mov   edi,24 -16
     shl   edi,2
     add   edi,irq_owner
     mov   edx,[edi]
     mov   eax,[0x3010]
     mov   eax,[eax+0x4]
     cmp   edx,eax
     jne   no_eventoccur24
     cmp   [0x2e8000],dword 0
     jz    no_eventoccur24
     popa
     mov   eax,24
     ret
   no_eventoccur24:

     mov   edi,[0x3010]
     test  [edi],dword 0x02000000
     jz    no_eventoccur25
     mov   edi,25 -16
     shl   edi,2
     add   edi,irq_owner
     mov   edx,[edi]
     mov   eax,[0x3010]
     mov   eax,[eax+0x4]
     cmp   edx,eax
     jne   no_eventoccur25
     cmp   [0x2e9000],dword 0
     jz    no_eventoccur25
     popa
     mov   eax,25
     ret
   no_eventoccur25:

     mov   edi,[0x3010]
     test  [edi],dword 0x04000000
     jz    no_eventoccur26
     mov   edi,26 -16
     shl   edi,2
     add   edi,irq_owner
     mov   edx,[edi]
     mov   eax,[0x3010]
     mov   eax,[eax+0x4]
     cmp   edx,eax
     jne   no_eventoccur26
     cmp   [0x2eA000],dword 0
     jz    no_eventoccur26
     popa
     mov   eax,26
     ret
   no_eventoccur26:

     mov   edi,[0x3010]
     test  [edi],dword 0x08000000
     jz    no_eventoccur27
     mov   edi,27 -16
     shl   edi,2
     add   edi,irq_owner
     mov   edx,[edi]
     mov   eax,[0x3010]
     mov   eax,[eax+0x4]
     cmp   edx,eax
     jne   no_eventoccur27
     cmp   [0x2eB000],dword 0
     jz    no_eventoccur27
     popa
     mov   eax,27
     ret
   no_eventoccur27:

     mov   edi,[0x3010]
     test  [edi],dword 0x10000000
     jz    no_eventoccur28
     mov   edi,28 -16
     shl   edi,2
     add   edi,irq_owner
     mov   edx,[edi]
     mov   eax,[0x3010]
     mov   eax,[eax+0x4]
     cmp   edx,eax
     jne   no_eventoccur28
     cmp   [0x2eC000],dword 0
     jz    no_eventoccur28
     popa
     mov   eax,28
     ret
   no_eventoccur28:

     mov   edi,[0x3010]
     test  [edi],dword 0x20000000
     jz    no_eventoccur29
     mov   edi,29 -16
     shl   edi,2
     add   edi,irq_owner
     mov   edx,[edi]
     mov   eax,[0x3010]
     mov   eax,[eax+0x4]
     cmp   edx,eax
     jne   no_eventoccur29
     cmp   [0x2eD000],dword 0
     jz    no_eventoccur29
     popa
     mov   eax,29
     ret
   no_eventoccur29:

     mov   edi,[0x3010]
     test  [edi],dword 0x40000000
     jz    no_eventoccur30
     mov   edi,30 -16
     shl   edi,2
     add   edi,irq_owner
     mov   edx,[edi]
     mov   eax,[0x3010]
     mov   eax,[eax+0x4]
     cmp   edx,eax
     jne   no_eventoccur30
     cmp   [0x2eE000],dword 0
     jz    no_eventoccur30
     popa
     mov   eax,30
     ret
   no_eventoccur30:

     mov   edi,[0x3010]
     test  [edi],dword 0x80000000
     jz    no_eventoccur31
     mov   edi,31 -16
     shl   edi,2
     add   edi,irq_owner
     mov   edx,[edi]
     mov   eax,[0x3010]
     mov   eax,[eax+0x4]
     cmp   edx,eax
     jne   no_eventoccur31
     cmp   [0x2eF000],dword 0
     jz    no_eventoccur31
     popa
     mov   eax,31
     ret
   no_eventoccur31:


     popa                   ; no event
     mov   eax,0
     ret





; text

dtext:

; eax x & y
; ebx color
; ecx start of text
; edx length
; edi 1 force


	cmp    edx,0    ; zero length ?
	jnz    dsok
	ret

      dsok:

	pusha

      letnew:

	push   eax
	push   ecx
	push   edx

	movzx  ebx,ax
	shr    eax,16
	movzx  edx,byte [ecx]
	mov    ecx,[esp+3*4+32-16]
	call   drawletter

	pop    edx
	pop    ecx
	pop    eax

	add    eax,6*65536
	add    ecx,1
	dec    edx
	jnz    letnew

	popa

	ret




drawletter:

; eax  x
; ebx  y
; ecx  color
; edx  letter

	pusha

	mov   eax,0
	mov   ebx,0  ; 0x4000+eax+ebx*8
      chc:
	mov   edx,ebx
	shl   edx,3
	add   edx,eax
	add   edx,0x4000+8

	mov   ecx,[esp+32-12]
	shl   ecx,6
	add   edx,ecx

	cmp   [edx],byte 'o'
	jnz   nopix

	push  eax
	push  ebx

	add   eax,[esp+32+2*4-4]
	add   ebx,[esp+32+2*4-16]
	mov   ecx,[esp+32+2*4-8]
	call  putpixel

	pop   ebx
	pop   eax

      nopix:
	add   eax,1
	cmp   eax,5
	jnz   chc
	mov   eax,0
	add   ebx,1
	cmp   ebx,7
	jnz   chc

	popa

	ret



; fileread - sys floppy
;
; eax points to filename 11 chars
; ebx start of read in file
; ecx read nro of bytes
; edx mem location to return data
;
; ret eax = size or 0xffffffff file not found

fileread:


	; floppy

      frfloppy:
	push   eax
	push   ebx
	push   ecx
	push   edx
	push   esi
	push   edi

	mov    edi,0x100000+512*18+512  ;Point at directory
	mov    edx,224 +1

.20:    sub    edx,1
	cmp    edx,0
	jnz    .21
	mov    [esp+20],dword 0xffffffff
	jmp    frnoread
      .21
	mov    esi,eax                  ;Name of file we want
	xor    ecx,ecx
	mov    cl, 11
	cld
	rep    cmpsb                    ;Found the file?
	je     fifound                  ;Yes
	add    cl, 21                   ;Offset to next directory entry
	add    edi, ecx                 ;Advance to next entry
	jmp    .20
     fifound:

	mov    ebx,[edi-11+28]          ;file size
	mov    [esp+20],ebx

	add    edi,0xf
	mov    eax,[edi]
	and    eax,65535
	mov    edi,eax                  ;edi=cluster

      frnew:

	add    eax,31                   ;bootsector+2*fat+filenames
	shl    eax,9                    ;*512
	add    eax,0x100000             ;image base

	mov    ebx,[esp+8]
	mov    ecx,512                  ;[esp+4]
	call   memmove

	shl    edi,1                    ;find next cluster from FAT
	add    edi,0x280000
	mov    eax,[edi]
	and    eax,4095
	mov    edi,eax

	cmp    edi,dword 4095           ; last cluster ?
	jz     frnoread

	mov    ebx,[esp+8]
	add    ebx,512
	mov    [esp+8],ebx

	jmp    frnew

      frnoread:

	pop    edi
	pop    esi
	pop    edx
	pop    ecx
	pop    ebx
	pop    eax

	ret




; filedelete - sys floppy
;
; eax points to filename 11 chars


filedelete:


	push   eax
	push   ebx
	push   ecx
	push   edx
	push   esi
	push   edi

	mov    edi,0x100000+512*18+512  ;Point at directory
	mov    edx,224 +1

     .20d:
	sub    edx,1
	cmp    edx,0
	jnz    .21d
	jmp    frnoreadd
     .21d:
	mov    esi,eax                  ;Name of file we want
	mov    ecx, 11
	cld
	rep    cmpsb                    ;Found the file?
	je     fifoundd                 ;Yes
	add    cl, 21                   ;Offset to next directory entry
	add    edi, ecx                 ;Advance to next entry
	jmp    .20d
     fifoundd:

	mov    [edi-11],byte 0xE5       ;mark filename deleted

	add    edi,0xf
	mov    eax,[edi]
	and    eax,65535
	mov    edi,eax                  ;edi = cluster

      frnewd:

	shl    edi,1                    ;find next cluster from FAT
	add    edi,0x280000
	mov    eax,[edi]

	mov    [edi],word 0x0           ;clear fat chain cluster

	and    eax,4095
	mov    edi,eax

	cmp    edi,dword 4095           ;last cluster ?
	jz     frnoreadd2

	jmp    frnewd

      frnoreadd2:

	pop    edi
	pop    esi
	pop    edx
	pop    ecx
	pop    ebx
	pop    eax

	jmp    filedelete

      frnoreadd:

	pop    edi
	pop    esi
	pop    edx
	pop    ecx
	pop    ebx
	pop    eax

	ret


; filesave - sys floppy
;
; eax points to filename 11 chars
;
;        eax      ; pointer to file name
;        ebx      ; buffer
;        ecx      ; count to write in bytes
;        edx      ; 0 create new , 1 append


filesave:

	cmp  edx,0
	jnz  fsdel

	pusha
	call filedelete
	popa

      fsdel:

	push   eax
	push   ebx
	push   ecx
	push   edx
	push   esi
	push   edi

	mov    edi,0x100000+512*18+512  ;Point at directory
	mov    edx,224 +1

	; find an empty spot for filename in the root dir

     .20ds:
	sub    edx,1
	cmp    edx,0
	jnz    .21ds
	jmp    frnoreadds
     .21ds:
	cmp    [edi],byte 0xE5
	jz     fifoundds
	cmp    [edi],byte 0x0
	jz     fifoundds
	add    edi,32                   ; Advance to next entry
	jmp    .20ds
     fifoundds:

	push   edi                      ; move the filename to root dir  
	mov    esi,[esp+4+20]
	mov    ecx,11
	cld
	rep    movsb
	pop    edi

	mov    edx,edi                  
	add    edx,11+0xf               ; edx <- cluster save position

	mov    ebx,[esp+12]             ; save file size
	mov    [edi+28],ebx

	mov    ebx,1                    ; first cluster

      frnewds:

	add    ebx,1

	mov    edi,ebx                  ; find free cluster in FAT
	shl    edi,1                    
	add    edi,0x280000
	mov    eax,[edi]
	and    eax,4095
	cmp    eax,0x0
	jnz    frnewds

	mov    [edx],bx                 ; save next cluster pos. to prev cluster
	mov    edx,edi                  ; next save pos abs mem add

	pusha                           ; move save to floppy cluster
	add    ebx,31                   
	shl    ebx,9                    
	add    ebx,0x100000             
	mov    eax,[esp+32+16]
	mov    ecx,512                  
	call   memmove
	popa

	mov    eax,[esp+12]
	cmp    eax,512
	jb     flnsa
	sub    eax,512
	mov    [esp+12],eax

	mov    eax,[esp+16]
	add    eax,512
	mov    [esp+16],eax

	jmp    frnewds

     flnsa:
	mov    [edi],word 4095          ; mark end of file - last cluster

      frnoreadds:

	pop    edi
	pop    esi
	pop    edx
	pop    ecx
	pop    ebx
	pop    eax

	ret






; check pixel limits

cplimit:


	push eax
	push ebx

	mov  al,[0xe000]
	mov  bl,1
	cmp  al,bl
	jz   cplc

	pop  ebx
	pop  eax

	mov ecx,0
	ret

      cplc:

	pop  ebx
	pop  eax

	pusha

	mov edi,[0x3010]
	mov esi,0xd000-0x3000
	add edi,esi

	mov edx,[edi+0]    ;xs
	cmp edx,eax
	jbe cpl1
	popa
	mov ecx,1
	ret
      cpl1:
	mov edx,[edi+4]    ;ys
	cmp edx,ebx
	jbe cpl2
	popa
	mov ecx,1
	ret
      cpl2:
	mov edx,[edi+8]    ;xe
	cmp eax,edx
	jbe cpl3
	popa
	mov ecx,1
	ret
      cpl3:
	mov edx,[edi+12]   ;ye
	cmp ebx,edx
	jbe cpl4
	popa
	mov ecx,1
	ret
      cpl4:

	popa

	mov ecx,0          ; passed the limits

	ret



; check if pixel is allowed to be drawn


checkpixel:

	push  eax
	push  ebx

	mov   eax,[0x3000]
	shl   eax,5
	shl   eax,1
	add   eax,0xc000
	movzx edx,word [eax]

	movzx ecx,word [0x3004]
	cmp   edx,ecx
	jz    ncpok

	jmp   ncheckinside

      ncpok:                  ; first in stack

	pop   ebx
	pop   eax

	mov   ecx,0
	ret

      ncheckinside:

	pop   ebx
	pop   eax



	call  cplimit
	cmp   ecx,0
	jz    ncpc1
	ret

      ncpc1:

	push eax
	push ebx
	push edx

	cld
	push eax
	mov  eax,[0xfe00]
	inc  eax
	mul  ebx
	pop  ebx
	add  eax,ebx
	add  eax,0x400000

	xor  ecx,ecx
	mov  cl,[eax]

	mov  eax,[0x3000]

	shl  eax,5
	add  eax,0x3000+0xe
	movzx eax,byte[eax]

	cmp  eax,ecx
	jz   cpcok
	mov  ecx,1

	pop  edx
	pop  ebx
	pop  eax

	ret

      cpcok:
	mov  ecx,0

	pop  edx
	pop  ebx
	pop  eax

	ret




; ***************************************
; ******** VESA 2.0 PIXEL RELATED *******
; ***************************************


; putpixel
;
; eax = x coordinate
; ebx = y coordinate
; ecx = ?? RR GG BB    ; 0x01000000 negation
; edi = 0x00000001 force

putpixel:

	mov  [novesachecksum],dword 0

	pusha
	test edi,1                 ; force ?
	jz   noforce
	jmp  drawok2
      noforce:
	call checkpixel
	cmp  ecx,0
	jz   drawok2
	popa
	ret
      drawok2:
	mov edx,[0xfe00]
	cmp edx,eax
	jb  ppr
	mov edx,[0xfe04]
	cmp edx,ebx
	jb  ppr
	jmp ppok
      ppr:
	popa
	ret
      ppok:
	; check if negation
	test ecx,0x01000000
	jz   noneg
	call getpixel
	not  ecx
	mov  [esp+32-8],ecx
      noneg:
	; OK to set pixel
	call [0xe020]
	popa
	ret

%ifdef CFG_VIDEO_VESA20
Vesa20_putpixel24:

	mov  edi,eax ; x
	mov  eax,ebx ; y
	lea  edi,[edi+edi*2]
	mov  ebx,[0xfe08]
	mul  ebx
	add  edi,eax
	add  edi,[0xfe80]
	mov  eax,[esp+32-8+4]
	stosw
	shr  eax,16
	mov  [edi],al

	ret



Vesa20_putpixel32:

	mov  edi,eax ; x
	mov  eax,ebx ; y
	shl  edi,2
	mov  ebx,[0xfe08]
	mul  ebx
	add  edi,eax
	add  edi,[0xfe80]
	mov  ecx,[esp+32-8+4]
	mov  [edi],ecx

	ret
%endif ; CFG_VIDEO_VESA20




; getpixel
;
; in:
; eax = x coordinate
; ebx = y coordinate
;
; ret:
; ecx = 00 RR GG BB

getpixel:


	push    eax
	push    ebx
	push    edx
	push    edi

	call    [0xe024]

	pop     edi
	pop     edx
	pop     ebx
	pop     eax

	ret

%ifdef CFG_VIDEO_VESA20
Vesa20_getpixel24:

	mov  edi,eax ; x
	mov  eax,ebx ; y
	lea  edi,[edi+edi*2]
	mov  ebx,[0xfe08]
	mul  ebx
	add  edi,eax
	add  edi,[0xfe80]
	mov  ecx,[edi]
	and  ecx,0xffffff

	ret


Vesa20_getpixel32:

	mov  edi,eax ; x
	mov  eax,ebx ; y
	shl  edi,2
	mov  ebx,[0xfe08]
	mul  ebx
	add  edi,eax
	add  edi,[0xfe80]
	mov  ecx,[edi]
	and  ecx,0xffffff

	ret
%endif ; CFG_VIDEO_VESA20



%ifdef CFG_VIDEO_VESA12

; ***************************************
; ********* BANKED PIXEL RELATED ********
; ***************************************


set_bank:

   cli

   cmp al,[0xfff2]
   je retsb

   mov [0xfff2],al

   push ax
   push dx
   mov ah,al
   mov dx,0x03D4
   mov al,0x39
   out dx,al
   inc dl
   mov al,0xA5
   out dx,al
   dec dl
   mov al,6Ah
   out dx,al
   inc dl
   mov al,ah

   ;uncomment next line if smth 'll be wrong
   ; shl al,2

   out dx,al
   dec dl
   mov al,0x39
   out dx,al
   inc dl
   mov al,0x5A
   out dx,al
   dec dl
   pop dx
   pop ax

retsb:

  ret




Vesa12_putpixel24:

	mov  edi,eax ; x
	mov  eax,ebx ; y
	lea  edi,[edi+edi*2]
	mov  ebx,[0xfe08]
	mul  ebx
	add  edi,eax

	mov  eax,edi
	shr  eax,16
	call set_bank
	and  edi,65535
	add  edi,0xa0000

	mov  eax,[esp+28]
	stosw
	shr  eax,16
	mov  [edi],al
	sti

	ret



Vesa12_putpixel32:

	mov  edi,eax ; x
	mov  eax,ebx ; y
	shl  edi,2
	mov  ebx,[0xfe08]
	mul  ebx
	add  edi,eax

	mov  eax,edi
	shr  eax,16
	call set_bank
	and  edi,65535
	add  edi,0xa0000

	mov  ecx,[esp+28]
	mov  [edi],ecx
	sti

	ret




Vesa12_getpixel24:

	mov  edi,eax ; x
	mov  eax,ebx ; y
	lea  edi,[edi+edi*2]
	mov  ebx,[0xfe08]
	mul  ebx
	add  edi,eax

	mov  eax,edi
	shr  eax,16
	call set_bank
	and  edi,65535
	add  edi,0xa0000
	mov  ecx,[edi]
	and  ecx,255*256*256+255*256+255
	sti

	ret


Vesa12_getpixel32:

	mov  edi,eax ; x
	mov  eax,ebx ; y
	shl  edi,2
	mov  ebx,[0xfe08]
	xor  edx,edx
	mul  ebx
	add  edi,eax

	mov  eax,edi
	shr  eax,16
	call set_bank
	and  edi,65535
	add  edi,0xa0000

	mov  ecx,[edi]
	and  ecx,255*256*256+255*256+255
	sti

	ret
%endif ; CFG_VIDEO_VESA12






; DRAWLINE


draw_line:

    push eax
    push ebx

    push esi
    push edi
   
    push eax
    push ebx
    push ecx
    push edx
   
    mov  cx,[esp+12]
    mov  dx,[esp+14]
    cmp  cx,dx
    jge  dll1

    mov  [esp+12],dx
    mov  [esp+14],cx
    mov  cx,[esp+8]
    mov  dx,[esp+10]
    mov  [esp+8],dx
    mov  [esp+10],cx

 dll1:
   
    mov   eax,[esp+12]
    mov   ebx,[esp+8]
   
    movzx edi,ax
    shr   eax,16
    sub   edi,eax
    movzx esi,bx
    shr   ebx,16
    sub   esi,ebx

    cmp  esi,0
    jnz  dll2
    cmp  edi,0
    jnz  dll2
    mov  ecx,1
    jmp  dlcont

  dll2:

    cmp  esi,edi
    jg   b_esi

    mov  ecx,edi
    xor  edx,edx
    mov  eax,0x10000
    mov  ebx,edi
    div  ebx
    xor  edx,edx
    mov  ebx,eax
    mov  eax,esi
    mul  ebx
    mov  esi,0x10000
    mov  edi,eax
   
    jmp  dlcont

  b_esi:
   
    mov  ecx,esi
    xor  edx,edx
    mov  eax,0x10000
    mov  ebx,esi
    div  ebx
    xor  edx,edx
    mov  ebx,eax
    mov  eax,edi
    mul  ebx
    mov  edi,0x10000
    mov  esi,eax
   
  dlcont:
   
    mov  eax,[esp+12]
    mov  ebx,[esp+8]
   
   
  dll0:
   
    pusha
    shr  eax,16
    shr  ebx,16
    mov  ecx,[esp+32+4]
    mov  edi,[esp+32+16]
    call putpixel
    popa
   
    add  eax,esi
    add  ebx,edi
   
    loop dll0
   
    pop  edx
    pop  ecx
    add  esp,8   
    pop  edi
    pop  esi
    pop  ebx
    pop  eax

    ret





; *********************************
; ********* DRAWBAR RELATED *******
; *********************************


; eax x beginning
; ebx y beginning
; ecx x end
; edx y end
; edi color


drawbar:
%ifdef CFGSYS_VIDEO_VESA_12AND20

    cmp  [0xfe0c],word 0100000000000000b
    jge  dbv20
    cmp  [0xfe0c],word 0x13
    je   dbv20

%endif

%ifdef CFG_VIDEO_VESA12
    call vesa12_drawbar
    ret
%endif

%ifdef CFGSYS_VIDEO_VESA_12AND20

  dbv20:

%endif

%ifdef CFG_VIDEO_VESA20
    call vesa20_drawbar
    ret
%endif

%ifdef CFG_VIDEO_VESA20
vesa20_drawbar:


     mov [novesachecksum],dword 0

     sub edx,ebx ; edx:=Yend-Ystart=H
     sub ecx,eax ; ecx:=Xend-Xstat=B

     push ebp ; +24
     push esi ; +20
     push edi ; +16
     push eax ; +12
     push ebx ; +8
     push ecx ; +4
     push edx ; +0


     mov ecx,[0x3010] ;
     add eax,[ecx-0x1000] ; eax:=Xwin+x
     add ebx,[ecx-0x1000+4] ; ebx:=Ywin+y

     mov ecx, eax ; ecx:=(x+Xwin)
     mov eax, [0xfe08] ; BytesPerScanLine
     mul ebx ; *(y+Ywin)
     mov edi, eax ; edi:=BytesPerScanLine*(y+Ywin)

     mov eax, [0xfe00] ; ScreenXSize
     inc eax ; +1
     mul ebx ; *(y+Ywin)
     mov ebp, eax ; 
     add ebp, ecx ; +(x+Win)
     add ebp, 0x400000 ; ebp:=(y+Ywin)*(ScreenXSize+1)+(x+Xwin)+AddrBuffer

     add edi,ecx ; +(x+Xwin)
     add edi,ecx ; +(x+Xwin)
     add edi,ecx ; +(x+Xwin)

     cmp [0xfbf1],byte 24 ; 24 or 32 bpp ? - x start
     jz dbpi24
     add edi,ecx ; +(x+Xwin)
   dbpi24:
     add edi,[0xfe80]     ; +AddrLFB

; edi:=(y+Ywin)*Screen_Y_multiplier+X*BytesPerPixel
; ebp:=(y+Ywin)*(ScreenXSize+1)+(x+Xwin)+AddrBuffer

; x size

     mov eax,[esp+4] ; B [esp+4]

     mov ecx,eax ; B
     add ecx,eax ; +B
     add ecx,eax ; +B

     cmp [0xfbf1],byte 24 ; 24 or 32 bpp ? - x size
     jz dbpi243
     add ecx,eax ; +B if 32bpp
     ; ecx:=B*BytesPerPixel, eax:=B
   dbpi243:

     mov ebx,[esp+0] ; H

     ; check limits ?

     push eax
     push ecx

     mov eax,[0x3010]

     mov ecx,[eax+0xd000-0x3000+0]
     cmp ecx,0
     jnz dbcblimitlset

     mov ecx,[eax+0xd000-0x3000+4]
     cmp ecx,0
     jnz dbcblimitlset

     mov ecx,[eax+0xd000-0x3000+8]
     cmp ecx,[0xfe00]
     jnz dbcblimitlset

     mov ecx,[eax+0xd000-0x3000+12]
     cmp ecx,[0xfe04]
     jnz dbcblimitlset

     pop ecx
     pop eax

     push dword 0

     jmp dbcblimitlno
 
   dbcblimitlset:

     pop ecx ; B*BytesPerPixel
     pop eax ; B

     push dword 1

   dbcblimitlno:


     cmp [0xfbf1],byte 24 ; 24 or 32 bpp ?
     jz dbpi24bit
     jmp dbpi32bit




; DRAWBAR 24 BBP


dbpi24bit:

     mov ecx,eax ; B
     cld

   dbnewpi:

     push ebx
     push edi
     push ecx ; B

     mov  ebx,[0x3010]
     movzx ebx,byte[ebx+0xe]

     cld

     dbnp24:

	 mov dl,[ds:ebp]

	 push eax
	 push ecx

	 cmp dl,bl
	 jnz dbimp24no

	 mov eax,[esp+8+3*4+16+4]    ; NoCurrentWin|NoCurrentWin

	 mov ecx,[esp+5*4]
	 cmp ecx,0
	 jz dbimp24yes

	 call voodoodbcplimit
	 jnz  dbimp24no

       dbimp24yes:

	 mov  [edi],ax
	 shr  eax,16    
	 mov  [edi+2],al

       dbimp24no:

	 pop ecx
	 pop eax

	 add edi,3                   ; ptrScreen+=3
	 inc ebp

	 dec  ecx
	 jnz  dbnp24                 ; while ecx > 0

       ;

       mov eax,[esp+3*4+16+4]        ; color glide
       test eax,0x80000000
       jz nodbgl24
       cmp al,0
       jz nodbgl24
       dec eax
       mov [esp+3*4+16+4],eax

    nodbgl24:

     pop ecx ; ecx:=B
     pop edi
     pop ebx

     add edi,[0xfe08] ; ptrScreen+=BytesPerScanLine

     add ebp,[0xfe00] ;
     sub ebp, ecx ;
     inc ebp ; ptrBuffer:=ptrBuffer-B+BytesPerPixel+1

     dec ebx ; H--
     jz dbnonewpi ; H <> 0
     jmp dbnewpi

dbnonewpi:

     add esp,7*4
     pop ebp

     ret


; DRAWBAR 32 BBP


dbpi32bit:

     cld
     shr ecx,2

   dbnewpi32:

     push ebx
     push edi
     push ecx

     mov   ebx,[0x3010]
     movzx ebx,byte[ebx+0xe]

     cld

     dbnp32:

       mov dl,[ds:ebp]

       push eax
       push ecx

       cmp dl,bl
       jnz dbimp32no

       mov eax,[esp+8+3*4+16+4]
 
       mov ecx,[esp+5*4]
       cmp ecx,0
       jz dbimp32yes

       call voodoodbcplimit
       jnz  dbimp32no

     dbimp32yes:

       mov  [edi],ax
       shr  eax,16    
       mov  [edi+2],al

     dbimp32no:

       pop ecx
       pop eax

       add edi,4                ; ptrScreen+=4
       inc ebp

       dec ecx
       jnz dbnp32

   dbnp32d:

     mov eax,[esp+12+16+4]      ; color glide
     test eax,0x80000000
     jz nodbgl32
     cmp al,0
     jz nodbgl32
     dec eax
     mov [esp+12+16+4],eax
   nodbgl32:

     pop ecx
     pop edi
     pop ebx

     add edi,[0xfe08]            ; ptrScreen+=BytesPerScanLine

     add ebp,[0xfe00]            ;
     sub ebp, ecx                ;
     inc ebp                     ; ptrBuffer:=ptrBuffer-B+BytesPerPixel+1

     dec ebx                     ; H--
     jz nodbnewpi32              ; H<>0

     jmp dbnewpi32

   nodbnewpi32:

     add esp,7*4
     pop ebp

     ret


voodoodbcplimit:

; ebp:=(y+Ywin)*(ScreenXSize+1)+(x+Xwin)+AddrBuffer


     pusha

     xor edx,edx
     mov eax,ebp
     mov ebx,[0xfe00] ; Screen_X_size
     add ebx,1 ; +1
     sub eax,0x400000 ; -AddrBuffer
     div ebx ;
     mov ebx,eax ; ebx:=Y
     mov eax,edx ; eax:=X
     call cplimit

     cmp  ecx,0
     jne  dbcpl12
     popa
     clc
     ret
   dbcpl12:
     popa
     stc
     ret
%endif ; CFG_VIDEO_VESA20


%ifdef CFG_VIDEO_VESA12
dbcplimit:

	pusha

	xor  edx,edx
	mov  ebx,[0xfe00]
	add  ebx,1
	sub  eax,0x400000
	div  ebx
	mov  ebx,eax
	mov  eax,edx
	call cplimit

	cmp  ecx,0
	jne  dbcpl1
	popa
	clc
	ret
     dbcpl1:
	popa
	stc
	ret




vesa12_drawbar:


    mov  [novesachecksum],dword 0

    sub  edx,ebx
    sub  ecx,eax

    push esi
    push edi
    push eax
    push ebx
    push ecx
    push edx


    mov  ecx,[0x3010]
    add  eax,[ecx-0x1000]
    add  ebx,[ecx-0x1000+4]

    push eax
    mov  eax,ebx         ; y
    mov  ebx,[0xfe08]
    mul  ebx
    pop  ecx

    add  eax,ecx         ; x
    add  eax,ecx
    add  eax,ecx

    cmp  [0xfbf1],byte 24     ; 24 or 32 bpp ? - x start
    jz   dbpi2412
    add  eax,ecx
  dbpi2412:
    add  eax,[0xfe80]
    mov  edi,eax


    ; x size

    mov  eax,[esp+4] ; [esp+6]

    mov  ecx,eax
    add  ecx,eax
    add  ecx,eax

    cmp  [0xfbf1],byte 24     ; 24 or 32 bpp ? - x size
    jz   dbpi24312
    add  ecx,eax
  dbpi24312:

    mov  ebx,[esp+0]  

    ; check limits ?

    push eax
    push ecx

    mov  eax,[0x3010]
	
    mov  ecx,[eax+0xd000-0x3000+0]
    cmp  ecx,0
    jnz  dbcblimitlset12

    mov  ecx,[eax+0xd000-0x3000+4]
    cmp  ecx,0
    jnz  dbcblimitlset12

    mov  ecx,[eax+0xd000-0x3000+8]
    cmp  ecx,[0xfe00]
    jnz  dbcblimitlset12

    mov  ecx,[eax+0xd000-0x3000+12]
    cmp  ecx,[0xfe04]
    jnz  dbcblimitlset12

    pop  ecx
    pop  eax

    push dword 0

    jmp  dbcblimitlno12

  dbcblimitlset12:

    pop  ecx
    pop  eax

    push dword 1
    
  dbcblimitlno12:


    cmp  [0xfbf1],byte 24     ; 24 or 32 bpp ?
    jz   dbpi24bit12
    jmp  dbpi32bit12



; DRAWBAR 24 BBP


dbpi24bit12:

    push eax
    push ebx
    push edx

    mov  eax,ecx
    mov  ebx,3
    div  ebx
    mov  ecx,eax

    pop  edx
    pop  ebx
    pop  eax

    cld

  dbnewpi12:

    push ebx
    push edi
    push ecx

      xor  edx,edx
      mov  eax,edi
      sub  eax,[0xfe80]
      mov  ebx,3
      div  ebx
      add  eax,0x400000

      mov   ebx,[0x3010]
      movzx ebx,byte[ebx+0xe]

      cld

      dbnp2412:

	mov  dl,[eax]

	push eax
	push ecx

	cmp  dl,bl
	jnz  dbimp24no12

	cmp  [esp+5*4],dword 0
	jz   dbimp24yes12

	call dbcplimit
	jnz  dbimp24no12

     dbimp24yes12:

	push edi

	mov  eax,edi
	sub  eax,[0xfe80]
	shr  eax,16
	call set_bank
	and  edi,0xffff
	add  edi,0xa0000
	
	mov  eax,[esp+8+3*4+16+4+4]
	stosw
	shr  eax,16
	stosb

	sti

	pop  edi

	add  edi,3

	pop  ecx
	pop  eax

	inc  eax

	loop dbnp2412,ecx
	jmp  dbnp24d12

      dbimp24no12:

	pop  ecx
	pop  eax
	cld

	add  edi,3
	inc  eax

	loop dbnp2412,ecx

      dbnp24d12:


	mov  eax,[esp+3*4+16+4]
	test eax,0x80000000
	jz   nodbgl2412
	cmp  al,0
	jz   nodbgl2412
	dec  eax
	mov  [esp+3*4+16+4],eax
      nodbgl2412:


    pop  ecx
    pop  edi
    pop  ebx

    add  edi,[0xfe08]

    dec  ebx
    jz   dbnonewpi12
    jmp  dbnewpi12

  dbnonewpi12:
  
    add  esp,7*4

    ret


; DRAWBAR 32 BBP


  dbpi32bit12:

    cld
    shr  ecx,2

   dbnewpi3212:

    push ebx
    push edi
    push ecx

      mov  eax,edi
      sub  eax,[0xfe80]
      shr  eax,2
      add  eax,0x400000

      mov   ebx,[0x3010]
      movzx ebx,byte[ebx+0xe]

      cld

      dbnp3212:

	mov  dl,[eax]

	push eax
	push ecx

	cmp  dl,bl
	jnz  dbimp32no12

	cmp  [esp+5*4],dword 0
	jz   dbimp32yes12

	call dbcplimit
	jnz  dbimp32no12

      dbimp32yes12:

	push edi

	mov  eax,edi
	sub  eax,[0xfe80]
	shr  eax,16
	call set_bank
	and  edi,0xffff
	add  edi,0xa0000
	
	mov  eax,[esp+8+3*4+16+4+4]
	stosw
	shr  eax,16
	stosb

	sti

	pop  edi

	add  edi,4
	inc  ebp

	pop  ecx
	pop  eax

	inc  eax

	loop dbnp3212,ecx
	jmp  dbnp32d12

      dbimp32no12:

	pop  ecx
	pop  eax

	inc  eax

	add  edi,4
	inc  ebp

	loop dbnp3212,ecx

      dbnp32d12:

	mov  eax,[esp+12+16+4]
	test eax,0x80000000
	jz   nodbgl3212
	cmp  al,0
	jz   nodbgl3212
	dec  eax
	mov  [esp+12+16+4],eax
      nodbgl3212:

    pop  ecx
    pop  edi
    pop  ebx

    add  edi,[0xfe08]

    dec  ebx
    jz   nodbnewpi3212

    jmp  dbnewpi3212

  nodbnewpi3212:

    add  esp,7*4

    ret
%endif ; CFG_VIDEO_VESA12




; activate window

windowactivate:


	; esi = abs mem position in stack 0xC400+

	pusha

	push  esi

	xor   eax,eax
	mov   ax,[esi] ; ax <- process no
	shl   eax,1
	mov   ebx,eax
	add   ebx,0xc000
	xor   eax,eax
	mov   ax,[ebx] ; ax <- position in window stack

	mov   esi,0                 ; drop others

      waloop:

	cmp   esi,dword[0x3004]
	jb    waok
	jmp   wacont

      waok:

	inc   esi

	mov   edi,esi
	shl   edi,1
	add   edi,0xc000

	mov   bx,[edi]
	cmp   bx,ax
	jbe   wanoc
	dec   bx
	mov   [edi],bx

      wanoc:

	jmp   waloop

      wacont:
			    ; set to no 1

	pop   esi

	xor   eax,eax
	mov   ax,[esi]
	shl   eax,1
	add   eax,0xc000

	mov   bx,[0x3004]
	mov   [eax],bx

	; update on screen -window stack

	mov   esi,0

      waloop2:

	mov   edi,[0x3004]
	cmp   esi,edi
	jb    waok2
	jmp   wacont2

      waok2:

	inc   esi

	mov   edi,esi
	shl   edi,1
	add   edi,0xc000

	movzx ebx,word [edi]

	shl   ebx,1
	add   ebx,0xc400

	mov   ecx,esi
	mov   [ebx],cx

	jmp   waloop2

      wacont2:

	mov   [0xf400],byte 0           ; empty keyboard buffer
	mov   [0xf500],byte 0           ; empty button buffer

	popa

	ret



; check misc

checkmisc:



    cmp   [0xfff0],byte 0               ; background update ?
    jz    nobackgr
    cmp   [0xfff4],byte 0
    jz    ndmb
    call  drawmouseunder
  ndmb:
    mov   [0xfff0],byte 2
    call  change_task
    mov   [0xfff0],byte 0
    mov   [0xd000+32+0],dword 0
    mov   [0xd000+32+4],dword 0
    mov   eax,[0xfe00]
    mov   ebx,[0xfe04]
    mov   [0xd000+32+8],eax
    mov   [0xd000+32+12],ebx
    call  drawbackground
    mov   [0xfff4],byte 0
  nobackgr:


    

    mov   eax,[0x3004]                  ; zombie / termination ?
    mov   ebx,0x3020+0xa
    mov   esi,1

  newct:
    mov   cl,[ebx]
    cmp   cl,byte 3
    jz    terminate
    cmp   cl,byte 4
    jz    terminate

    add   ebx,0x20
    inc   esi
    dec   eax
    jnz   newct

    jmp   nech


; ****** TERMINATE APPLICATION ******


terminate:


    cli

    push  esi ; process no to terminate
    push  esi
    push  esi

    ; save for draw screen

    pusha

    cld
    shl   esi,5
    add   esi,0x2000

    mov   ax,[esi+0]
    mov   [dlx],ax
    mov   bx,[esi+8]
    add   ax,bx
    mov   [dlxe],ax

    mov   ax,[esi+4]
    mov   [dly],ax
    mov   bx,[esi+12]
    add   ax,bx
    mov   [dlye],ax

    popa


    ; process info away

    sub   ebx,0xa
    inc   eax

    mov   eax,ebx           ; 0x3000
    add   eax,0x20

    mov   ecx,0x3ff0
    sub   ecx,eax
    call  memmove

    sub   ebx,0x1000        ; 0x2000
    sub   eax,0x1000
    call  memmove

    add   ebx,0x1000*0xB    ; 0xD000
    add   eax,0x1000*0xB
    call  memmove

    mov   ebx,[esp]         ; 0x80000
    shl   ebx,8
    add   ebx,0x80000
    mov   eax,ebx
    add   eax,256
    mov   ecx,0x8fff0
    sub   ecx,eax
    call  memmove

    mov   ebx,[esp]         ; tss away
    mov   esi,ebx
    shl   ebx,7
    add   ebx,0x40000
    mov   eax,ebx
    add   eax,128
    mov   ecx,0x4fff0
    sub   ecx,eax
    call  memmove



    ; buttons away

    pop   ecx

  newba2:

    mov   edi,[0xfe88]
    mov   eax,edi

    cld

    movzx ebx,word [edi]
    inc   bx

  newba:

    dec   bx
    jz    nmba

    add   eax,0x10
    cmp   cx,[eax]
    jnz   newba

    pusha
    mov   ecx,ebx
    inc   ecx
    shl   ecx,4
    mov   ebx,eax
    add   eax,0x10
    call  memmove
    dec   dword [edi]
    popa

    jmp   newba2

  nmba:

    ; on buttons: decrease bigger process numbers

    pop   esi

    mov   edi,[0xfe88]

    cld
    movzx eax,word [edi]
    inc   eax

    mov   ebx,[0xfe88]

  newuws3:
    dec   eax
    jz    nnws3

    add   ebx,0x10
    movzx ecx,word [ebx]
    cmp   ecx,esi
    jbe   newuws3

    dec   ecx
    mov   [ebx],cx
    jmp   newuws3

  nnws3:


    ; update windowing stacks


    ; 0xC000 I - move processes

    pop   esi

    mov   ecx,esi    ; save for II
    shl   ecx,1
    add   ecx,0xc000
    xor   edx,edx
    mov   dx,[ecx]
    push  edx

    mov   eax,esi
    shl   eax,1
    add   eax,0xC000

    mov   ebx,eax
    add   eax,2

    mov   ecx,[0x3004]
    sub   ecx,esi
    add   ecx,2
    shl   ecx,1

    call  memmove

    ; 0xC000 II - decrease biggers

    pop   edx

    xor   ecx,ecx
    mov   eax,[0x3004]
    inc   eax
    inc   eax
    mov   ebx,0xc000
  newuws:
    dec   eax
    jz    nnws
    add   ebx,2
    mov   cx,[ebx]
    cmp   ecx,edx
    jbe   newuws
    dec   cx
    mov   [ebx],cx
    jmp   newuws
  nnws:

    ; C000 --> C400

    mov   eax,0xc000
    mov   esi,0

    cld

  lc40:

    add   eax,2

    inc   esi
    mov   ebx,[0x3004]
    cmp   ebx,esi
    jb    lc41

    xor   ecx,ecx
    mov   cx,[eax]
    shl   ecx,1
    add   ecx,0xc400
    mov   edx,esi
    mov   [ecx],dx

    jmp   lc40


  lc41:

    dec   dword [0x3004]            ; decrease number of processes

    mov   [0xf400],byte 0           ; empty keyboard buffer
    mov   [0xf500],byte 0           ; empty button buffer

    sti  ; .. and life goes on


    movzx eax,word [dlx]
    movzx ebx,word [dly]
    movzx ecx,word [dlxe]
    movzx edx,word [dlye]
    call  calculatescreen

    mov   esi,0
    call  redrawscreen

    ret


  nech:


    ; system shutdown request

    cmp  [0xFF00],byte 1
    jnz  noshutdown
    jmp  system_shutdown
  noshutdown:


    ret




; shut down the system

system_shutdown:

%ifdef CFG_CDAUDIO
    mov  eax,3            ; stop playing cd
    call sys_cd_audio
%endif

    cli
    cld

    mov  ecx,0x3fff00/4    ; darken screen
    push ecx
    mov  esi,[0xfe80]
    mov  edi,16*0x100000
    cld
  sdnewpix:
    mov  eax,[esi]
    add  esi,4
    shr  eax,1
    and  eax,0x7f7f7f7f
    stosd
    loop sdnewpix
    pop  ecx
    mov  esi,16*0x100000
    mov  edi,[0xfe80]
    cld
    rep  movsd


    mov  eax,[0xfe00]
    shr  eax,1
    sub  eax,220

    mov  ebx,[0xfe04]
    shr  ebx,1
    mov  [shutdownpos],ebx
    sub  ebx,120

    mov  edi,1
    mov  ecx,0x0000ff

  sdnewpix2:

    call putpixel

    inc  eax
    mov  esi,[0xfe00]
    shr  esi,1
    add  esi,220
    cmp  eax,esi
    jnz  sdnewpix2

    dec  ecx

    mov  eax,[0xfe00]
    shr  eax,1
    sub  eax,220

    inc  ebx

    mov  edx,[shutdownpos]
    add  edx,105
    cmp  ebx,edx
    jnz  sdnewpix2


;    mov  esi,[0xfe00]      ; menuet version
;    shr  esi,1
;    sub  esi,220
;    add  esi,27
;    shl  esi,16
;    mov  eax,esi
;    add  eax,[shutdownpos]
;    sub  eax,105
;    mov  ebx,0xffffff
;    mov  ecx,version
;    mov  edx,16
;    mov  edi,1
;    call dtext


    mov  esi,[0xfe00]      ; 'it is safe..'
    shr  esi,1
    sub  esi,220
    add  esi,27
    shl  esi,16
    mov  eax,esi
    add  eax,[shutdownpos]
    add  eax,43
    mov  esi,5
    mov  ebx,0xffffff      
    mov  ecx,shutdowntext
    mov  edx,40
    mov  edi,1
  newsdt:
    call dtext
    add  eax,10
    add  ecx,40
    sub  esi,1
    cmp  esi,0
    jnz  newsdt


    mov  eax,endsplash_img          ; load endsplash image
    mov  ebx,0
    mov  ecx,16800
    mov  edx,0x90000
    mov  esi,0
    call fileread

    mov  esi,[0xfe00]       ; draw endsplash image
    shr  esi,1
    add  esi,20
    shl  esi,16
    mov  eax,esi
    add  eax,[shutdownpos]
    sub  eax,110

    mov  ebx,0xff0000
    mov  ecx,0x90001
    mov  edx,27
    mov  edi,1

   nrl:
    call dtext
    sub  ebx,0x050000
    add  eax,8
    add  ecx,31
    cmp  ecx,dword 0x90001+25*31
    jnz  nrl

    call checkEgaCga 

    cli

    mov  eax,0x2F0000
    mov  ebx,0x0000
    mov  ecx,0xffff
    call memmove

    call restorefatchain

    mov  eax,pr_mode_exit
    mov  [0x467+0],ax
    mov  [0x467+2],word 0x1000

    mov  al,0x0F
    out  0x70,al
    mov  al,0x05
    out  0x71,al

    mov  al,0xFE
    out  0x64,al
    hlt


BITS 16


pr_mode_exit:


    mov AX,0x1000
    mov DS,AX
    mov ES,AX
    mov FS,AX
    mov GS,AX
    mov SS,AX

    mov al,2   
    out 0x21,al
    mov al,0                 
    out 0xA1,al

    jmp real_mode

old_ints_h
	dw      4*0x20
	dd      0
	dw      0

real_mode:

    lidt [cs:old_ints_h]

    mov  sp,0xfff0

    sti



  nbw:                     ; wait for key up to avoid ps/2 mouse confusion
    xor  ax,ax
    in   al,0x60
    cmp  ax,128
    jb   nbw

    sub  ax,128

%ifdef CFG_FLOPPY_WRITE
    cmp  al,2              ; write floppy
    jnz  no_floppy_write
    call floppy_write
    jmp  nbw2
  no_floppy_write:
%endif

%ifdef CFG_APM_POWEROFF
    cmp  al,3              ; poweroff
    jnz  no_apm_off
    call APM_PowerOff
  no_apm_off:
%endif

    cmp  al,4              ; boot
    jnz  no_sys_boot
    jmp  0xffff:0
  no_sys_boot:

  nbw2:                    ; wait for key down
    xor  ax,ax
    in   al,0x60
    cmp  ax,128
    jge  nbw2

    jmp  nbw


%ifdef CFG_APM_POWEROFF
APM_PowerOff:

	mov     ax,5304h
	sub     bx,bx
	int     15h
	mov     ax,5302h
	sub     bx,bx
	int     15h
	mov     ax,5308h
	mov     bx,1
	mov     cx,bx
	int     15h
	mov     ax,530Dh
	mov     bx,1
	mov     cx,bx
	int     15h
	mov     ax,530Fh
	mov     bx,1
	mov     cx,bx
	int     15h
	mov     ax,530Eh
	sub     bx,bx
	mov     cx,102h
	int     15h
	mov     ax,5307h
	mov     bx,1
	mov     cx,3
	int     15h
	ret
%endif ; CFG_APM_POWEROFF


%ifdef CFG_FLOPPY_WRITE
; write diskette image to physical floppy


floppy_write:


	pusha

	mov     ax,0x0000               ; reset drive
	mov     dx,0x0000
	int     0x13

	mov     cx,0x0001               ; startcyl,startsector
	mov     dx,0x0000               ; starthead,drive
	push    word 80*2               ; read no of sect

      fwwrites:
	pusha

	; move 1mb+ -> 0:a000

	pusha
	mov     si,fwmovedesc
	push    word 0x1000
	pop     es
	mov     cx,256*18
	mov     ah,0x87
	int     0x15
	mov     eax,[es:fwmovedesc+0x12]
	add     eax,512*18
	mov     [es:fwmovedesc+0x12],eax
	popa

	xor     si,si
      fwnewwrite:
	push    word 0x0
	pop     es
	mov     bx,0xa000               ; es:bx -> data area
	mov     ax,0x0300+18            ; read, no of sectors to read
	int     0x13

	cmp     ah,0
	jz      fwgoodwrite

	add     si,1
	cmp     si,10
	jnz     fwnewwrite

	add     esp,32+2

	popa                            ; can't access diskette

	ret

      fwgoodwrite:

	popa

	cld

	inc     dh
	cmp     dh,2
	jnz     fwbb2
	mov     dh,0
	inc     ch

      fwbb2:

	cld
	pop     ax
	dec     ax
	push    ax
	cmp     ax,0
	jnz     fwrs

	pop     ax

	jmp     fwwritedone
      fwrs:
	jmp     fwwrites

      fwmovedesc

	db      0x00,0x00,0x0,0x00,0x00,0x00,0x0,0x0
	db      0x00,0x00,0x0,0x00,0x00,0x00,0x0,0x0
	db      0xff,0xff,0x0,0x00,0x10,0x93,0x0,0x0
	db      0xff,0xff,0x0,0xa0,0x00,0x93,0x0,0x0
	db      0x00,0x00,0x0,0x00,0x00,0x00,0x0,0x0
	db      0x00,0x00,0x0,0x00,0x00,0x00,0x0,0x0
	db      0x00,0x00,0x0,0x00,0x00,0x00,0x0,0x0
	db      0x00,0x00,0x0,0x00,0x00,0x00,0x0,0x0

      fwwritedone:

	popa

	ret
%endif ; CFG_FLOPPY_WRITE


BITS 32


shutdownpos dd 0x0

shutdowntext
    db   "IT'S SAFE TO POWER OFF COMPUTER OR      "
    db   '                                        '
    db   '1) SAVE RAMDISK TO FLOPPY               '
    db   '2) APM - POWEROFF                       '
    db   '3) REBOOT                               '

endsplash_img
    db CFG_ENDSPLASH_IMG




; draw a window

drawwindow_I:

	pusha

	mov   ecx,[edx+24]
	mov   eax,[edx+00]
	shl   eax,16
	add   eax,[edx+00]
	add   eax,[edx+08]
	add   eax,1
	mov   ebx,[edx+04]
	shl   ebx,16
	add   ebx,[edx+04]
	mov   edi,0
	call  draw_line

	mov   ecx,[edx+24]
	mov   eax,[edx+00]
	shl   eax,16
	add   eax,[edx+00]
	add   eax,[edx+08]
	add   eax,1
	mov   ebx,[edx+04]
	add   ebx,[edx+12]
	shl   ebx,16
	add   ebx,[edx+04]
	add   ebx,[edx+12]
	mov   edi,0
	call  draw_line

	mov   ecx,[edx+24]
	mov   eax,[edx+00]
	shl   eax,16
	add   eax,[edx+00]
	mov   ebx,[edx+04]
	shl   ebx,16
	add   ebx,[edx+04]
	add   ebx,[edx+12]
	add   ebx,1
	mov   edi,0
	call  draw_line

	mov   ecx,[edx+24]
	mov   eax,[edx+00]
	add   eax,[edx+08]
	shl   eax,16
	add   eax,[edx+00]
	add   eax,[edx+08]
	mov   ebx,[edx+04]
	shl   ebx,16
	add   ebx,[edx+04]
	add   ebx,[edx+12]
	add   ebx,1
	mov   edi,0
	call  draw_line

	mov   ecx,[edx+20]   ; grab bar
	push  ecx

	mov   esi,edx
	mov   edx,[esi+04]   
	add   edx,1
	mov   ebx,[esi+04]
	add   ebx,21

	mov   eax,[esi+04]
	add   eax,[esi+12]

	cmp   ebx,eax
	jb    wdsizeok
	mov   ebx,eax
      wdsizeok:

	push  ebx
      drwi:
	mov   eax,edx
	shl   eax,16
	add   eax,edx
	mov   ebx,eax
	mov   eax,[esi+00]
	inc   eax
	shl   eax,16
	add   eax,[esi+00]
	add   eax,[esi+08]

	push  edx            ; color glide ?
	cld
	clc
	mov   edx,0x80000000
	mov   ecx,[esi+20]
	and   ecx,edx
	cmp   ecx,edx
	jnz   nofa

	mov   ecx,[esi+20]
	sub   ecx,0x00040404
	mov   [esi+20],ecx
	and   ecx,0x00ffffff
	jmp   faj

      nofa:
	mov   ecx,[esi+20]
	and   ecx,0x00ffffff

      faj:

	pop   edx

	mov   edi,0
	call  draw_line
	inc   edx
	pop   ebx
	push  ebx
	cmp   edx,ebx
	jnz   drwi
	pop   ebx

	pop   ecx
	mov   [esi+20],ecx


	mov   edx,[esi+04]      ; inside 
	add   edx,21
	mov   ebx,[esi+04]
	add   ebx,[esi+12]
	cmp   edx,ebx
	jbe   inside
	jmp   noinside
      inside:
	mov   eax,1
	mov   ebx,21
	mov   ecx,[esi+8]
	mov   edx,[esi+12]
	mov   edi,[esi+16]
	call  drawbar
      noinside:


;        mov   eax,[esi+08] ; resize knob
;        sub   eax,1
;        mov   ebx,[esi+12]
;        sub   ebx,12
;        mov   ecx,[esi+08]
;        mov   edx,[esi+12]
;        mov   edi,[esi+24]
;      rsk1:
;        pusha  
;        call  drawbar
;        popa
;        sub   eax,1
;        add   ebx,1
;        cmp   ebx,[esi+12]
;        jnz   rsk1
;      noresizebar:


	popa

	ret





; negative button image

negativebutton:

	pusha

	movzx ebx,word [eax+0]
	shl   ebx,5
	add   ebx,0x2000
	movzx ecx,word [ebx+0]     ; window x start
	movzx edx,word [eax+4]     ; button x start
	add   ecx,edx
	push  ecx

	mov   dx,[eax+6]     ; button x size
	add   cx,dx
	mov   esi,ecx
	add   esi,1
	mov   cx,[ebx+4]     ; window y start
	mov   dx,[eax+8]     ; button y start
	add   ecx,edx
	mov   ebx,ecx
	mov   dx,[eax+10]    ; button y size
	add   dx,cx
	add   dx,1

	pop   eax
	mov   ecx,0x01000000

	; eax x beginning
	; ebx y beginning
	; esi x end
	; edx y end
	; ecx color

	sub   edx,1
	push  edx
	add   edx,1
	sub   esi,1
	push  esi
	add   esi,1

	push  eax
	push  ebx
	push  ecx
	push  edx
	push  edi

      bdbnewline:
	mov   edi,1    ; force
	cmp   eax,[esp+16]
	jz    bneg
	cmp   eax,[esp+20]
	jz    bneg
	cmp   ebx,[esp+12]
	jz    bneg
	cmp   ebx,[esp+24]
	jz    bneg
	jmp   nbneg

      bneg:

	call  putpixel

      nbneg:

	add   eax,1
	cmp   eax,esi
	jnz   bdbnewline
	mov   eax,[esp+16]
	add   ebx,1
	cmp   ebx,edx
	jnz   bdbnewline

	add   esp,28

	popa

	ret



; check buttons


; 0000 word process number
; 0002 word button id number
; 0004 word x start
; 0006 word x size
; 0008 word y start
; 000A word y size
;
; button table in 0x10 increments
;
; first at 0x10



checkbuttons:

    pusha

    mov   bl,[0xfb40]   ; mouse buttons pressed ?
    mov   al,0
    cmp   bl,al
    jnz   cbc

    popa
    ret

  cbc:

    mov   esi,0

    xor   edx,edx
    mov   edi,[0xfe88]
    mov   dx,[edi]

    mov   ecx,0
    cmp   edx,ecx
    jz    nobuttoncheck

    push  esi

    inc   edx
    push  edx

    jmp   buttonnewcheck


  nobuttoncheck:

    popa
    ret


  buttonnewcheck:

    pop   edx
    pop   esi

    inc   esi

    cmp   edx,esi
    jnz   bch

    popa                 ; no button pressed

    ret

  bch:

    push  esi
    push  edx

    mov   eax,esi

    shl   eax,4
    add   eax,edi

    xor   ecx,ecx
    xor   ebx,ebx

    mov   bx,[eax+0]     ; top of windowing stack ?
    shl   ebx,1
    add   ebx,0xc000
    mov   cx,[ebx]
    mov   dx,[0x3004]
    cmp   ecx,edx
    jnz   buttonnewcheck

    ; first in windowing stack -> check coordinates

    xor   ebx,ebx        ; mouse x >= button x ?
    mov   bx,[eax+0]
    shl   ebx,5
    add   ebx,0x2000
    xor   ecx,ecx
    mov   cx,[ebx+0]     ; window x start
    xor   edx,edx
    mov   dx,[eax+4]     ; button x start
    add   edx,ecx
    mov   cx,[0xfb0a]
    cmp   edx,ecx
    jbe   buc1
    jmp   buttonnewcheck
   buc1:

    xor   ebx,ebx
    mov   bx,[eax+6]     ; button x size
    add   edx,ebx
    cmp   ecx,edx
    jbe   buc11
    jmp   buttonnewcheck
   buc11:


    xor   ebx,ebx        ; mouse y >= button y ?
    mov   bx,[eax+0]
    shl   ebx,5
    add   ebx,0x2000
    xor   ecx,ecx
    mov   cx,[ebx+4]     ; window y start
    xor   edx,edx
    mov   dx,[eax+8]     ; button y start
    add   edx,ecx
    mov   cx,[0xfb0c]
    cmp   edx,ecx
    jbe   buc2
    jmp   buttonnewcheck
   buc2:


    xor   ebx,ebx
    mov   bx,[eax+10]    ; button y size
    add   edx,ebx
    cmp   ecx,edx
    jbe   buc21
    jmp   buttonnewcheck
   buc21:


    ; mouse on button


    pop   edx
    pop   esi

    mov   bx,[eax+2]       ; save id number
    push  bx

    pusha
    call  drawmouseunder
    mov   [0xfff4],byte 0  ; no mouse background
    mov   [0xfff5],byte 1  ; do not draw mouse
    call  checkmouse
    popa

    call negativebutton

    pusha

  cbwaitmouseup:

    call  checkidle

    call  checkmouse

    cmp   [0xfb40],byte 0  ; mouse buttons pressed ?
    jnz   cbwaitmouseup

    popa

    call  negativebutton

    mov   [0xfff4],byte 0  ; no mouse background
    mov   [0xfff5],byte 0  ; draw mouse

    mov   [0xf500],byte 1
    pop   bx
    mov   [0xf501],bl

    popa

    ret






; check if window is necessary to draw

checkwindowdraw:


	; edi = position in 0x2000+

	mov   esi,edi
	sub   esi,0x2000
	shr   esi,5

	; esi = process nro

	mov   edx,esi
	shl   edx,1
	add   edx,0xc000

	xor   ecx,ecx
	mov   cx,[edx]     
	shl   ecx,1
	add   ecx,0xc400   ; position in windowing stack, checks from here ->

	mov   esi,ecx

	pusha

	xor   eax,eax
	mov   ax,[esi]
	shl   eax,5
	add   eax,0x2000
	mov   esi,eax

	popa


	push  esi

      wdn0:

	pop   esi
	add   esi,2
	push  esi

	mov   eax,[0x3004]
	shl   eax,1
	add   eax,0xc400

	cmp   esi,eax
	jbe   wdn1

	pop   esi

	mov   ecx,0       ; passed all windows to top
	ret

      wdn1:

	xor   eax,eax
	mov   ax,[esi]
	shl   eax,5
	add   eax,0x2000
	mov   esi,eax


	mov   eax,[edi+0]
	mov   ebx,[edi+4]
	mov   ecx,[edi+8]
	mov   edx,[edi+12]
	add   ecx,eax
	add   edx,ebx

	mov   ecx,[esi+4]    ; y check
	cmp   ecx,edx
	jb    wici1
	jmp   wdn0
      wici1:
	mov   eax,[esi+12]
	add   ecx,eax
	cmp   ebx,ecx
	jbe   wici2
	jmp   wdn0
      wici2:

	mov   eax,[edi+0]
	mov   ebx,[edi+4]
	mov   ecx,[edi+8]
	mov   edx,[edi+12]
	add   ecx,eax
	add   edx,ebx

	mov   edx,[esi+0]    ; x check
	cmp   edx,ecx
	jb    wici3
	jmp   wdn0
      wici3:
	mov   ecx,[esi+8]
	add   edx,ecx
	cmp   eax,edx
	jbe   wici4
	jmp   wdn0
      wici4:

	pop   esi

	mov   ecx,1   ; overlap some window

	ret




; if redraw necessary at activate

waredraw:

	pusha

	call  checkwindowdraw      ; draw window on activation ?

	cmp   ecx,0
	jnz   wand2                ; yes
	jmp   wand                 ; no

      wand2:

	popa

	call  windowactivate

	; update screen info

	pusha
	mov   edi,[0x3004]
	shl   edi,1
	add   edi,0xc400
	mov   esi,[edi]
	and   esi,65535
	shl   esi,5
	add   esi,0x2000

	mov   eax,[esi+00]
	mov   ebx,[esi+04]
	mov   ecx,[esi+08]
	mov   edx,[esi+12]

	add   ecx,eax
	add   edx,ebx

	mov   edi,[0x3004]
	shl   edi,1
	add   edi,0xc400
	mov   esi,[edi]
	and   esi,255
	shl   esi,5
	add   esi,0x3000+0xe
	movzx esi,byte[esi]
	call  calculatescreen ; setscreen

	popa

	cmp   [0xff01],dword 1
	jbe   wand5

	mov   eax,10            ; wait for putimages to finish
	call  delay_hs

	mov   [edi+31],byte 1  ; redraw flag for app

	ret

      wand5:

	pusha
	call  drawmouseunder
	mov   [0xfff4],byte 0  ; no mouse background
	mov   [0xfff5],byte 1  ; do not draw mouse
	call  checkmouse
	mov   [0xfff4],byte 0  ; no mouse background
	mov   [0xfff5],byte 1  ; do not draw mouse
	popa

	mov   eax,5            ; wait for putimages to finish
	call  delay_hs

	mov   [0xfff4],byte 0  ; no mouse background
	mov   [0xfff5],byte 1  ; do not draw mouse

	mov   [edi+31],byte 1  ; redraw flag for app

	mov   ecx,100

       cwwaitflagdown:

	dec   ecx
	jz    cwnowait

	mov   [0xfff4],byte 0  ; no mouse background
	mov   [0xfff5],byte 1  ; do not draw mouse

	mov   eax,2
	call  delay_hs
	cmp   [edi+31],byte 0  ; wait flag to drop
	jnz   cwwaitflagdown

      cwnowait:

	mov   eax,10           ; wait for draw to finish
	call  delay_hs

	mov   [0xfff4],byte 0  ; no mouse background
	mov   [0xfff5],byte 0  ; draw mouse

	ret

      wand:

	popa

	call  windowactivate

	ret



; check window touch

checkwindows:


	pusha

	cmp  [0xff01],dword 1  ; activate request from app ?
	jbe  cwl1

	mov  edi,[0xff01]
	shl  edi,5
	add  edi,0x2000

	mov  ebx,[0xff01]
	shl  ebx,1
	add  ebx,0xc000
	mov  esi,[ebx]
	and  esi,65535

	shl  esi,1
	add  esi,0xc400

	call waredraw

	mov  [0xff01],dword 0  

	popa

	ret




      cwl1:

	mov   bl,[0xfb40]
	cmp   bl,0
	jnz   cwm

	popa
	ret

      cwm:

	mov   esi,dword[0x3004]
	inc   esi

      cwloop:

	mov   edx,dword 2
	cmp   edx,esi
	jb    cwok

	popa
	ret

      cwok:
	dec   esi

	mov   eax,esi
	shl   eax,1
	add   eax,0xc400

	xor   ebx,ebx
	mov   bx,[eax]
	shl   ebx,5
	add   ebx,0x2000
	mov   edi,ebx

	mov   ax,[0xfb0a]
	mov   bx,[0xfb0c]

	mov   cx,[edi+0]
	mov   dx,[edi+4]

	cmp   cx,ax
	jb    cw1
	jmp   cwloop
      cw1:
	cmp   dx,bx
	jb    cw2
	jmp   cwloop
      cw2:

	add   cx,[edi+8]
	add   dx,[edi+12]

	cmp   ax,cx
	jb    cw3
	jmp   cwloop
      cw3:
	cmp   bx,dx
	jb    cw4
	jmp   cwloop
      cw4:

	pusha

	mov   eax,esi
	mov   bx,[0x3004]

	cmp   ax,bx

	jnz   nomovwin
	jmp   movwin

       nomovwin:

	; eax = position in windowing stack

	; redraw must ?

	shl   esi,1
	add   esi,0xc400

	call  waredraw

	add   esp,32

	popa

	ret


	; ******** MOVE WINDOW **********

      movwin:

	popa

	mov   dx,[edi+4] ; check if touch on bar
	add   dx,21
	cmp   bx,dx
	jb    cw5

	popa

	ret

      cw5:

	xor   ecx,ecx
	xor   edx,edx

	mov   cx,[edi+0]
	mov   dx,[edi+4]

	pusha
	mov   [dlx],ecx      ; save for drawlimits
	mov   [dly],edx
	mov   eax,[edi+8]
	add   ecx,eax
	mov   eax,[edi+12]
	add   edx,eax
	mov   [dlxe],ecx
	mov   [dlye],edx
	popa

	sub   ax,cx
	sub   bx,dx

	pusha
	call  drawmouseunder
	popa

	mov   esi,[0xfb0a]
	mov   [0xf300],esi

	mov   [0xfff5],byte 1  ; do not draw mouse

	call  checkmouse

	pusha           ; wait for putimages to finish
	mov   eax,5
	call  delay_hs

	mov   eax,[edi+0]
	mov   [npx],eax
	mov   eax,[edi+4]
	mov   [npy],eax

	popa


	pusha                  ; save old coordinates

	mov   ax,[edi+00]
	mov   [oldc+00],ax

	mov   ax,[edi+04]
	mov   [oldc+04],ax

	mov   ax,[edi+08]
	mov   [oldc+08],ax
	mov   [npxe],ax

	mov   ax,[edi+12]
	mov   [oldc+12],ax
	mov   [npye],ax

	popa


	call  drawwindowframes


	; move window


      newchm:

	call  checkidle

	call  checkEgaCga 

	mov   [0xfff4],byte 0
	call  checkmouse

	mov   esi,[0xf300]
	cmp   esi,dword [0xfb0a]
	jnz   cwb2
	jmp   cwb

      cwb2:

	mov   cx,[0xfb0a]
	mov   dx,[0xfb0c]
	sub   cx,ax
	sub   dx,bx
	push  ax
	push  bx

	call  drawwindowframes

	mov   ax,[0xfe00]
	mov   bx,[0xfe04]

	mov   [npx],word 0     ; x repos ?
	cmp   ax,cx
	jb    noreposx
	sub   ax,[npxe]
	mov   [npx],ax
	cmp   ax,cx
	jb    noreposx
	mov   [npx],cx
      noreposx:

	mov   [npy],word 0     ; y repos ?
	cmp   bx,dx
	jb    noreposy
	sub   bx,[npye]
	mov   [npy],bx
	cmp   bx,dx
	jb    noreposy
	mov   [npy],dx
      noreposy:

	pop   bx
	pop   ax
	call  drawwindowframes

	mov   esi,[0xfb0a]
	mov   [0xf300],esi

      cwb:

	cmp   [0xfb40],byte 0
	jz    nonewchm
	jmp   newchm

      nonewchm:               ; new position done

	call  drawwindowframes

	mov   eax,[npx]
	mov   [edi+0],eax
	mov   eax,[npy]
	mov   [edi+4],eax
	mov   eax,[npxe]
	mov   [edi+8],eax
	mov   eax,[npye]
	mov   [edi+12],eax

	pusha
	mov   eax,[oldc+00]
	mov   ebx,[oldc+04]
	mov   ecx,[oldc+08]
	mov   edx,[oldc+12]
	add   ecx,eax
	add   edx,ebx
	call  calculatescreen
	popa

	pusha
	mov   eax,[edi+00]
	mov   ebx,[edi+04]
	mov   ecx,[edi+08]
	mov   edx,[edi+12]
	add   ecx,eax
	add   edx,ebx
	mov   edi,[0x3004]
	shl   edi,1
	add   edi,0xc400
	mov   esi,[edi]
	and   esi,255
	shl   esi,5
	add   esi,0x3000+0xe
	movzx esi,byte [esi]
	call  calculatescreen ; setscreen
	popa

	mov   [edi+31],byte 1 ; mark first as redraw


	mov   esi,edi
	call  redrawscreen

	mov   ecx,100         ; wait to avoid mouse residuals
	cld
      waitre2:
	mov   eax,5
	call  delay_hs
	mov   [0xfff5],byte 1 ; no mouse pointer
	mov   [0xfff4],byte 0 ; no mouse under
	call  checkmouse
	cmp   [edi+31],byte 0
	jz    retwm
	loop  waitre2
      retwm:

	mov   [0xfff5],byte 0 ; mouse pointer
	mov   [0xfff4],byte 0 ; no mouse under

	popa

	ret

oldc    dd 0x0,0x0,0x0,0x0

dlx     dd 0x0
dly     dd 0x0
dlxe    dd 0x0
dlye    dd 0x0

npx     dd 0x0
npy     dd 0x0
npxe    dd 0x0
npye    dd 0x0

mpx     dd 0x0
mpy     dd 0x0








; draw negative window frames

drawwindowframes:

	pusha

	mov   eax,[npx]
	shl   eax,16
	add   eax,[npx]
	add   eax,[npxe]
	add   eax,65536*1
	mov   ebx,[npy]
	shl   ebx,16
	add   ebx,[npy]
	mov   ecx,0x01000000
	push  edi
	mov   edi,1
	call  draw_line
	pop   edi

	mov   eax,[npx]
	shl   eax,16
	add   eax,[npx]
	add   eax,[npxe]
	add   eax,65536*1
	mov   ebx,[npy]
	add   ebx,[npye]
	shl   ebx,16
	add   ebx,[npy]
	add   ebx,[npye]
	mov   ecx,0x01000000
	push  edi
	mov   edi,1
	call  draw_line
	pop   edi


	mov   eax,[npx]
	shl   eax,16
	add   eax,[npx]
	mov   ebx,[npy]
	shl   ebx,16
	add   ebx,[npy]
	add   ebx,[npye]
	add   ebx,1
	mov   ecx,0x01000000
	push  edi
	mov   edi,1
	call  draw_line
	pop   edi


	mov   eax,[npx]
	add   eax,[npxe]
	shl   eax,16
	add   eax,[npx]
	add   eax,[npxe]
	mov   ebx,[npy]
	shl   ebx,16
	add   ebx,[npy]
	add   ebx,[npye]
	add   ebx,1
	mov   ecx,0x01000000
	push  edi
	mov   edi,1
	call  draw_line
	pop   edi

	popa

	ret



; redraw screen

redrawscreen:

; esi , if process base is esi, do not set flag

	 pusha

	 mov   eax,5
	 call  delay_hs

	 mov   ecx,0               ; redraw flags for apps
       newdw2:

	 inc   ecx
	 push  ecx

	 mov   eax,ecx
	 shl   eax,5
	 add   eax,0x2000
				   ; check if window in redraw area
	 mov   edi,eax

	 cmp   ecx,1               ; limit for background
	 jz    bgli

	 mov   eax,[edi+0]
	 mov   ebx,[edi+4]
	 mov   ecx,[edi+8]
	 mov   edx,[edi+12]
	 add   ecx,eax
	 add   edx,ebx

	 mov   ecx,[dlye]   ; ecx = area y end     ebx = window y start
	 cmp   ecx,ebx
	 jb    ricino

	 mov   ecx,[dlxe]   ; ecx = area x end     eax = window x start
	 cmp   ecx,eax
	 jb    ricino

	 mov   eax,[edi+0]
	 mov   ebx,[edi+4]
	 mov   ecx,[edi+8]
	 mov   edx,[edi+12]
	 add   ecx,eax
	 add   edx,ebx

	 mov   eax,[dly]    ; eax = area y start     edx = window y end
	 cmp   edx,eax
	 jb    ricino

	 mov   eax,[dlx]    ; eax = area x start     ecx = window x end
	 cmp   ecx,eax
	 jb    ricino

	bgli:

	 cmp   edi,esi
	 jz    ricino

	 mov   eax,edi

	 add   eax,0xd000-0x2000

	 mov   ebx,[dlx]          ; set limits
	 mov   [eax+0],ebx
	 mov   ebx,[dly]
	 mov   [eax+4],ebx
	 mov   ebx,[dlxe]
	 mov   [eax+8],ebx
	 mov   ebx,[dlye]
	 mov   [eax+12],ebx

	 sub   eax,0xd000-0x2000

	 mov   [eax+31],byte 1    ; mark as redraw


       ricino:

	 pop   ecx

	 cmp   cx,word [0x3004]
	 jz    newdw3
	 jmp   newdw2

       newdw3:

	 cmp   esi,1
	 je    newdw8
	 call  drawbackground
       newdw8:

	 popa

	 ret



;   check mouse
;
;
;   FB00  ->   FB0F   mouse memory 00 chunk count - FB0A-B x - FB0C-D y
;   FB10  ->   FB17   mouse color mem
;   FB21              x move
;   FB22              y move
;   FB30              color temp
;   FB28              high bits temp
;   FB4A  ->   FB4D   FB4A-B x-under - FB4C-D y-under
;   FC00  ->   FCFE   com1/ps2 buffer
;   FCFF              com1/ps2 buffer count starting from FC00

mousecount  dd  0x0
mousedata   dd  0x0

checkmouse:

	pusha

	mov    [mousecount],dword 0x2e0000+12*4096
	mov    [mousedata],dword 0x2e0000+12*4096+0x10
%ifdef CFG_MOUSE_COM1
	cmp    [0xF604],byte 2
	jne    nocom1mouse
	mov    [mousecount],dword 0x2e0000+4*4096
	mov    [mousedata],dword 0x2e0000+4*4096+0x10
      nocom1mouse:
%endif
%ifdef CFG_MOUSE_COM2
	cmp    [0xF604],byte 3
	jne    nocom2mouse
	mov    [mousecount],dword 0x2e0000+3*4096
	mov    [mousedata],dword 0x2e0000+3*4096+0x10
      nocom2mouse:
%endif


      uusicheckmouse:

	mov    ebx,[mousecount]                ; anything at buffer for mouse
	cmp    [ebx],byte 0
	jnz    c_byte

	jmp    checkmouseret

      c_byte:

	; first byte of comX or ps2 ?

	cmp    [0xF604],byte 2
%ifdef CFG_MOUSE_COM1 CFG_MOUSE_COM2
	jge    com1mousefirst
%endif
%ifdef CFG_MOUSE_PS2
	jmp    ps2mousefirst
%endif

%ifdef CFG_MOUSE_COM1 CFG_MOUSE_COM2

	; ******************************************
	; *********** COMX mouse driver ************
	; ******************************************

       com1mousefirst:

	mov    edi,[mousedata]
	mov    dl,byte [edi] ; first com1 ?
	and    dl,64
	cmp    dl,64
	jnz    cm2
	mov    bl,0             ; zero mouse block count
	mov    [0xfb00],byte bl
       cm2:
	xor    ebx,ebx

	mov    bl,[0xfb00]
	add    bl,1
	mov    [0xfb00],bl
	mov    eax,0xfb00
	add    eax,ebx
	mov    edi,[mousedata]
	mov    dl,byte [edi]
	mov    [eax],byte dl
	cmp    bl,3             ; three ?
	jz     com1mouse

	jmp    decm

      com1mouse:

	; buttons

	movzx  eax,byte [0xfb01]
	shr    eax,4
	and    eax,3
	mov    [0xfb40],al


	; com1 mouse
	; x

	mov    dl,[0xfb01]        ; x high bits
	movzx  eax,dl
	and    al,3
	shl    al,6
	mov    dl,byte[0xfb02]    ; x low bits
	add    al,dl
	mov    [0xfb21],byte al
	movzx  ebx,word[0xfb0a]

	mov    al,byte [0xfb01]   ; + or - ?
	and    al,2
	cmp    al,2
	jnz    x_add

       x_sub:
	movzx  ebx,word[0xfb0a]   ; x-
	movzx  eax,byte [0xfb21]
	sub    bx,255
	add    bx,ax
	push   ebx
	mov    [0xfb00],byte 0
	jmp    my_event
       x_add:
	movzx  ebx,word[0xfb0a]   ; x+
	movzx  eax,byte [0xfb21]
	add    bx,ax
	push   ebx
	mov    [0xfb00],byte 0


	; y


      my_event:

	mov    dl,[0xfb01]       ; y high bits
	movzx  eax,dl
	and    al,12
	shl    al,4
	mov    dl,byte[0xfb03]   ; y low bits
	add    al,dl
	mov    [0xfb22],byte al
	movzx  ebx,word[0xfb0c]

	mov    al,byte [0xfb01]  ; + or - ?
	and    al,8
	cmp    al,8
	jnz    y_add

      y_sub:
	movzx  ebx,word[0xfb0c]  ; y-
	movzx  eax,byte [0xfb22]
	sub    bx,255
	add    bx,ax
	push   ebx
	mov    [0xfb00],byte 0
	jmp    mdraw
      y_add:
	movzx  ebx,word[0xfb0c]  ; y+
	movzx  eax,byte [0xfb22]
	add    bx,ax
	push   ebx
	mov    [0xfb00],byte 0
	jmp    mdraw

	; end of com1 mouse

%endif ; CFG_MOUSE_COM1 or CFG_MOUSE_COM2

%ifdef CFG_MOUSE_PS2
	; ******************************************
	; ********  PS2 MOUSE DRIVER  **************
	; ******************************************

      ps2mousefirst:

	movzx  edx,byte [0x2E0000+4096*12+0x10]   ; first ps2 ?
	cmp    edx,40
	jne    cm3
	mov    [0xfb00],byte 0  ; zero mouse block count
      cm3:

	movzx  ebx,byte [0xfb00]
	add    ebx,1
	mov    [0xfb00],bl
	mov    eax,0xfb00
	add    eax,ebx
	mov    dl,byte [0x2E0000+4096*12+0x10]
	mov    [eax],byte dl

	cmp    bl,3             ; full packet of three bytes ?
	jz     ps2mouse
	jmp    decm


      ps2mouse:

	mov    [0xfb00],byte 0  ; zero mouse block count

	; buttons

	movzx  eax,byte [0xfb01]
	and    eax,3
	mov    [0xfb40],al

	; x

	movzx  eax,word [0xfb0a]
	movzx  edx,byte [0xfb02]
	cmp    edx,128
	jb     ps2xp
	shl    edx,1
	add    eax,edx
	cmp    eax,512
	jge    ps2xsok
	mov    eax,0
	jmp    ps2xready
       ps2xsok:
	sub    eax,512
	jmp    ps2xready
       ps2xp:
	shl    edx,1
	add    eax,edx
	jmp    ps2xready
       ps2xready:
	push   eax

	; y

	movzx  eax,word [0xfb0c]
	movzx  edx,byte [0xfb03]
	cmp    edx,128
	jb     ps2yp
	add    eax,512
	shl    edx,1
	sub    eax,edx
	jmp    ps2yready
       ps2yp:
	shl    edx,1
	cmp    edx,eax
	jb     ps201
	mov    edx,eax
       ps201:
	sub    eax,edx
	jmp    ps2yready
       ps2yready:
	push   eax

	jmp    mdraw


	; end of ps2 mouse
%endif ; CFG_MOUSE_PS2

	; FIXME: no mouse needs testing: probably have to push some values etc.


	; ****************************
	; ***** CHECK FOR LIMITS *****
	; ****************************

      mdraw:

	pop    ebx
	pop    eax

	mov    dx,0                   ; smaller than zero
	cmp    bx,dx
	jge    mnb11
	mov    bx,0
      mnb11:
	mov    [0xfb0c],word bx

	mov    dx,0
	cmp    ax,dx
	jge    mnb22
	mov    ax,0
      mnb22:
	mov    [0xfb0a],word ax

	mov    edx,[0xfe04]           ; bigger than maximum
	cmp    ebx,edx
	jb     mnb1
	mov    bx,[0xfe04]
      mnb1:
	mov    [0xfb0c],word bx

	mov    edx,[0xfe00]
	cmp    eax,edx
	jb     mnb2
	mov    ax,[0xfe00]
      mnb2:
	mov    [0xfb0a],word ax

	push   eax
	push   ebx



	; ****************************
	; ****  DRAW POINTER  ********
	; ****************************

	mov    al,[0xfff4]
	cmp    al,0
	jz     nmeka
	call   drawmouseunder
      nmeka:

	pop    ebx
	pop    eax
	call   meka
	jmp    decm


	; return old picture

      drawmouseunder:

	pusha

	xor    ecx,ecx
	xor    edx,edx

      mres:

	movzx  eax,word [0xfb4a]
	movzx  ebx,word [0xfb4c]

	add    eax,ecx
	add    ebx,edx

	push   ecx
	push   edx
	push   eax
	push   ebx

	mov    eax,edx
	shl    eax,6
	shl    ecx,2
	add    eax,ecx
	add    eax,mouseunder
	mov    ecx,[eax]

	pop    ebx
	pop    eax

	push   edi
	mov    edi,1 ;force
	call   putpixel
	pop    edi

	pop    edx
	pop    ecx

	add    ecx,1
	cmp    ecx,16
	jnz    mres
	xor    ecx,ecx
	add    edx,1
	cmp    edx,16
	jnz    mres

	popa

	ret


      meka:

	mov    [0xfff4],byte 1


	; save & draw

	mov    [0xfb4a],ax
	mov    [0xfb4c],bx

	push   eax
	push   ebx

	mov    ecx,0
	mov    edx,0

      drm:

	push   eax
	push   ebx
	push   ecx
	push   edx


	pusha

	add    eax,ecx  ; save picture under mouse
	add    ebx,edx

	push   ecx
	call   getpixel
	mov    [0xfb30],ecx
	pop    ecx
	mov    eax,edx
	shl    eax,6
	shl    ecx,2
	add    eax,ecx
	add    eax,mouseunder
	mov    ebx,[0xfb30]
	mov    [eax],ebx

	popa

	mov    eax,edx    ; is there a pixel to draw in mouse pointer
	shl    eax,4
	add    eax,ecx
	mov    ebx,[0xf200]
	add    eax,ebx
	mov    bl,[eax]

	cmp    [0xfff5],byte 0
	jz     mokc
	mov    bl,0

      mokc:
	cmp    bl,0
	jz     no_draw
	cmp    bl,1
	jnz    d1
	mov    eax,0xfb10
	mov    [eax],dword 0x00ffffff
	jmp    yes_draw
       d1:
	cmp    bl,2
	jnz    d2
	mov    eax,0xfb10
	mov    [eax],dword 0x00000000
	jmp    yes_draw
       d2:
	mov    eax,0xfb10
	mov    [eax],dword 0x01000000
	jmp    no_draw

      no_draw:

	pop    edx
	pop    ecx
	pop    ebx
	pop    eax

	jmp    mnext

      yes_draw:

	pop    edx
	pop    ecx
	pop    ebx
	pop    eax

	add    eax,ecx
	add    ebx,edx

	pusha
	mov    ecx,dword [0xfb10] ; draw pixel of mouse
	mov    edi,1
	call   putpixel
	popa

      mnext:

	mov    ebx,[esp+0]
	mov    eax,[esp+4]

	add    ecx,1
	cmp    ecx,16
	jnz    mdrm
	mov    ecx,0
	add    edx,1
	cmp    edx,16
	jnz    mdrm

	jmp    outpointer

     mdrm:
	jmp    drm


      outpointer:

	pop    ebx
	pop    eax

	ret


	; ****   NEXT DATA BYTE FROM MOUSE BUFFER   ****


       decm:

	mov    edi,[mousecount]         ; decrease counter
	mov    eax,[edi]
	sub    eax,1
	mov    [edi],eax

	mov    esi,[mousedata]
	mov    edi,esi
	add    esi,1
	mov    ecx,250
	cld
	rep    movsb

	jmp    uusicheckmouse

      checkmouseret:

	popa

	ret




draw_mouse_pointer:


      pusha

      movzx  ebx,word [0xfb0c]
      movzx  eax,word [0xfb0a]

      call   meka

      popa

      ret




; background

calculatebackground:


	; all black

	mov   [0x400000-8],dword 4      ; size x
	mov   [0x400000-4],dword 2      ; size y

	mov   edi,0x300000              ; set background to black
	mov   eax,0
	mov   ecx,0x0fff00 / 4
	cld
	rep   stosd

	mov   edi,0x400000              ; set os to use all pixels
	mov   eax,0x01010101
	mov   ecx,0x1fff00 / 4
	cld
	rep   stosd

	ret


drawbackground:

%ifdef CFGSYS_VIDEO_VESA_12AND20

       cmp  [0xfe0c],word 0100000000000000b
       jge  dbrv20
       cmp  [0xfe0c],word 0x13
       je   dbrv20

%endif

%ifdef CFG_VIDEO_VESA12
       call  vesa12_drawbackground
       ret
%endif


%ifdef CFGSYS_VIDEO_VESA_12AND20

     dbrv20:

%endif

%ifdef CFG_VIDEO_VESA20

       cmp   [0x400000-12],dword 1
       jne   bgrstr

       call  vesa20_drawbackground_tiled

       ret

     bgrstr:

       call  vesa20_drawbackground_stretch

       ret
%endif ; CFG_VIDEO_VESA20

calculate_edi_esi:

	push  eax
	push  ebx
	xor   edx,edx
	mov   esi,eax
	mov   eax,[0xfe00]
	inc   eax
	mul   ebx
	add   eax,esi
	mov   edi,eax

	pop   ebx
	pop   eax

	ret



;--------------vbe voodoo ------------------------------------------------
vesa20_drawbackground_tiled:

     push ebp
     push eax
     push ebx
     push ecx
     push edx

     mov edx,dword [0x400000-8] ; B
     add edx,dword [0x400000-8] ; +B
     add edx,dword [0x400000-8] ; +B
     push edx

     mov ebp,[0xd000+32+0] ; x start:=(x+Xwin)
     mov ebx,[0xd000+32+4] ; y start:=(y+Ywin)

     mov eax,[0xfe08]
     mul ebx
     xchg ebp, eax  ; BytesPerScanLine*(Ywin+y)
     add ebp, eax   ; +X
     add ebp, eax   ; +X
     add ebp, eax   ; +X

     cmp [0xfbf1],byte 24 ; 24 or 32 bpp ? - x size
     jz bgl243 
     add ebp,eax ; +X
   bgl243:
     add ebp,[0xfe80]  ; +LFB

     ; ebp:=Y*BytesPerScanLine+X*BytesPerPixel+AddrLFB

     call calculate_edi_esi


   dp3:                            ; MAIN LOOP

     cmp [edi+0x400000],byte 1     ; ptrBuffer^<>byte(1)
     je  ybgp

     jmp nbgp
     
   ybgp:

     push eax
     push ebx

     mov ecx,dword [0x400000-8]    ; B
     xor edx,edx                   ; edx:=0
     div ecx                       ; Xstart/B

     ; eax=Int(qn) edx:=Rem

     lea esi,[edx+edx*2]           ; esi:=edx*3

     mov ecx,dword [0x400000-4]    ; ecx:=H
     mov eax,[esp+0]               ; eax:=Ystart
     xor edx,edx                   ;
     div ecx                       ; Ystart/H

     mov eax,edx                   ; eax:=Rem
     xor edx,edx                   ;
     mov ebx,[esp+8]               ; ebx:=B*3
     mul ebx                       ;
     add esi,eax                   ;
     mov eax,[esi+0x300000]
     and eax,0xffffff

     xchg edi, ebp
     stosw
     shr eax,16
     stosb
     xchg ebp, edi                 ; ebp+=3
     cmp [0xfbf1],byte 24          ; 24 or 32 bpp ? - x size
     jz bgl24312 
     inc ebp ; +1
   bgl24312:

     pop ebx
     pop eax

     jmp hook1


   nbgp:
     add ebp,3                     ; +3
     cmp [0xfbf1],byte 24          ; 24 or 32 bpp ? - x size
     jz bgl243121 
     inc ebp ; +1
   bgl243121:

   hook1:

     inc edi                       ; ptrBuffer++
     add esi,3                     ; ptrImage+=3
     inc eax
     cmp eax,[0xd000+32+8]         ; X > xend?
     jg nodp3
     jmp dp3

   nodp3:

     mov ebp,[0xd000+32+0]

     inc ebx

     mov  eax,[0xfe08]
     mul  ebx
     xchg ebp, eax                 ; BytesPerScanLine*(Ywin+y)
     add  ebp, eax                 ; +X
     add  ebp, eax                 ; +X=X*2
     add  ebp, eax                 ; +X=X*3
     cmp  [0xfbf1],byte 24         ; 24 or 32 bpp ? - x size
     jz   bgl2431 
     add  ebp,eax                  ; +X=X*4
   bgl2431:
     add ebp,[0xfe80]              ; +LFB

     ; ebp:=Y*BytesPerScanLine+X*BytesPerPixel+AddrLFB

     call calculate_edi_esi

     cmp ebx,[0xd000+32+12]
     jg  dp4

     jmp dp3

   dp4:

     add esp,4

     pop edx
     pop ecx
     pop ebx
     pop eax
     pop ebp

     ret

; ----------



vesa20_drawbackground_stretch:


     push ebp
     push eax
     push ebx
     push ecx
     push edx

     mov edx,dword [0x400000-8] ; B
     add edx,dword [0x400000-8] ; +B
     add edx,dword [0x400000-8] ; +B
     push edx

     mov ebp,[0xd000+32+0] ; x start:=(x+Xwin)
     mov ebx,[0xd000+32+4] ; y start:=(y+Ywin)

     mov eax,[0xfe08]
     mul ebx
     xchg ebp, eax  ; BytesPerScanLine*(Ywin+y)
     add ebp, eax   ; +X
     add ebp, eax   ; +X
     add ebp, eax   ; +X

     cmp [0xfbf1],byte 24 ; 24 or 32 bpp ? - x size
     jz  sbgl243 
     add ebp,eax ; +X
   sbgl243:
     add ebp,[0xfe80] ; +LFB

     ; ebp:=Y*BytesPerScanLine+X*BytesPerPixel+AddrLFB

     call calculate_edi_esi


   sdp3:                            ; MAIN LOOP

     cmp [edi+0x400000],byte 1     ; ptrBuffer^<>byte(1)
     je  sybgp

     jmp snbgp
     
   sybgp:

     push eax
     push ebx

     mov   eax,dword [0x400000-8]
     xor   edx,edx
     mov   ebx,[esp+4]
     mul   ebx
     xor   edx,edx
     mov   ebx,[0xfe00]
     div   ebx
     lea   esi,[eax+eax*2]
     mov   eax,dword [0x400000-4]
     xor   edx,edx
     mov   ebx,[esp+0]
     mul   ebx
     xor   edx,edx
     mov   ebx,[0xfe04]
     div   ebx

     xor   edx,edx
     mov   ebx,[esp+8]
     mul   ebx
     add   esi,eax
     mov   eax,[esi+0x300000]
     and   eax,0xffffff

     xchg edi, ebp
     stosw
     shr eax,16
     stosb
     xchg ebp, edi                 ; ebp+=3
     cmp [0xfbf1],byte 24          ; 24 or 32 bpp ? - x size
     jz  sbgl24312 
     inc ebp ; +1
   sbgl24312:

     pop ebx
     pop eax

     jmp shook1

   snbgp:
     add  ebp,3                     ; +3
     cmp [0xfbf1],byte 24           ; 24 or 32 bpp ? - x size
     jz  sbgl243121 
     inc ebp ; +1
   sbgl243121:

   shook1:

     inc edi                       ; ptrBuffer++
     add esi,3                     ; ptrImage+=3
     inc eax
     cmp eax,[0xd000+32+8]         ; X > xend?
     jg  snodp3
     jmp sdp3

   snodp3:

     mov ebp,[0xd000+32+0]

     inc ebx

     mov  eax,[0xfe08]
     mul  ebx
     xchg ebp, eax                 ; BytesPerScanLine*(Ywin+y)
     add  ebp, eax                 ; +X
     add  ebp, eax                 ; +X=X*2
     add  ebp, eax                 ; +X=X*3
     cmp  [0xfbf1],byte 24         ; 24 or 32 bpp ? - x size
     jz   sbgl2431 
     add  ebp,eax                  ; +X=X*4
   sbgl2431:
     add ebp,[0xfe80]          ; +LFB

     ; ebp:=Y*BytesPerScanLine+X*BytesPerPixel+AddrLFB

     call calculate_edi_esi

     cmp ebx,[0xd000+32+12]
     jg  sdp4

     jmp sdp3

   sdp4:

     add esp,4

     pop edx
     pop ecx
     pop ebx
     pop eax
     pop ebp

     ret





%ifdef CFG_VIDEO_VESA12
vesa12_drawbackground:


	push  eax
	push  ebx
	push  ecx
	push  edx

	xor   edx,edx
	mov   eax,dword[0x400000-8]
	mov   ebx,dword[0x400000-4]
	mul   ebx
	mov   ebx,3
	mul   ebx
	mov   [imax],eax

	mov   eax,[0xd000+32+0]
	mov   ebx,[0xd000+32+4]

	mov   edi,0 ;no force

      v12dp3:

	push  eax
	push  ebx

	mov   esi,0x300000

	mov   ecx,dword [0x400000-8]
	sub   ecx,1
	and   eax,ecx
	mov   ecx,dword [0x400000-4]
	sub   ecx,1
	and   ebx,ecx

	push  eax

	mov   eax,ebx
	xor   edx,edx

	mov   ebx,dword [0x400000-8]
	add   ebx,dword [0x400000-8]
	add   ebx,dword [0x400000-8]

	mul   ebx
	mov   esi,eax
	pop   eax
	add   esi,eax
	add   esi,eax
	add   esi,eax
	add   esi,0x300000

	pop   ebx
	pop   eax

      v12di4:

	mov   ecx,[esi]

	pusha

	mov   esi,eax
	mov   edi,ebx
	mov   eax,[0xfe00]
	add   eax,1
	mul   ebx
	add   eax,esi
	add   eax,0x400000

	cmp   [eax],byte 1
	jnz   v12nbgp

	mov   eax,[0xfe08]
	mov   ebx,edi
	mul   ebx
	add   eax,esi
	add   eax,esi
	add   eax,esi


	cmp   [0xFBF1],byte 24
	jz    v12bgl3
	add   eax,esi
      v12bgl3:

	push ebx
	push eax
	sub  eax,[0xfe80]
	shr  eax,16
	call set_bank
	pop  eax
	and  eax,65535
	add  eax,0xa0000
	pop  ebx

	mov   [eax],cx
	add   eax,2
	shr   ecx,16
	mov   [eax],cl
	sti


      v12nbgp:

	popa


	add   esi,3
	inc   eax
	cmp   eax,[0xd000+32+8]
	jg    v12nodp31
	jmp   v12dp3

      v12nodp31:

	mov   eax,[0xd000+32+0]

	inc   ebx
	cmp   ebx,[0xd000+32+12]
	jg    v12dp4

	jmp   v12dp3

      v12dp4:

	pop   edx
	pop   ecx
	pop   ebx
	pop   eax

	ret
%endif ; CFG_VIDEO_VESA12


imax    dd 0x0



; delay in 1/1000 sec

delay_ms:

	push  eax
	push  ecx

	mov   ecx,esi
	imul  ecx,66

	in    al,0x61
	and   al,0x10
	mov   ah,al
	cld

 cnt1:  in    al,0x61
	and   al,0x10
	cmp   al,ah
	jz    cnt1

	mov   ah,al
	loop  cnt1

	pop   ecx
	pop   eax

	ret


set_app_param:


	pusha

	mov  edi,[0x3010]
	mov  [edi],eax

	popa

	ret




; delay in 1/100 secs

delay_hs:


	push  eax
	push  ecx
	push  edx

	mov   edx,[0xfdf0]
	add   edx,eax

      newtic:
	mov   ecx,[0xfdf0]
	cmp   edx,ecx
	jbe   zerodelay

	call  change_task

	jmp   newtic

      zerodelay:
	pop   edx
	pop   ecx
	pop   eax

	ret


; memory move in bytes

memmove:

; eax = from
; ebx = to
; ecx = no of bytes

    pusha

    cld

    ; ecx no to move

    mov  esi,eax
    mov  edi,ebx
    rep  movsb

    popa

    ret


%ifdef CFG_RANDOM_WIN
random_shaped_window:

;
;  eax = 0    giving address of data area
;      ebx    address
;  ebx = 1    shape area scale
;      ebx    2^ebx scale

     cmp  eax,0
     jne  rsw_no_address
     mov  eax,[0x3000]
     shl  eax,8

     mov  [eax+0x80000+0x80],ebx
   rsw_no_address:

     cmp  eax,1
     jne  rsw_no_scale
     mov  eax,[0x3000]
     shl  eax,8
     mov  [eax+0x80000+0x84],bl
   rsw_no_scale:

     ret
%endif ; CFG_RANDOM_WIN

; calculate fat chain

calculatefatchain:

   pusha

   mov  esi,0x100000+512
   mov  edi,0x280000

  fcnew:
   xor  eax,eax
   xor  ebx,ebx
   xor  ecx,ecx
   xor  edx,edx
   mov  al,[esi+0]  ; 1
   mov  bl,[esi+1]
   and  ebx,15
   shl  ebx,8
   add  eax,ebx
   mov  [edi],ax
   add  edi,2

   xor  eax,eax
   xor  ebx,ebx
   xor  ecx,ecx
   xor  edx,edx
   mov  bl,[esi+1]  ; 2
   mov  cl,[esi+2]
   shr  ebx,4
   shl  ecx,4
   add  ecx,ebx
   mov  [edi],cx
   add  edi,2

   add  esi,3

   cmp  edi,0x280000+4100*4
   jnz  fcnew

   popa

   ret


; restore fat chain

restorefatchain:

   pusha

   mov  esi,0x280000
   mov  edi,0x100000+512

  fcnew2:
   cld
   xor  eax,eax
   xor  ebx,ebx                    
   xor  ecx,ecx                    ;   esi  XXXXxxxxxxxx  yyyyyyyyYYYY
   xor  edx,edx                    
   mov  ax,[esi]                   ;   edi  xxxxxxxx YYYYXXXX yyyyyyyy
   mov  bx,ax
   shr  bx,8
   and  ebx,15
   mov  [edi+0],al  ; 1 -> 1 & 2
   mov  [edi+1],bl
   add  esi,2

   xor  eax,eax
   xor  ebx,ebx
   xor  ecx,ecx
   xor  edx,edx

   cld
   mov  bx,[esi]
   mov  cx,bx

   shr  ecx,4
   mov  [edi+2],cl

   and  ebx,15
   shl  ebx,4
   mov  edx,[edi+1]
   add  edx,ebx
   mov  [edi+1],dl  ; 2 -> 2 & 3
   add  esi,2

   add  edi,3

   cmp  edi,0x100000+512+0x1200
   jb   fcnew2

   mov  eax,0x100000+512           ; duplicate fat chain
   mov  ebx,0x100000+512+0x1200
   mov  ecx,0x1200
   call memmove

   popa

   ret




read_floppy_file:

; as input
;
; eax pointer to file    
; ebx file lenght
; ecx start 512 byte block number
; edx number of blocks to read
; esi pointer to return/work area (atleast 20 000 bytes)
;
;
; on return
;
; eax = 0 command succesful
;       1 no fd base and/or partition defined
;       2 yet unsupported FS
;       3 unknown FS
;       4 partition not defined at hd
;       5 file not found
; ebx = size of file


     pusha

     mov  edi,esi
     add  edi,1024

     mov  esi,0x100000+19*512
     sub  ecx,1
     shl  ecx,9
     add  esi,ecx

     shl  edx,9
     mov  ecx,edx

     cld
     rep  movsb

     popa

     ret



%ifdef CFG_HD
read_hd_file:

; as input
;
; eax pointer to file   (0 = read only first sector of drive: eg 'label')
; ebx file lenght
; ecx start 512 byte block number
; edx number of blocks to read
; esi pointer to return/work area (atleast 20 000 bytes)
;
;
; on return
;
; eax = 0 command succesful
;       1 no hd base and/or partition defined
;       2 yet unsupported FS
;       3 unknown FS
;       4 partition not defined at hd
;       5 file not found
; ebx = size of file


    push ebx
    push eax 
    push ebx
    push ecx
    push edx
    push esi
    push edi

    cmp  [hdbase],dword 0
    jz   hdnotok
    cmp  [fat32part],dword 0
    jz   hdnotok

    jmp  hdrok

  hdnotok:
    mov  eax,1
    jmp  fs_unsupport

  hdrok:

    mov  eax,0          ; read master boot record
    mov  ebx,[esp+4]
    call hd_read

    mov  ecx,[fat32part]

    mov  esi,[esp+4]    ; save partition to be read
    add  esi,8
    mov  [esi],ecx

    mov  ebx,0
    mov  esi,[esp+4]
    add  esi,0x1be+8
    cld

  partition_nro:
    cmp  [esi],dword 0
    jnz  partok
    mov  eax,4
    jmp  fs_unsupport
  partok:
    add  ebx,[esi]
    add  esi,0x10
    loop partition_nro,ecx

    mov  esi,[esp+4]    ; save partition start
    add  esi,12
    mov  [esi],ebx

    mov  eax,ebx        ; root dir at partition X
    mov  ebx,[esp+4]
    add  ebx,512
    call hd_read

    ; NTFS ?
    mov  esi,[esp+4]
    mov  eax,[esi+512+3]
    cmp  eax,'NTFS'
    jnz  nontfs
    mov  esi,[esp+4]
    mov  [esi],dword 'NTFS'
    add  esi,4
    mov  [esi],dword '    '
    mov  eax,2
    jmp  fs_unsupport
  nontfs:

    ; FAT16 ?
    mov  esi,[esp+4]
    mov  eax,[esi+512+55]
    cmp  eax,'AT16'
    jnz  nofat16
    mov  esi,[esp+4]
    mov  [esi],dword 'FAT1'
    add  esi,4
    mov  [esi],dword '6   '
    mov  eax,2
    jmp  fs_unsupport
  nofat16:

    ; FAT32 ?
    mov  esi,[esp+4]
    mov  eax,[esi+512+0x53]
    cmp  eax,'AT32'
    jnz  nofat32
    mov  eax,[esp+20] ; read only 1. sector of drive ?
    cmp  eax,0
    jz   nf32r
    call read_fat32_file
  nf32r:
    jmp  fs_unsupport

  nofat32:

    ; unknown fs

    mov  esi,[esp+4]
    mov  [esi],dword 'FS ?'
    add  esi,4
    mov  [esi],dword '    '
    mov  eax,3
    jmp  fs_unsupport

  fs_unsupport:
    pop  edi
    pop  esi
    pop  edx
    pop  ecx
    pop  ebx
    add  esp,4
    pop  ebx

    ret


read_fat32_file:

  mov  esi,[esp+8]        ; fat start cluster
  add  esi,512+0xE
  xor  ecx,ecx
  mov  cx,[esi]
  mov  esi,[esp+8]
  add  esi,12
  add  ecx,[esi]
  mov  edi,[esp+8]
  add  edi,16
  mov  [edi],ecx

  mov  esi,[esp+8]        ; root start block
  add  esi,512+0x2C
  mov  ecx,[esi]
  mov  edi,[esp+8]
  add  edi,24
  mov  [edi],ecx

  mov  esi,[esp+8]        ; number of fats
  add  esi,512+0x10
  xor  ecx,ecx
  mov  cl,[esi]
  mov  edi,[esp+8]
  add  edi,28
  mov  [edi],ecx

  mov  esi,[esp+8]        ; fat size
  add  esi,512+0x24
  mov  ecx,[esi]
  mov  edi,[esp+8]
  add  edi,32
  mov  [edi],ecx

  mov  esi,[esp+8]        ; hidden sectors
  add  esi,512+0x1C
  mov  ecx,[esi]
  mov  edi,[esp+8]
  add  edi,36
  mov  [edi],ecx

			  ; block start cluster :=
			  ; number_of_fats*fat_size+fat_start_cluster;
  xor  edx,edx
  mov  edi,[esp+8]
  add  edi,28
  mov  eax,[edi]
  mov  edi,[esp+8]
  add  edi,32
  mov  ebx,[edi]
  mul  ebx
  mov  edi,[esp+8]
  add  edi,16
  add  eax,[edi]
  mov  edi,[esp+8]
  add  edi,40
  mov  [edi],eax

  mov  esi,[esp+8]       ; sectors per cluster
  add  esi,512+0xd
  xor  ecx,ecx
  mov  cl,[esi]
  mov  edi,[esp+8]
  add  edi,44
  mov  [edi],ecx


  ; all needed hd variables are now read and saved



  ; FIND THE FIRST CLUSTER OF FILE

  mov  edi,[esp+8]           ; start of file search in directory que
  mov  [edi+48],dword 1

  mov  esi,[esp+8]           ; eax <- first cluster of root
  add  esi,24
  mov  eax,[esi]

  mov  ebx,[esp+20]
  cmp  ebx,0
  jnz  yesreadmore
  jmp  hdfilefound

 yesreadmore:

  mov  ebx,[esp+8]
  add  ebx,1024
  call read_fat32_block

  mov  edi,ebx                  ; Point at directory

  push eax                      ; number of entries to check

  xor  edx,edx
  mov  eax,16
  mov  esi,[esp+8+4]
  mov  ebx,[esi+44]
  mul  ebx
  mov  edx,eax

 hdsearch:
  mov  esi,[esp+24+4]             ; Name of file we want
  mov  ebx,[esp+8+4]
  add  esi,[ebx+48]

  xor  ecx,ecx
  mov  cl, 11
  cld
  rep  cmpsb                    ; Found the file?
  je   hdfifound                ; Yes
  add  cl, 21                   ; Offset to next directory entry
  add  edi,ecx                  ; Advance to next entry
  sub  edx,1
  cmp  edx,0
  jz   nextclustercheck
  jmp  hdsearch

 nextclustercheck:

  pop  eax

  call read_next_in_fat

  mov  ecx,0x0f000000              ; file not found
  cmp  eax,ecx
  jb   nohdeof2
  mov  eax,6
  ret


 nohdeof2:

  jmp  yesreadmore

 hdfifound:

  pop  eax

  mov  ebx,[edi-11+28]             ; save file size
  mov  [esp+8+16+4],ebx

  mov  ax,[edi-11+20]              ; first cluster of find file
  shl  eax,16
  mov  ax,[edi-11+26]

  mov  ebx,[esp+20]
  sub  ebx,12
  mov  [esp+20],ebx
  cmp  ebx,0
  jz   hdlastsearch

  mov  ecx,[esp+8]                  ; adjust to read next directory
  mov  ebx,[ecx+48]
  add  ebx,12
  mov  [ecx+48],ebx

  jmp  yesreadmore

 hdlastsearch:


  ; FIRST CLUSTER OF FILE IS FOUND - READ THE FILE


 hdfilefound:


 search_next_fat:

  mov  edi,[esp+8]           ; ebx <- sectors per cluster
  mov  ebx,[edi+44]
  mov  ecx,[esp+12+4]        ; esi <- wanted cluster
  cmp  ecx,ebx
  jbe  hdfilerightblock

  call read_next_in_fat

  mov  ebx,0x0f000000        ; eof ?
  cmp  eax,ebx
  jb   nohdeof
  mov  eax,5
  ret

 nohdeof:
  mov  ecx,[esp+12+4]        ; sub clusters_per_sector from wanted cluster
  mov  edi,[esp+8]
  sub  ecx,[edi+44]
  mov  [esp+12+4],ecx

  jmp  search_next_fat


 hdfilerightblock:

  mov  ebx,[esp+8]
  add  ebx,1024
  call read_fat32_block

  mov  esi,[esp+12+4]        ; give the wanted block
  sub  esi,1
  cmp  esi,0
  jz   rightblock2
  shl  esi,9
  add  esi,1024
  add  esi,[esp+8]
  mov  edi,[esp+8]
  add  edi,1024
  mov  ecx,128
  cld
  rep  movsd

 rightblock2:

  mov eax,0

  ret


read_next_in_fat:

  pusha

  shl  eax,2

  mov  ecx,eax
  shr  eax,9    ; cluster no

  mov  esi,[esp+12+32]
  add  esi,16
  add  eax,[esi]
  mov  ebx,[esp+12+32]
  add  ebx,512
  call hd_read

  and  ecx,511  ; in cluster
  add  ecx,ebx

  mov  eax,[ecx]

  mov  [esp+28],eax

  popa

  ret




read_fat32_block:

  pusha

  push ebx

  sub  eax,2
  xor  edx,edx

  mov  edi,[esp+12+32+4]    ; * sectors per cluster
  add  edi,44
  mov  ebx,[edi]
  mul  ebx

  mov  edi,[esp+12+32+4]    ; + block start cluster
  add  edi,40
  add  eax,[edi]      

  mov  edi,[esp+12+32+4]
  add  edi,44
  mov  ecx,[edi]

  pop  ebx

 nf32b:

  call hd_read

  add  eax,1

  add  ebx,512

  loop nf32b,ecx

  popa

  ret




; read hd


hd_read:

;   eax block to read
;   ebx destination



    pusha                 ; scan cache

    mov  ecx,[0xFE10]     ; entries in cache
    cmp  ecx,0
    jz   nohdincache

    mov  esi,0x600000+8
    mov  edi,1

  hdreadcache:

    mov  edx,[esi]
    cmp  edx,eax
    jnz  nohdcache

    mov  edx,[hdpos]
    cmp  dl,[esi+4]
    jnz  nohdcache

    mov  esi,edi
    shl  esi,9
    add  esi,0x600000+65536
    mov  edi,ebx
    mov  ecx,512
    cld
    rep  movsb
    popa
    ret

  nohdcache:

    add  esi,8
    add  edi,1
    loop hdreadcache

  nohdincache:
    popa


    push  esi
    push  edi

    push  eax
    push  ebx
    push  ecx
    push  edx


  newhdread:

    xor   eax,eax

    mov   edx,[hdbase]

    add   edx,1
    mov   al,0
    out   dx,al

    add   edx,1
    mov   al,1
    out   dx,al

    add   edx,1
    mov   al,[esp+12]
    out   dx,al

    add   edx,1
    mov   al,[esp+13]
    out   dx,al

    add   edx,1
    mov   al,[esp+14]
    out   dx,al

    add   edx,1
    mov   al,[esp+15]
    and   al,1+2+4+8
    mov   bl,[hdid]
    add   al,bl
    add   al,128+64+32
    out   dx,al

    add   edx,1
    mov   al,20h
    out   dx,al

  hdwait:

    in    al,dx
    mov   bl,al
    and   bl,128
    cmp   bl,128
    jz    hdwait

    mov   edi,[esp+8]
    mov   ecx,256
    mov   edx,[hdbase]

  hdloop2:
    in    ax,dx
    stosw            
    loop  hdloop2,ecx


    mov   esi,edi          ; move sector to cache
    sub   esi,512

    mov   ecx,512

    mov   edi,[0xFE10]     ; inc counter
    cmp   edi,3600
    jbe   hdenok
    mov   edi,0
  hdenok:
    add   edi,1
    mov   [0xFE10],edi

    shl   edi,9
    add   edi,0x600000+65536

    cld
    rep   movsb            ; move data


    mov   eax,[esp+12]     ; move the read sector to memory
    mov   edi,[0xFE10]
    shl   edi,3
    add   edi,0x600000
    mov   [edi],eax
    mov   al,[hdpos]
    mov   [edi+4],al


   hde:

    pop   edx
    pop   ecx
    pop   ebx
    pop   eax

    pop   edi
    pop   esi

    ret
%endif ; CFG_HD


sys_programirq:


    mov   edx,ebx
    shl   edx,2
    add   edx,irq_owner
    mov   edx,[edx]

    mov   edi,[0x3010]
    mov   edi,[edi+0x4]

    cmp   edx,edi
    je    spril1

    mov   ecx,1
    ret

  spril1:

    mov   esi,eax
    shl   ebx,6
    add   ebx,irq00read
    mov   edi,ebx
    mov   ecx,16
    cld
    rep   movsd

    mov   ecx,0
    ret




get_irq_data:

     mov   edx,eax           ; check for correct owner
     shl   edx,2
     add   edx,irq_owner
     mov   edx,[edx]

     mov   edi,[0x3010]
     mov   edi,[edi+0x4]

     cmp   edx,edi
     je    gidril1

     mov   ecx,2
     ret

  gidril1:

     mov   ebx,eax
     shl   ebx,12
     add   ebx,0x2e0000
     mov   eax,[ebx]
     mov   ecx,1
     cmp   eax,0
     jz    gid1

     sub   eax,1
     mov   esi,ebx
     mov   [ebx],eax
     movzx ebx,byte [ebx+0x10]
     add   esi,0x10
     mov   edi,esi
     add   esi,1
     mov   ecx,4000 / 4
     cld
     rep   movsd
     mov   ecx,0
   gid1:
     ret




r_f_port_area:

     cmp   eax,0
     je    r_port_area
     jmp   free_port_area

   r_port_area:

     pusha

     cmp   ebx,ecx            ; beginning > end ?
     jg    rpal1

     mov   esi,[0x2d0000]

     cmp   esi,0              ; no reserved areas ?
     je    rpal2

     cmp   esi,255            ; max reserved
     jge   rpal1

   rpal3:

     mov   edi,esi
     shl   edi,4
     add   edi,0x2d0000

     cmp   ebx,[edi+8]
     jg    rpal4
     cmp   ecx,[edi+4]
     jb    rpal4

     jmp   rpal1

   rpal4:

     dec   esi
     jnz   rpal3

     jmp   rpal2


   rpal1:

     popa

     mov   eax,1
     ret


   rpal2:

     popa

     mov   edi,[0x2d0000]
     add   edi,1
     mov   [0x2d0000],edi

     shl   edi,4
     add   edi,0x2d0000

     mov   esi,[0x3010]
     mov   esi,[esi+0x4]

     mov   [edi],esi

     mov   [edi+4],ebx
     mov   [edi+8],ecx

     mov   eax,0

     ret




free_port_area:

     pusha
     
     mov   esi,[0x2d0000]     ; no reserved areas ?
     cmp   esi,0
     je    frpal2

     mov   edx,[0x3010]
     mov   edx,[edx+4]

   frpal3:

     mov   edi,esi
     shl   edi,4
     add   edi,0x2d0000

     cmp   edx,[edi]
     jne   frpal4
     cmp   ebx,[edi+4]
     jne   frpal4
     cmp   ecx,[edi+8]
     jne   frpal4

     jmp   frpal1

   frpal4:

     dec   esi
     jnz   frpal3

   frpal2:

     popa

     mov   eax,1
     ret


   frpal1:

     mov   ecx,256
     sub   ecx,esi
     shl   ecx,4

     mov   esi,edi
     add   esi,16

     cld
     rep   movsb

     popa

     mov   eax,0

     ret



reserve_free_irq:

     cmp   eax,0
     jz    reserve_irq

     mov   edi,ebx
     shl   edi,2
     add   edi,irq_owner
     mov   edx,[edi]

     mov   eax,[0x3010]
     mov   eax,[eax+0x4]

     mov   ecx,1

     cmp   edx,eax
     jne   fril1

     mov   [edi],dword 0

     mov   ecx,0

   fril1:

     ret


  reserve_irq:

     mov   edi,ebx
     shl   edi,2
     add   edi,irq_owner
     mov   edx,[edi]
     mov   ecx,1
     cmp   edx,0
     jne   ril1

     mov   edx,[0x3010]
     mov   edx,[edx+0x4]
     mov   [edi],edx

     mov   ecx,0

   ril1:

     ret


%ifdef CFG_FAT32

;***************************************
; BEGIN FAT32 by Paolo Minazzi
;***************************************


;       CLUSTER[0]  CLUSTER[1]  CLUSTER[2] CLUSTER[3]
;
;###    ######## -->######## -->########   ########
;#2#--> #      # |  #      # |  #      #   #      #
;###    #      # |  #      # |  #      #   #      #
;       #      # |  #      # |  #      #   #      #
;       #      # |  #      # |  #      #   #      #
;       #      # |  #      # |  #      #   #      #          CLUSTER
;       #      # |  #      # |  #      #   #      #
;       #      # |  #      # |  #      #   #      #
;       #      # |  #      # |  #      #   #      #
;       ######## |  ######## |  ########   ########
;                |           |
;                |           ------------------
;                |                            |
;                --------------------         |   
;                                   |         |
;      ########    ########    ########   ########
;      # xxxx #    # xxxx #    #   3  #   #   4  #
;      ########    ########    ########   ########          FAT
;
;       FAT[0]       FAT[1]      FAT[2]    FAT[3]
;
;


RESERVED_SECTOR:     dd 0x20
PARTITION_START:     dd 0x3f
FAT32_SIZE:          dd 0x1f3a
NUMBER_OF_FATS:      dd 0x2
SECTOR_PER_CLUSTER:  dd 0x8
buffer:              resb 512
cluster:             dd 0
cluster1:            dd 0
valid_fat:           db 0

set_FAT32_variables:

    push eax
    push ebx
    xor eax,eax                        ;read MBR
    mov ebx,buffer
    call hd_read_new

    mov  ecx,[fat32part]
    cmp  ecx,0
    jnz  sfvok
    pop  ebx
    add  esp,4
    mov  eax,4
    ret
  sfvok:

    mov  eax,0
    mov  esi,ebx
    add  esi,0x1be+8
  sf32vl:
    add  eax,[esi]
    add  esi,0x10
    loop sf32vl,ecx
    mov  [PARTITION_START],eax

    call hd_read_new                       ;read boot sector of partition

    mov  [valid_fat],byte 0
    cmp [ebx+0x53],dword 'AT32'
    jz  sfvl
    pop ebx
    pop eax
    ret
  sfvl:
    mov  [valid_fat],byte 1

    movzx eax,word [ebx+0xe]
    mov [RESERVED_SECTOR],eax

    mov eax,[ebx+0x24]
    mov [FAT32_SIZE],eax

    movzx eax,byte [ebx+0x10]
    mov [NUMBER_OF_FATS],eax

    movzx eax,byte [ebx+0xd]
    mov [SECTOR_PER_CLUSTER],eax
    pop ebx
    pop eax
    ret

set_FAT:
;--------------------------------
; set_FAT
;
; input :  EAX = #block
;          EDX = value to save
; output:  nothing
;--------------------------------
   push eax ;[esp+16]
   push ebx ;[esp+12]
   push ecx ;[esp+8]
   push edx ;[esp+4]
   push esi ;[esp]

   shl eax,2
   xor edx,edx
   mov ebx,512
   div ebx
   mov esi,edx

   add eax,[RESERVED_SECTOR]
   add eax,[PARTITION_START]
   mov ebx,buffer

   mov ecx,[NUMBER_OF_FATS]
   mov edx,[esp+4]
set_an_other_FAT:
   call hd_read_new
	
   mov [buffer+esi],edx
   call hd_write_new

   add eax,[FAT32_SIZE]
   loop set_an_other_FAT

   pop esi
   pop edx
   pop ecx
   pop ebx
   pop eax
   ret
	

get_FAT:
;--------------------------------
; set_FAT
;
; input :  EAX = #block
; output:  EAX = #block
;--------------------------------
   push ebx
   push edx
   
   shl eax,2
   xor edx,edx
   mov ebx,512
   div ebx
   
   add eax,[RESERVED_SECTOR]
   add eax,[PARTITION_START]
   mov ebx,buffer

   call hd_read_new

   mov eax,[buffer+edx]

   pop edx
   pop ebx
   ret

find_pointer_in_FAT:
;--------------------------------
; For each item of FAT do
;   if (item read == EAX)
;     return #item
;     carry=0
;   else
;     carry=1;
;
; input :  EAX = #block
; output:  if CARRY=0 then EAX = #block
;          if CARRY=1 then EAX = distroyed
;--------------------------------
  add eax,2
  push eax
  push ecx
  push edx

  mov ecx,[FAT32_SIZE]
  shl ecx,9-2                 ; *512 for sector   /4 for dword 
  mov edx,2

adr78:
  mov eax,edx
  call get_FAT
  cmp eax,[esp+8]
  je found_pointer
  inc edx
  loop adr78

  pop edx
  pop ecx
  add esp,4
  stc
  ret

found_pointer:
  mov eax,edx
  pop edx
  pop ecx
  add esp,4
  clc
  ret

analyze_directory:
;--------------------------------
; input  : EAX = first cluster of the directory
;          EBX = pointer to filename
; output : IF CARRY=0 EAX = sector where th file is found
;                     EBX = pointer in buffer
;                     [buffer .. buffer+511]
;                     ECX,EDX,EDI,EDI not changed
;          IF CARRY=1 
;--------------------------------
   push ebx ;[esp+16]
   push ecx
   push edx
   push esi
   push edi


adr56:
   mov [cluster_tmp],eax

   mov ecx,[FAT32_SIZE]
   imul ecx,[NUMBER_OF_FATS]
   add ecx,[RESERVED_SECTOR]
	
   sub eax,2
   imul eax,[SECTOR_PER_CLUSTER]
   add eax,[PARTITION_START]
   add eax,ecx

   mov edx,[SECTOR_PER_CLUSTER]
   
new_sector_in_cluster:
   mov ebx,buffer
   call hd_read_new

   mov ecx,512/32

adr1_analyze:
   mov esi,[esp+16]
   mov edi,ebx
   cld
   push ecx
   mov ecx,11
   rep cmpsb
   pop ecx
   je found_file_analyze

   add ebx,32
   loop adr1_analyze

   inc eax
   dec edx
   jne new_sector_in_cluster
    
   mov eax,[cluster_tmp]
   call get_FAT
   mov [cluster_tmp],eax
   cmp eax,0x0fffffff
   jne adr56


   pop edi
   pop esi
   pop edx
   pop ecx
   add esp,4
   stc        ;file not found
   ret

found_file_analyze:
   pop edi
   pop esi
   pop edx
   pop ecx
   add esp,4
   clc        ;file found
   ret

cluster_tmp: dd 0 ; common variable used by
		  ; analyze_directory and analyze_directory_to_write

analyze_directory_to_write:
;--------------------------------
; input  : EAX = first cluster of the directory
; output : IF CARRY=0 EAX = sector where th file is found
;                     EBX = pointer in buffer
;                     [buffer .. buffer+511]
;                     ECX,EDX,EDI,EDI not changed
;          IF CARRY=1 
;--------------------------------

   push ecx
   push edx
   push esi
   push edi

adr561:
   mov [cluster_tmp],eax

   mov ecx,[FAT32_SIZE]
   imul ecx,[NUMBER_OF_FATS]
   add ecx,[RESERVED_SECTOR]

   sub eax,2
   imul eax,[SECTOR_PER_CLUSTER]
   add eax,[PARTITION_START]
   add eax,ecx

   mov edx,[SECTOR_PER_CLUSTER]

new_sector_in_cluster1:
   mov ebx,buffer
   call hd_read_new

   mov ecx,512/32

adr1_analyze1:
   cmp byte [ebx],0
   jne avanti
   jmp found_file_analyze1

avanti:
   add ebx,32
   loop adr1_analyze1

   inc eax
   dec edx
   jne new_sector_in_cluster1

   mov eax,[cluster_tmp]
   call get_FAT
   mov [cluster_tmp],eax
   cmp eax,0x0fffffff
   jne adr561j
   jmp av3
 adr561j:
   jmp adr561  
 av3:

   mov eax,2                       ;this block of code add a new cluster
   call get_free_FAT               ;for the directory because the directory
				   ;is full

   mov edx,0x0fffffff
   call set_FAT

   mov edx,eax
   mov eax,[cluster_tmp]
   call set_FAT

   mov ecx,512/4
   xor eax,eax
   mov edi,buffer
   cld
   rep stosd

   mov eax,[SECTOR_PER_CLUSTER]
   imul eax,edx
   add eax,[PARTITION_START]
   add eax,[RESERVED_SECTOR]
   mov ecx,[FAT32_SIZE]
   imul ecx,[NUMBER_OF_FATS]
   add eax,ecx
   mov ebx,buffer
   mov ecx,[SECTOR_PER_CLUSTER]

set_empty_directory:
   call hd_write_new
   inc eax
   loop set_empty_directory

   sub eax,[SECTOR_PER_CLUSTER]

found_file_analyze1:

   pop edi
   pop esi
   pop edx
   pop ecx
   clc        ;file found
   ret

get_data_cluster:
;--------------------------------
; input :  EAX = # block
;          EBX = pointer to buffer
;          ECX = # bytes to read in buffer
; output:  nothing
;--------------------------------       
%define ORG_EAX [esp+20]
%define ORG_EBX [esp+16]
%define ORG_ECX [esp+12]
%define ORG_EDX [esp+8]
%define ORG_ESI [esp+4]
%define ORG_EDI [esp]

   push eax ; ORG_EAX
   push ebx ; ORG_EBX
   push ecx ; ORG_ECX
   push edx ; ORG_EDX
   push esi ; ORG_ESI
   push edi ; ORG_EDI

   mov edi,ebx

   xor edx,edx
   mov eax,ecx
   mov ebx,512
   div ebx

   mov ecx,[FAT32_SIZE]
   imul ecx,[NUMBER_OF_FATS]
   add ecx,[RESERVED_SECTOR]
	
   mov ebx,ORG_EAX
   sub ebx,2
   imul ebx,[SECTOR_PER_CLUSTER]
   add ebx,[PARTITION_START]
   add ebx,ecx
   
   mov ecx,eax

   mov eax,ebx
   mov ebx,buffer
   jecxz not_whole

read:
   call hd_read_new

   mov esi,ebx
   cld
   push ecx
   mov ecx,512
   rep movsb
   pop ecx

   inc eax
   loop read

not_whole:
   call hd_read_new
   mov esi,ebx
   mov ecx,edx
   cld
   rep movsb

   pop edi
   pop esi
   pop edx
   pop ecx
   pop ebx
   pop eax
   ret

set_data_cluster:
;--------------------------------
; input :  EAX = # block
;          EBX = pointer to buffer
;          ECX = # bytes to read in buffer
; output:  nothing
;--------------------------------       
   push eax ; ORG_EAX
   push ebx ; ORG_EBX
   push ecx ; ORG_ECX
   push edx ; ORG_EDX
   push esi ; ORG_ESI
   push edi ; ORG_EDI

   mov esi,ebx

   xor edx,edx
   mov eax,ecx
   mov ebx,512
   div ebx

   mov ecx,[FAT32_SIZE]
   imul ecx,[NUMBER_OF_FATS]
   add ecx,[RESERVED_SECTOR]
	
   mov ebx,ORG_EAX
   sub ebx,2
   imul ebx,[SECTOR_PER_CLUSTER]
   add ebx,[PARTITION_START]
   add ebx,ecx
   
   mov ecx,eax

   mov eax,ebx
   mov ebx,buffer
   jecxz not_whole2

write:
   mov edi,ebx
   cld
   push ecx
   mov ecx,512
   rep movsb
   pop ecx
   call hd_write_new

   inc eax
   loop write

not_whole2:
   mov edi,ebx
   mov ecx,edx
   cld
   rep movsb
   call hd_write_new

   pop edi
   pop esi
   pop edx
   pop ecx
   pop ebx
   pop eax
   ret

get_free_FAT:
;------------------------------------------
; input  :  EAX = # cluster for start the searching
; output :  EAX = # first cluster found free
;-------------------------------------------
   push ebx
   push ecx
   push edx

   mov ecx,eax

   shl eax,2
   xor edx,edx
   mov ebx,512
   div ebx

   add eax,[RESERVED_SECTOR]
   add eax,[PARTITION_START]
   mov ebx,buffer
   call hd_read_new
	
check_new:
   cmp dword [buffer+edx],0
   je found
   add edx,4
   inc ecx
   cmp edx,512
   jb check_new
	
   inc eax
   call hd_read_new
   xor edx,edx
   jmp check_new
	
found:  
   mov eax,ecx
   pop edx
   pop ecx
   pop ebx
   ret
	
file_write:
;--------------------------------          
; EBX = pointer to filename  PAR0
; EDX = pointer to buffer    PAR1
; ECX = file length          PAR2
; ESI = cluster init         PAR3
; call write
;--------------------------------

    cmp  [fat32part],dword 0
    jnz  sfwok
    mov  eax,4
    ret
  sfwok:

    cmp  [valid_fat],byte 1
    jz   sfwok2
    mov  eax,3
    ret
  sfwok2:

   pushad

   pusha
   push eax
   push dword 2
   call file_delete
   add  esp,8
   popa

   mov [0xfe10],dword 0


%define PAR0 [esp+28]        ;EAX
%define PAR1 [esp+24]        ;ECX
%define PAR2 [esp+16]        ;EBX
%define PAR3 [esp+20]        ;EDX

   ;mov eax,PAR3
   ;sub eax,2

   ;imul eax,[SECTOR_PER_CLUSTER]
   ;add eax,[RESERVED_SECTOR]
   ;add eax,[PARTITION_START]

   ;mov ecx,[FAT32_SIZE]
   ;imul ecx,[NUMBER_OF_FATS]
   ;add eax,ecx

   ;mov ebx,buffer
   ;call hd_read_new

   ;mov ecx,20
;adr1:
   ;cmp byte [ebx],0
   ;je found1
   ;add ebx,32
   ;loop adr1

   mov eax,PAR3
   mov ebx,PAR0
   call analyze_directory_to_write
   
found1:

   push eax
   mov eax,2
   call get_free_FAT
   mov [cluster],eax
   pop eax

   push ebx
   mov ebx,buffer
   call hd_read_new
   pop ebx

   mov esi,PAR0             ;file name
   mov edi,ebx              
   mov ecx,11
   cld
   rep movsb

   mov ecx,PAR2
   mov [ebx+28],ecx         ;file size
   mov ecx,[cluster]      
   mov [ebx+26],cx          ;16 bits low  of cluster
   shr ecx,16            
   mov [ebx+20],cx          ;16 bits high of cluster
   mov byte [ebx+11],0x20   ;attrib

   mov ebx,buffer           ;save the directory name,length,cluster
   call hd_write_new

   imul esi,[SECTOR_PER_CLUSTER],512

   mov eax,[cluster]
   mov ebx,PAR1
   mov ecx,esi
   cmp PAR2,esi
   jae adr3
   mov ecx,PAR2
adr3:
   call set_data_cluster
   
adr2:
   add PAR1,ecx
   sub PAR2,ecx

   jz file_saved_OK

   mov eax,[cluster]
   inc eax
   call get_free_FAT
   mov [cluster1],eax

   mov eax,[cluster]
   mov edx,[cluster1]
   mov ecx,esi
   cmp PAR2,esi
   jae adr4
   mov ecx,PAR2
adr4:
   call set_FAT

   mov eax,[cluster1]
   mov ebx,PAR1
   call set_data_cluster

   mov eax,[cluster1]
   mov [cluster],eax

   jmp adr2

file_saved_OK:

   mov eax,[cluster]
   mov edx,0x0fffffff
   call set_FAT

   popad
   ret 

file_read:
;--------------------------------          
; EBX = pointer to filename  PAR0
; EDX = pointer to buffer    PAR1
; ECX = file length          PAR2
; ESI = cluster init         PAR3
; call file_read
;--------------------------------
   pushad

%define PAR0 [esp+28]        ;EAX
%define PAR1 [esp+24]        ;ECX
%define PAR2 [esp+16]        ;EBX
%define PAR3 [esp+20]        ;EDX

   mov eax,PAR3
   mov ebx,PAR0
   call analyze_directory
   jc file_to_read_not_found


found1r:

   mov ax,[ebx+20]
   shl eax,16
   mov ax,[ebx+26]

   imul esi,[SECTOR_PER_CLUSTER],512

   mov ebx,PAR1
   mov ecx,esi
   cmp PAR2,esi
   jae adr3r
   mov ecx,PAR2
adr3r:

   call get_data_cluster

adr2r:
   add PAR1,ecx
   sub PAR2,ecx

   jz file_read_OK

   call get_FAT

   mov ebx,PAR1
   mov ecx,esi
   cmp PAR2,esi
   jae adr4r
   mov ecx,PAR2
adr4r:
   call get_data_cluster

   jmp adr2r

file_read_OK:
   popad
   ret
   
file_to_read_not_found:
   popad
   stc
   ret

file_delete:
;--------------------------------          
; INPUT  :  push <pointer to filename> PAR0
;           push <cluster init>        PAR1
;           call file_delete
; OUTPUT :  CARRY=0 OK
;           CARRY=1 ERROR
;--------------------------------
%define PAR0 [esp+8  +32]
%define PAR1 [esp+4  +32]

   pushad

   mov eax,PAR1
   mov ebx,PAR0
   call analyze_directory
   jc file_to_delete_not_found

found_delete:

   mov byte [ebx],0

   mov edx,ebx
   mov ebx,buffer
   call hd_write_new
   mov ebx,edx

   mov ax,[ebx+20]
   shl eax,16
   mov ax,[ebx+26]
   xor edx,edx

clean_new_chain:
   mov ebx,eax

   call get_FAT
   mov ecx,eax

   mov eax,ebx
   call set_FAT

   mov eax,ecx
   cmp eax,0x0fffffff
   je delete_OK

   jmp clean_new_chain

delete_OK:
   popad
   clc
   ret

file_to_delete_not_found:

   popad
   stc
   ret

;**************************************************************************
hd_read_new:

; eax = block
; ebx = pointer to memory

    push  edi    ; [esp+12] +13 +14 +15 
    push  eax    ; [esp+8]  +9 +10 +11
    push  ecx    ; [esp+4]  +5 +6 +7
    push  edx    ; [esp]    +1 +2 +3

    xor   eax,eax
    mov   edx,[hdbase]
    inc   edx
    out   dx,al
    inc   edx
    mov   al,1
    out   dx,al
    inc   edx
    mov   al,[esp+8]
    out   dx,al
    inc   edx
    mov   al,[esp+9]
    out   dx,al
    inc   edx
    mov   al,[esp+10]
    out   dx,al
    inc   edx
    mov   al,[esp+11]
    and   al,1+2+4+8
    add   al,[hdid]
    add   al,128+64+32
    out   dx,al
    inc   edx
    mov   al,20h
    out   dx,al

hdwait_read:
    in    al,dx
    and   al,128
    jnz    hdwait_read

    mov   edi,ebx
    mov   ecx,256
    mov   edx,[hdbase]

    cld
    rep insw

    pop edx
    pop ecx
    pop eax
    pop edi
    ret
;****************************************************************
hd_write_new:
; eax = block
; ebx = pointer to memory

    push  esi    ; [esp+12] +13 +14 +15
    push  eax    ; [esp+8]  +9  +10 +11
    push  ecx    ; [esp+4]  +5 +6 +7
    push  edx    ; [esp]    +1 +2 +3

    xor   eax,eax
    mov   edx,[hdbase]
    inc   edx
    out   dx,al
    inc   edx
    mov   al,1
    out   dx,al
    inc   edx
    mov   al,[esp+8]
    out   dx,al
    inc   edx
    mov   al,[esp+9]
    out   dx,al
    inc   edx
    mov   al,[esp+10]
    out   dx,al
    inc   edx
    mov   al,[esp+11]
    and   al,1+2+4+8
    add   al,[hdid]
    add   al,128+64+32
    out   dx,al
    inc   edx
    mov   al,30h
    out   dx,al

    mov   esi,ebx
    mov   ecx,256

    mov   edx,[hdbase]
    cld
    rep outsw  ;port[dx] <- word [esi]

    add edx,7
hdwait_write:
    in    al,dx
    and   al,128
    jnz    hdwait_write
    
    pop edx
    pop ecx
    pop eax
    pop esi
    ret
%endif ; CFG_FAT32

%ifdef CFG_SOUND_SB16
    
;***********************************************************
;*************   SB16-CODE-by-Paolo-Minazzi   **************
;***********************************************************


%macro out_simple 2
	mov al,%2
	out %1,al
%endm


code_SB16_load_music:

     cmp byte [SB16_Status],1
     je nol
     mov edi,SB16Buffer
     mov esi,eax
     mov ecx,65536/4
     cld
     rep movsd
nol: ret
		    

code_SB16_play_music:
;-------------------------------------
;set DMA by mixer registers
;-------------------------------------
	cmp byte [SB16_Status],1
	jne contsb16
	jmp retserve
      contsb16:

	mov dx,[sb16]
	add dx,4
	mov ax,((01<<DMA)<<8) + 81h
	out dx,ax

	mov ax,0f280h   ;enable irq5
	out dx,ax
;-------------------------------------
; set speaker on
;-------------------------------------
adr1_SB:   mov dx,[sb16]
	   add dx,0ch
	   in al,dx
	   and al,080h
	   jnz adr1_SB

	   mov al,0d1h
	   out dx,al

;-------------------------------------
; set dma
;-------------------------------------
	out_simple 0ah,DMA+4
	out_simple 0ch,0
	out_simple 0bh,048h+DMA
	
	out_simple DMA*2,0
	out_simple DMA*2,0
	
%if DMA=0
	out_simple 0x87,DMAPage
%endif

%if DMA=1
	out_simple 0x83,DMAPage
%endif

; %if DMA=2
;        out_simple 0x81,DMAPage
; %endif

%if DMA=3
	out_simple 0x82,DMAPage
%endif

	out_simple DMA*2+1,(DataLength-1) & 0ffh
	out_simple DMA*2+1,(DataLength-1)>>8

	out_simple 0ah,DMA
;-------------------------------------
; set dsp in dma mode
;-------------------------------------
	  mov dx,[sb16]
	  add dx,0ch

adr2_SB:  in al,dx                                ;wait
	  and al,080h
	  jnz adr2_SB

	  mov al,40h
	  out dx,al

adr3_SB:  in al,dx                                ;wait
	  and al,080h
	  jnz adr3_SB

	mov al,256-1000000/Rate
	out dx,al

adr4_SB:   in al,dx                                ;wait
	and al,080h
   jnz adr4_SB

	mov al,14h
	out dx,al

adr5:   in al,dx                                ;wait
	and al,080h
	jnz adr5

	mov al,(DataLength-1) & 0ffh
	out dx,al

adr6:   in al,dx                                ;wait
	and al,080h
	jnz adr6

	mov al,(DataLength-1)>>8
	out dx,al
;        mov byte [SB16_Status],1

    retserve:

     ret



;****************************************
; END CODE SB16 by Minazzi Paolo
;***************************************

%endif ; CFG_SOUND_SB16



%ifdef CFG_LEAVE_UNUSED_CODE
; system manegement mode

smm:

     cli

     mov   ax,0x0003
     int   0x10

     jmp $

%endif ; CFG_LEAVE_UNUSED_CODE




; keyboard_read


keyboard_read:

	push    ecx
	push    edx
	mov     ecx,0ffffh
      key_read_loop:
	in      al,64h
	test    al,1
	jnz     key_read_ready
	loop    key_read_loop
	mov     ah,1
	jmp     key_read_exit
      key_read_ready:
	push    ecx
	mov     ecx,32
      key_read_delay:
	loop    key_read_delay
	pop     ecx
	in      al,60h
	xor     ah,ah
      key_read_exit:
	pop     edx
	pop     ecx
	ret

; keyboard_write



keyboard_write:

	push    ecx
	push    edx
	mov     dl,al
	mov     ecx,0ffffh
      kbd_wrt_loop1:
	in      al,64h
	test    al,20h
	jz      kbd_wrt_ok1
	loop    kbd_wrt_loop1
	mov     ah,1
	jmp     kbd_wrt_exit
      kbd_wrt_ok1:
	in      al,60h
	mov     ecx,0ffffh
      kbd_wrt_loop:
	in      al,64h
	test    al,2
	jz      kbd_wrt_ok
	loop    kbd_wrt_loop
	mov     ah,1
	jmp     kbd_wrt_exit
      kbd_wrt_ok:
	mov     al,dl
	out     60h,al
	mov     ecx,0ffffh
      kbd_wrt_loop3:
	in      al,64h
	test    al,2
	jz      kbd_wrt_ok3
	loop    kbd_wrt_loop3
	mov     ah,1
	jmp     kbd_wrt_exit
      kbd_wrt_ok3:
	mov     ah,8
      kbd_wrt_loop4:
	mov     ecx,0ffffh
      kbd_wrt_loop5:
	in      al,64h
	test    al,1
	jnz     kbd_wrt_ok4
	loop    kbd_wrt_loop5
	dec     ah
	jnz     kbd_wrt_loop4
      kbd_wrt_ok4:
	xor     ah,ah
      kbd_wrt_exit:
	pop     edx
	pop     ecx
	ret

; keyboard_cmd


keyboard_cmd:

	mov     ecx,0ffffh
      cmd_wait:
	in      al,64h
	test    al,2
	jz      cmd_send
	loop    cmd_wait
	jmp     cmd_error
      cmd_send:
	mov     al,bl
	out     64h,al
	mov     ecx,0ffffh
      cmd_accept:
	in      al,64h
	test    al,2
	jz      cmd_ok
	loop    cmd_accept
      cmd_error:
	mov     ah,1
	jmp     cmd_exit
      cmd_ok:
	xor     ah,ah
      cmd_exit:
	ret




; set mouse

setmouse:

     ; set mousepicture -pointer

     mov     [0xf200],dword mousepointer

     cli

%ifdef CFG_MOUSE_PS2
     ; ps2 mouse enable

     mov     bl,0a8h                 ; enable mouse cmd
     call    keyboard_cmd
     call    keyboard_read           ; read status return

     mov     bl,20h                  ; get command byte
     call    keyboard_cmd
     call    keyboard_read
     or      al,3                    ; enable keyb. and mouse interrupt
     mov     bl,60h                  ; write back command byte
     push    eax
     call    keyboard_cmd
     pop     eax
     call    keyboard_write

     mov     bl,0d4h                 ; next command is for mouse
     call    keyboard_cmd
     mov     al,0f4h                 ; enable mouse device
     call    keyboard_write
     call    keyboard_read           ; read status return
%endif ; CFG_MOUSE_PS2

%ifdef CFG_MOUSE_COM1
     ; com1 mouse enable

     mov   bx,0x3f8 ; combase

     mov   dx,bx
     add   dx,3
     mov   al,0x80
     out   dx,al

     mov   dx,bx
     add   dx,1
     mov   al,0
     out   dx,al

     mov   dx,bx
     add   dx,0
     mov   al,0x30*2    ; 0x30 / 4
     out   dx,al

     mov   dx,bx
     add   dx,3
     mov   al,2         ; 3
     out   dx,al

     mov   dx,bx
     add   dx,4
     mov   al,0xb
     out   dx,al

     mov   dx,bx
     add   dx,1
     mov   al,1
     out   dx,al
%endif ; CFG_MOUSE_COM1

%ifdef CFG_MOUSE_COM2
     ; com2 mouse enable

     mov   bx,0x2f8 ; combase

     mov   dx,bx
     add   dx,3
     mov   al,0x80
     out   dx,al

     mov   dx,bx
     add   dx,1
     mov   al,0
     out   dx,al

     mov   dx,bx
     add   dx,0
     mov   al,0x30*2
     out   dx,al

     mov   dx,bx
     add   dx,3
     mov   al,2
     out   dx,al

     mov   dx,bx
     add   dx,4
     mov   al,0xb
     out   dx,al

     mov   dx,bx
     add   dx,1
     mov   al,1
     out   dx,al

%endif ; CFG_MOUSE_COM2

     ret


_rdtsc:

     cmp   [0xfbf3],byte 0
     jz    ret_rdtsc

     rdtsc

     ret

   ret_rdtsc:

     mov   edx,0xffffffff
     mov   eax,0xffffffff
     ret



rerouteirqs:

	cli
	in      al,0x21
	mov     ah,al
	in      al,0xA1
	mov     cx,ax

	mov     al,0x11          ; edge, cascade, icw4
	out     0x20,al
	out     0xEB,al
	out     0xA0,al
	out     0xEB,al

	mov     al,0x20
	out     0x21,al
	out     0xEB,al

	add     al,0x08
	out     0xA1,al
	out     0xEB,al

	mov     al,0x04
	out     0x21,al
	out     0xEB,al
	shr     al,1
	out     0xA1,al
	out     0xEB,al
	shr     al,1
	out     0x21,al
	out     0xEB,al
	out     0xA1,al
	out     0xEB,al

	mov     ax,cx
	out     0xA1,al
	mov     al,ah
	out     0x21,al
	mov     cx,0x1000
picl1:  out     0xEB,al
	loop    picl1

	mov     al,255              ; mask all irqs
	out     0xa1,al
	out     0x21,al

	ret


; serve apps

serveapp:


     cmp   edi,0                 ; DefineWindow
     jnz   nodrawwindow
     call  sys_drawwindow_I
     ret
   nodrawwindow:

     cmp   edi,1                 ; SetPixel
     jnz   noputpixel
     mov   edx,[0x3010]
     add   eax,[edx-0x1000]
     add   ebx,[edx-0x1000+4]
     mov   edi,0 ; no force
     call  putpixel
     ret
   noputpixel:

     cmp   edi,2                 ; GetKey
     jnz   nogetkey
     call  sys_getkey
     mov   [esp+26],eax
     ret
   nogetkey:

     cmp   edi,3                 ; GetTime
     jnz   noclock
     call  sys_clock
     mov   [esp+26],eax
     ret
   noclock:

     cmp   edi,4                 ; WriteText
     jnz   nodtext
     mov   edi,[0x3010]
     add   edi,0x10
     add   ecx,[edi]
     push  ebx
     push  ecx
     push  edx
     mov   edx,[0x3010]
     mov   ebx,[edx-0x1000]
     shl   ebx,16
     add   ebx,[edx-0x1000+4]
     add   eax,ebx
     pop   edx
     pop   ecx
     pop   ebx
     mov   edi,0
     call  dtext
     ret
   nodtext:

     cmp   edi,5                 ; DelayHs
     jnz   nodelay
     call  delay_hs
     ret
   nodelay:

     cmp   edi,6                 ; OpenRamdiskFile
     jnz   nofileread            
     mov   edi,[0x3010]
     add   edi,0x10
     add   eax,[edi]
     add   edx,[edi]
     call  fileread
     mov   [esp+26],eax
     ret
   nofileread:

     cmp   edi,7                 ; PutImage
     jnz   noputimage
     mov   edi,[0x3010]
     add   edi,0x10
     add   eax,[edi]
     mov   edx,ecx
     mov   ecx,ebx
     mov   ebx,eax
     call  sys_putimage
     mov   [esp+26],eax
     ret
   noputimage:

     cmp   edi,8                 ; DefineButton
     jnz   nobutton
     call  sys_button
     ret
   nobutton:

     cmp   edi,9                 ; GetProcessInfo
     jnz   nocpuusage
     mov   edi,[0x3010]
     add   edi,0x10
     add   eax,[edi]
     call  sys_cpuusage
     mov   [esp+26],eax
     ret
   nocpuusage:

     cmp   edi,10                ; WaitForEvent
     jnz   nowaitforevent
     call  sys_waitforevent
     mov   [esp+26],eax
     ret
   nowaitforevent:

     cmp   edi,11                ; CheckForEvent
     jnz   nogetevent
     call  sys_getevent
     mov   [esp+26],eax
     ret
   nogetevent:

     cmp   edi,12                ; BeginDraw and EndDraw
     jnz   noredrawstat
     call  sys_redrawstat
     ret
   noredrawstat:

     cmp   edi,13                ; DrawRect
     jnz   nobar
     mov   edi,ecx
     mov   cx,ax
     and   ecx,65535
     shr   eax,16
     mov   dx,bx
     and   edx,65535
     shr   ebx,16
     add   ecx,eax
     add   edx,ebx
     call  drawbar
     ret
   nobar:

     cmp   edi,14                ; GetScreenSize
     jnz   noscreensize
     mov   ax,[0xfe00]
     shl   eax,16
     mov   ax,[0xfe04]
     mov   [esp+26],eax
     ret
   noscreensize:

     cmp   edi,15                ; SetBackgroundSize, WriteBackgroundData,
     jnz   nosetbackground       ;  DrawBackground, SetBackgroundType and
     call  sys_background
     ret
   nosetbackground:

     cmp   edi,16                ; FlushFloppyCache
     jnz   nowritecachetodiskette
     call  sys_cachetodiskette
     ret
   nowritecachetodiskette:

     cmp   edi,17                ; GetButton
     jnz   nogetbutton           
     call  sys_getbutton
     mov   [esp+26],eax
     ret
   nogetbutton:

     cmp   edi,18                ; Shutdown, KillApp, WindowActivate,
     jnz   nosystem              ; GetIdle and GetTsc
     call  sys_system
     mov   [esp+26],eax
     ret
   nosystem:

     cmp   edi,19                ; StartApp
     jnz   nostart
     mov   edi,[0x3010]
     add   edi,0x10
     add   eax,[edi]
     cmp   ebx,0
     jz    noapppar
     add   ebx,[edi]
   noapppar:
     call  start_application_fl
     mov   [esp+26],eax
     ret
   nostart:

%ifdef CFG_SOUND_MIDI
     cmp   edi,20                ; ResetMidi and OutputMidi
     jnz   nomidi
     call  sys_midi
     mov   [esp+26],eax
     ret
   nomidi:
%else
	CFGSYS_UNSUPPORTED 20
%endif

				 ; SetMidiBase, SetKeymap, SetShiftKeymap,
				 ; SetKeyLayout, SetCdBase, SetSb16Base,
				 ; SetSysLanguage, SetWssBase, SetHdBase
     cmp   edi,21                ; and SetFat32Partition
     jnz   nosetup
     call  sys_setup
     mov   [esp+26],eax
     ret
   nosetup:

     cmp   edi,22                ; DefineNewWindow
     jnz   nodrawwindowII
     call  sys_drawwindow_II
     ret
   nodrawwindowII:

     cmp   edi,23                ; TimeOutWaitForEvent
     jnz   nowaitforeventwithtimeout
     call  sys_waitforeventwithtimeout
     mov   [esp+26],eax
     ret
   nowaitforeventwithtimeout:

%ifdef CFG_CDAUDIO
     cmp   edi,24                ; PlayCdTrack, StopCd and GetCdPlaylist
     jnz   nocdaudio
     call  sys_cd_audio
     mov   [esp+26],eax
     ret
   nocdaudio:
%else
	CFGSYS_UNSUPPORTED 24
%endif

%ifdef CFG_SOUND_SB16
     cmp   edi,25                ; SetSb16MixerIMainVol and SetSb16MixerICdVol
     jnz   nosb16
     call  sys_sb16
     mov   [esp+26],eax
     ret
   nosb16:
%else
	CFGSYS_UNSUPPORTED 25
%endif

				 ; GetMidiBase, GetKeymap, GetShiftKeymap,
				 ; GetKeyLayout, GetCdBase, GetSb16Base,
				 ; GetSysLanguage, GetWssBase, GetHdBase
     cmp   edi,26                ; and GetFat32Partition
     jnz   nogetsetup
     call  sys_getsetup
     mov   [esp+26],eax
     ret
   nogetsetup:


%ifdef CFG_SOUND_WSS
     cmp   edi,27                ; SetWssMainVol and SetWssCdVol
     jnz   nowss
     call  sys_wss
     mov   [esp+26],eax
     ret
   nowss:
%else
	CFGSYS_UNSUPPORTED 27
%endif

%ifdef CFG_SOUND_SB16
     cmp   edi,28                ; SetSb16MixerIIMainVol and SetSb16MixerIICdVol
     jnz   nosb16II
     call  sys_sb16II
     mov   [esp+26],eax
     ret
   nosb16II:
%else
	CFGSYS_UNSUPPORTED 28
%endif

     cmp   edi,29                ; GetDate
     jnz   nodate
     call  sys_date
     mov   [esp+26],eax
     ret
   nodate:

%ifdef CFG_HD
     cmp   edi,30                ; ReadHd
     jnz   nohdread
     mov   edi,[0x3010]
     add   edi,0x10
     add   esi,[edi]
     add   eax,[edi]
     call  read_hd_file
     mov   [esp+26],eax
     mov   [esp+22],ebx
     ret
   nohdread:
%else
	CFGSYS_UNSUPPORTED 30
%endif

%ifdef CFG_HD
     cmp   edi,31                ; StartHdApp
     jnz   nostarthd
     mov   edi,[0x3010]
     add   edi,0x10
     add   eax,[edi]
     add   ecx,[edi]
     call  start_application_hd
     mov   [esp+26],eax
     ret
   nostarthd:
%else
	CFGSYS_UNSUPPORTED 31
%endif

     cmp   edi,32                ; DelRamdiskFile
     jnz   nofiledelete          
     mov   edi,[0x3010]
     add   edi,0x10
     add   eax,[edi]
     call  filedelete
     ret
   nofiledelete:

     cmp   edi,33                ; WriteRamdiskFile
     jnz   nofilesave          
     mov   edi,[0x3010]
     add   edi,0x10
     add   eax,[edi]
     add   ebx,[edi]
     call  filesave
     ret
   nofilesave:

     cmp   edi,34                ; ReadFloppyDrive
     jnz   nofloppyread
     mov   edi,[0x3010]
     add   edi,0x10
     add   esi,[edi]
     add   eax,[edi]
     call  read_floppy_file
     mov   [esp+26],eax
     mov   [esp+22],ebx
     ret
   nofloppyread:

     cmp   edi,35                ; GetPixel
     jnz   noreadscreen
     call  read_screen_pixel
     mov   [esp+26],eax
     ret
   noreadscreen:

%ifdef CFG_READSTRING
     cmp   edi,36                ; ReadString (not yet ready
     jnz   noreadstring
     mov   edi,[0x3010]
     add   edi,0x10
     add   eax,[edi]
     call  read_string
     mov   [esp+26],eax
     ret
   noreadstring:
%else
	CFGSYS_UNSUPPORTED 36
%endif

				 ; GetMousePosition_ScreenRelative,
				 ; GetMousePosition_WindowRelative and
     cmp   edi,37                ; GetMouseStatus
     jnz   nomousepos
     call  readmousepos
     mov   [esp+26],eax
     ret
   nomousepos:

     cmp   edi,38                ; DrawLine
     jnz   nodrawline
     mov   edi,[0x3010]
     mov   dx,[edi-0x1000]
     shl   edx,16
     mov   dx,[edi-0x1000]
     add   eax,edx
     mov   dx,[edi-0x1000+4]
     shl   edx,16
     mov   dx,[edi-0x1000+4]
     add   ebx,edx
     mov   edi,0
     call  draw_line
     ret
   nodrawline:


				 ; GetBackgroundSize, ReadBackgroundData,
     cmp   edi,39                ; and GetBackgroundType
     jnz   nogetbackground
     call  sys_getbackground
     mov   [esp+26],eax
     ret
   nogetbackground:

     cmp   edi,40                ; WantEvents
     jnz   no_set_app_param
     call  set_app_param
     ret
   no_set_app_param:

     cmp   edi,41                ; GetIrqOwner
     jnz   no_get_irq_owner
     shl   eax,2
     add   eax,irq_owner
     mov   eax,[eax]
     mov   [esp+26],eax
     ret
   no_get_irq_owner:

     cmp   edi,42                 ; ReadIrqData
     jnz   no_get_irq_data
     call  get_irq_data
     mov   [esp+26],eax
     mov   [esp+22],ebx
     mov   [esp+18],ecx
     ret
   no_get_irq_data:

     cmp   edi,43                 ; SendDeviceData
     jnz   no_outport
     call  sys_outport
     mov   [esp+26],eax
     ret
   no_outport:

     cmp   edi,44                ; ProgramIrqs
     jnz   no_programirq
     push  ebx
     mov   ebx,[0x3010]
     add   ebx,0x10
     add   eax,[ebx]
     pop   ebx
     call  sys_programirq
     mov   [esp+26],ecx
     ret
   no_programirq:

     cmp   edi,45                ; ReserveIrq and FreeIrq
     jnz   no_reserveirq
     call  reserve_free_irq
     mov   [esp+26],ecx
     ret
   no_reserveirq:

     cmp   edi,46                ; ReservePortArea and FreePortArea 
     jnz   no_r_f_port_area
     call  r_f_port_area
     mov   [esp+26],eax
     ret
   no_r_f_port_area:
     
     cmp   edi,47                 ; WriteNum
     jnz   no_display_number
     call  display_number
     ret
   no_display_number:

     cmp   edi,48                 ; SetRedrawType and SetButtonType
     jnz   no_display_settings
     call  display_settings
     ret
   no_display_settings:

     cmp   edi,49                 ; AppInts
     jnz   no_app_ints
     cmp   eax,0
     jnz   no_app_ints
     mov   edi,[0x3010]
     mov   [edi+0xd000-0x3000+0x1c],ebx
     ret
   no_app_ints:

%ifdef CFG_RANDOM_WIN
     cmp   edi,50                    ; SetRandomShapeWindow and
     jne   no_random_shaped_window   ; SetRandomShapeScale
     call  random_shaped_window
     ret
   no_random_shaped_window:
%else
	CFGSYS_UNSUPPORTED 50
%endif

%ifdef CFG_SOUND_SB16
     cmp   edi,SB16_load_music       ; LoadSb16Music
     jnz   noSB16_load_music
     push  ebx
     mov   ebx,[0x3010]
     add   ebx,0x10
     add   eax,[ebx]
     pop   ebx
     call  code_SB16_load_music
     ret
   noSB16_load_music:
%else
	CFGSYS_UNSUPPORTED SB16_load_music
%endif

%ifdef CFG_SOUND_SB16
     cmp   edi,SB16_play_music       ; PlaySb16Music
     jnz   noSB16_play_music
     call  code_SB16_play_music
     ret
   noSB16_play_music:
%else
	CFGSYS_UNSUPPORTED SB16_play_music
%endif

%ifdef CFG_FAT32
     cmp   edi,SET_FAT32_DATA        ; SetFat32Data
     jnz   noSET_FAT32_DATA
     call  set_FAT32_variables
     ret
   noSET_FAT32_DATA:
%else
	CFGSYS_UNSUPPORTED SET_FAT32_DATA
%endif

%ifdef CFG_FAT32
     cmp   edi,READ_HD_FAT32_FILE     ; ReadFat32File
     jnz   noREAD_HD_FAT32_FILE  
     push  ebx
     mov   ebx,[0x3010]
     add   ebx,0x10
     add   eax,[ebx]
     add   ecx,[ebx]
     pop   ebx
     call  file_read
     ret
   noREAD_HD_FAT32_FILE:
%else
	CFGSYS_UNSUPPORTED READ_HD_FAT32_FILE
%endif

%ifdef CFG_FAT32
     cmp   edi,WRITE_HD_FAT32_FILE     ; WriteFat32File
     jnz   noWRITE_HD_FAT32_FILE  
     push  ebx
     mov   ebx,[0x3010]
     add   ebx,0x10
     add   eax,[ebx]
     add   ecx,[ebx]
     pop   ebx
     call  file_write
     ret
   noWRITE_HD_FAT32_FILE:
%else
	CFGSYS_UNSUPPORTED WRITE_HD_FAT32_FILE
%endif

     cmp   edi,-1                       ; EndApp
     jnz   noend
     call  sys_end
     ret
   noend:

     mov   eax,-1
     mov   [esp+26],eax                 ; UnknownSystemCall

     ret




; system calls


i40

      push eax
      push ebx
      push ecx
      push edx
      push esi
      push edi

      mov  ax,ds
      push ax
      mov  ax,os_data
      mov  ds,ax
      mov  es,ax
      mov  gs,ax
      mov  fs,ax

      mov  eax,[esp+2+20]

      push edi
      mov  edi,eax
      mov  eax,ebx
      mov  ebx,ecx
      mov  ecx,edx
      mov  edx,esi
      pop  esi

      call serveapp              

      pop  ax

      mov  ds,ax
      mov  es,ax
      mov  gs,ax
      mov  fs,ax

      pop  edi
      pop  esi
      pop  edx
      pop  ecx
      pop  ebx
      pop  eax

      iret

%ifdef CFG_LINUX_EMULATION
i_Linux_sys:

      cli

      pusha

      mov  ax,ds
      push ax
      mov  ax,os_data
      mov  ds,ax
      mov  es,ax
      mov  gs,ax
      mov  fs,ax

      mov  edi,[0x3010]
      mov  esi,[edi+0xd000-0x3000+0x1c]

      cmp  esi,0                     ; table not defined
      je   lil1
  
      add  esi,[edi+0x10]
      add  esi,0x80*4                ; esi <- int 0x80 swap area from mem seg 0

      mov  ebx,[esi]

      mov  eax,[esp+32+2]

      mov  [esi],eax
 
      mov  [esp+32+2],ebx


      pop  ax
      mov  ds,ax
      mov  es,ax
      mov  gs,ax
      mov  fs,ax

      popa

      iret


    lil1:        ; table not defined

      sti

      jmp   $ 
%endif ; CFG_LINUX_EMULATION


; cpu ints

s0
	push  dword 0
	jmp   errorhandle
s1
	push  dword 1
	jmp   errorhandle
s2
	push  dword 2
	jmp   errorhandle
s3
	push  dword 3
	mov   eax,3
	jmp   errorhandle
s4
	push  dword 4
	mov   eax,4
	jmp   errorhandle
s5
	push  dword 5
	jmp   errorhandle
s6
	push  dword 6
	jmp   errorhandle
s7
	push  dword 7
	jmp   errorhandle
s8
	push  dword 8
	jmp   errorhandle
s9
	push  dword 9
	jmp   errorhandle
sa
	push  dword 0xA
	jmp   errorhandle
sb
	push  dword 0xb
	jmp   errorhandle
sc
	push  dword 0xc
	jmp   errorhandle
sd
	push  dword 0xd
	jmp   errorhandle
se
	push  dword 0xe
	jmp   errorhandle
sf
	push  dword 0xf
	jmp   errorhandle
s10
	push  dword 0x10
	jmp   errorhandle
s11
	push  dword 0x11
	jmp   errorhandle


writehex:


      pusha

      push eax

      mov  ax,[printerrorat]
      shl  eax,16
      mov  ax,[esp+32+4+4]
      sub  ax,60

      pop  ecx
      mov  edx,1

      mov  edi,8
      mov  ebx,0xffffff

    whl1:

      push ecx

      and  ecx,0xf
      add  ecx,hexletters

      push edi
      mov  edi,1
      call dtext
      pop  edi

      pop  ecx

      cld
      shr  ecx,4
      sub  eax,6*65536
      dec  edi
      jnz  whl1

      popa

      ret


hexletters  db '0123456789ABCDEF'

errorhandle:

	cli

	mov   bx,os_data   ; Selector for program
	mov   ds,bx
	mov   es,bx
	mov   fs,bx
	mov   gs,bx
	mov   ss,bx

	jmp   os_code:ehl0
      ehl0:

	push  dword 110     ; eax
	call  writehex
	add   esp,4

	push  dword 120     ; ebx
	mov   eax,ebx
	call  writehex
	add   esp,4

	push  dword 130     ; ecx
	mov   eax,ecx
	call  writehex
	add   esp,4

	push  dword 140     ; edx
	mov   eax,edx
	call  writehex
	add   esp,4

	push  dword 150     ; esi
	mov   eax,esi
	call  writehex
	add   esp,4

	push  dword 160     ; edi
	mov   eax,edi
	call  writehex
	add   esp,4

	push  dword 180     ; int no
	call  writehex
	add   esp,4

	push  dword 190     ; process no
	mov   eax,[0x3000]
	call  writehex
	add   esp,4

	; stack dump

	mov   edx,210

      newstdu:

	pop   eax
	push  edx
	call  writehex
	pop   eax

	add   edx,10
	cmp   edx,410
	jnz   newstdu

	mov   eax,[printerrorat]
	add   eax,60
	cmp   eax,620
	jbe   ehl3
	mov   eax,300
      ehl3:
	mov   [printerrorat],eax

	cmp   [0x3004],dword 1
	jg    conterrorhandle

	cli
	jmp   $

      conterrorhandle:

	sti

	mov   edx,[0x3010]      ; mark process as zombie
	mov   [edx+0xa],byte 3  ; terminate this program

      ehl:                      ; wait for termination
	call  change_task
	jmp   ehl


printerrorat dd 300

; update cpu times

updatecputimes:

	pusha

	call _rdtsc

	mov  eax,[idleuse]       ; idle / sec
	mov  [idleusesec],eax
	mov  [idleuse],dword 0

	mov  ecx,[0x3004]
	mov  edi,0x3020

      newupdate:

	mov  ebx,[edi+0x14]
	mov  [edi+0x1c],ebx
	mov  [edi+0x14],dword 0

	add  edi,0x20

	loop newupdate

	popa

	ret


tscmem  dd 0x0


; reprogrammed irqs


irq0:   ; scheduler

	cli

	pusha

	mov   ax,ds
	push  ax

	mov   ax,os_data
	mov   ds,ax

	add   [0xfdf0],dword 1          ; increase hs counter

	xor   edx,edx                   ; update cpu times ?
	mov   eax,[0xfdf0]
	mov   ebx,100
	div   ebx
	cmp   edx,0
	jnz   nocounter
	call  updatecputimes
      nocounter:


	cmp   [0xffff],byte 0           ; yes task change ?
	jz    yestaskchange

	mov   [0xffff],byte 0 

	pop   ax
	mov   ds,ax

	mov   al,0x20                   ; ready for next irq
	mov   dx,0x20
	out   dx,al

	popa

	iret

      yestaskchange:

	mov   edx,[0x3010]

	push  edx                       ; calculate cpu usage
	mov   ebx,[edx+0x18]
	call  _rdtsc
	sub   eax,ebx
	pop   edx
	add   eax,[edx+0x14]
	mov   [edx+0x14],eax

	add   edx,0x20
	mov   ebx,[0x3000]              ; increase process counter
	inc   ebx
	mov   eax,[0x3004]
	inc   eax
	cmp   ebx,eax
	jnz   nsched0
	mov   ebx,1
	mov   edx,0x3020
      nsched0:
	mov   [0x3000],ebx
	mov   [0x3010],edx              ; process mem base in edx

	push  edx                       ; save time stamp count
	call  _rdtsc
	pop   edx
	mov   [edx+0x18],eax

	mov   ax,tss0
	shl   bx,3
	add   ax,bx
	mov   [tss_s],ax

	mov   al,0x20                   ; ready for next irq
	mov   dx,0x20
	out   dx,al

	pop   ax
	mov   ds,ax

	popa
	sti

	cmp   [0x3004],dword 1
	je    notaskchange

	db    0xea                      ; Jump Instruction
tss_t   dd    0                         ; Target Address
tss_s   dw    tss0                      ; Segment Selector

      notaskchange:

	iret



change_task:

	cli

	pusha

	mov   ax,ds
	push  ax

	mov   ax,os_data
	mov   ds,ax

	mov   [0xffff],byte 1           ; timer no task change

	mov   edx,[0x3010]

	push  edx                       ; calculate cpu usage
	mov   ebx,[edx+0x18]
	call  _rdtsc
	sub   eax,ebx
	pop   edx
	add   eax,[edx+0x14]
	mov   [edx+0x14],eax

	add   edx,0x20
	mov   ebx,[0x3000]              ; increase process counter
	inc   ebx
	mov   eax,[0x3004]
	inc   eax
	cmp   ebx,eax
	jnz   n2sched0
	mov   ebx,1
	mov   edx,0x3020
      n2sched0:
	mov   [0x3000],ebx
	mov   [0x3010],edx              ; process mem base in edx

	push  edx                       ; save time stamp count
	call  _rdtsc
	pop   edx
	mov   [edx+0x18],eax

	mov   ax,tss0
	shl   bx,3
	add   ax,bx
	mov   [tss_s2],ax

	pop   ax
	mov   ds,ax

	popa
	sti

	cmp   [0x3004],dword 1
	je    notaskchange2

	db    0xea                      ; Jump Instruction
tss_t2  dd    0                         ; Target Address
tss_s2  dw    tss0                      ; Segment Selector

      notaskchange2:

	ret


irq1 ; keyboard

	cli

	push  eax
	push  ebx
	push  ecx
	push  edx

	mov    cx,ds
	mov    ax,os_data
	mov    ds,ax
	mov    es,ax
	mov    fs,ax
	mov    gs,ax

	cld

	push   cx

	xor    eax,eax
	mov    dx,0x60
	in     al,dx

	cmp   al,29             ; CTRL
	jne   noctrlon
	mov   [ctrl],byte 1
	jmp   nokey
      noctrlon:

	cmp   al,29+128 
	jne   noctrloff
	mov   [ctrl],byte 0
	jmp   nokey
      noctrloff:

	cmp   al,56             ; ALT
	jnz   noalton1
	mov   [alt],byte 1
	jmp   nokey
      noalton1:

	cmp   al,184
	jnz   noaltoff1
	mov   [alt],byte 0
	jmp   nokey
      noaltoff1:

	cmp   al,42             ; SHIFT
	jnz   noshifton1
	mov   [shift],byte 1
	jmp   nokey
      noshifton1:
	cmp   al,54
	jnz   noshifton2
	mov   [shift],byte 1
	jmp   nokey
      noshifton2:

	cmp   al,42+128
	jnz   noshiftoff1
	mov   [shift],byte 0
	jmp   nokey
      noshiftoff1:
	cmp   al,54+128
	jnz   noshiftoff2
	mov   [shift],byte 0
	jmp   nokey
      noshiftoff2:

	cmp   al,byte 128
	jb    yeskey

	jmp   nokey

      yeskey:

	mov   ecx,eax              ; plain key
	add   ecx,keymap
	mov   bl,[ecx]

	cmp   [ctrl],byte 1        ; ctrl on ?
	jne   nco
	sub   bl,96
      nco:

	cmp   [shift],byte 1       ; shift on ?
	jnz   nso
	mov   ecx,eax
	add   ecx,keymap_shift
	mov   bl,[ecx]
      nso:

	cmp   [alt],byte 1         ; alt on ?
	jz    yao
	jmp   nao
      yao:
	mov   ecx,eax
	add   ecx,keymap_alt
	mov   bl,[ecx]

	; alt mouse ?

	xor   edx,edx
	cmp   bl,178
	jnz   noaltup
	mov   edx,5*65536
	mov   bl,255
      noaltup:
	cmp   bl,177
	jnz   noaltdown
	mov   edx,251*65536
	mov   bl,255
      noaltdown:
	cmp   bl,176
	jnz   noaltleft
	mov   edx,251*256
	mov   bl,255
      noaltleft:
	cmp   bl,179
	jnz   noaltright
	mov   edx,5*256
	mov   bl,255
      noaltright:
	cmp   bl,' '
	jnz   noaltmbd
	mov   bl,[altmouseb]
	add   bl,1
	and   bl,1
	mov   [altmouseb],bl
	mov   bl,255
      noaltmbd:
	cmp   bl,255
	jnz   nao
	mov   [0xF604],byte 1 ; ps2 data
	mov   [0xfb00],byte 0 ; ps2 chunk count
	mov   [0x2E0000+4096*12],word 3 ; mouse data count
	movzx ebx,byte [altmouseb]
	or    edx,ebx
	mov   [0x2E0000+4096*12+0x10],edx
	mov   bl,0
      nao:

	cmp   bl,0
	jz    nokey
	mov   [0xf401],bl
	mov   [0xf400],byte 1

      nokey:

	mov   al,0x20                   ; ready for next irq
	mov   dx,0x20
	out   dx,al

	pop   cx
	mov   ds,cx
	mov   es,cx
	mov   fs,cx
	mov   gs,cx

	pop   edx
	pop   ecx
	pop   ebx
	pop   eax

	iret



shift   db 0x0
ctrl    db 0x0
alt     db 0x0
altgr   db 0x0

altmouseb  db  0x0


sys_panic  db   'TIMES      '

%ifdef CFG_LEAVE_UNUSED_CODE
%ifdef CFG_SOUND_SB16

irq5 ; SB16

	cli

	pusha

	mov    cx,ds

	mov    ax,os_data
	mov    ds,ax
	mov    es,ax
	mov    fs,ax
	mov    gs,ax

	mov    dx,[sb16]
	add    dx,0xe
	in     al,dx
	mov    byte [SB16_Status],0

	mov    al,0x20
	out    0x20,al

	mov    ds,cx
	mov    es,cx
	mov    fs,cx
	mov    gs,cx

	popa

	iret

%endif ; CFG_SOUND_SB16
%endif ; CFG_LEAVE_UNUSED_CODE



irqD:

     cli

     pusha

     mov   cx,ds
     push  cx

     mov   ax,os_data
     mov   ds,ax
     mov   es,ax
     mov   fs,ax
     mov   gs,ax

     ; $F0 to $00

     mov   dx,0xf0
     mov   al,0
     out   dx,al

     mov   dx,0xa0
     mov   al,0x20
     out   dx,al

     mov   dx,0x20
     out   dx,al

     pop   cx
     mov   ds,cx
     mov   es,cx
     mov   fs,cx
     mov   gs,cx

     popa

     iret


p_irq2:

     cli                       ; 6-9
     pusha                     ; 5
     mov    edi,2              ; 1
     call   irqhandler         ; 2/5
     popa                      ; 5-9
     iret                      ; 5/8

p_irq3:

     cli
     pusha
     mov    edi,3
     call   irqhandler
     popa
     iret


p_irq4:

     cli
     pusha
     mov    edi,4
     call   irqhandler
     popa
     iret

p_irq5:

     cli
     pusha
     mov    edi,5
     call   irqhandler
     popa
     iret

p_irq6:

     cli
     pusha
     mov    edi,6
     call   irqhandler
     popa
     iret

p_irq7:

     cli
     pusha
     mov    edi,7
     call   irqhandler
     popa
     iret

p_irq8:

     cli
     pusha
     mov    edi,8
     call   irqhandler
     popa
     iret

p_irq9:

     cli
     pusha
     mov    edi,9
     call   irqhandler
     popa
     iret

p_irq10:

     cli
     pusha
     mov    edi,10
     call   irqhandler
     popa
     iret

p_irq11:

     cli
     pusha
     mov    edi,11
     call   irqhandler
     popa
     iret

p_irq12:

     cli
     pusha
     mov    edi,12
     call   irqhandler
     popa
     iret

p_irq13:

     cli
     pusha
     mov    edi,13
     call   irqhandler
     popa
     iret

p_irq14:

     cli
     pusha
     mov    edi,14
     call   irqhandler
     popa
     iret

p_irq15:

     cli
     pusha
     mov    edi,15
     call   irqhandler
     popa
     iret





irqhandler:

     push   edi

     mov    cx,ds            ; 1
     mov    ax,os_data       ; 1
     mov    ds,ax            ; 1  
     mov    es,ax            ; 1
     mov    fs,ax            ; 1 
     mov    gs,ax            ; 1

     cld                     ; 2

     mov    esi,edi          ; 1
     shl    esi,6            ; 1
     add    esi,irq00read    ; 1

     shl    edi,12           ; 1
     add    edi,0x2E0000

   irqnewread:

     mov    dx,[esi]         ; 2+ 

     cmp    dx,0             ; 1
     jz     irqover          

     cmp    [esi+3],byte 1   ; 2     ; byte read
     jne    noirqbyte        ; 4-11  

     in     al,dx

     mov    edx,[edi]
     cmp    edx,4000
     je     irqfull
     mov    ebx,edi
     add    ebx,0x10
     add    ebx,edx
     mov    [ebx],al
     add    edx,1
     mov    [edi],edx

     add    esi,4
     jmp    irqnewread

   noirqbyte:


     cmp    [esi+3],byte 2     ; word read
     jne    noirqword

     in     ax,dx

     mov    edx,[edi]
     cmp    edx,4000
     je     irqfull
     mov    ebx,edi
     add    ebx,0x10
     add    ebx,edx
     mov    [ebx],al
     add    edx,2
     mov    [edi],edx

     add    esi,4
     jmp    irqnewread

   noirqword:



   irqfull:
   irqover:

     mov    ds,cx
     mov    es,cx
     mov    fs,cx
     mov    gs,cx

     mov    al,0x20            ; ready for next irq
     out    0x20,al

     pop    ebx
     cmp    ebx,7
     jbe    noa0
     out    0xa0,al
   noa0:

     ret




i_unknown10
	push  dword $
	jmp   i_u_errorhandle
i_unknown11
	push  dword $
	jmp   i_u_errorhandle
i_unknown12
	push  dword $
	jmp   i_u_errorhandle
i_unknown13
	push  dword $
	jmp   i_u_errorhandle
i_unknown14
	push  dword $
	jmp   i_u_errorhandle
i_unknown15
	push  dword $
	jmp   i_u_errorhandle
i_unknown16
	push  dword $
	jmp   i_u_errorhandle
i_unknown17
	push  dword $
	jmp   i_u_errorhandle
i_unknown18
	push  dword $
	jmp   i_u_errorhandle
i_unknown19
	push  dword $
	jmp   i_u_errorhandle
i_unknown1a
	push  dword $
	jmp   i_u_errorhandle
i_unknown1b
	push  dword $
	jmp   i_u_errorhandle
i_unknown1c
	push  dword $
	jmp   i_u_errorhandle
i_unknown1d
	push  dword $
	jmp   i_u_errorhandle
i_unknown1e
	push  dword $
	jmp   i_u_errorhandle
i_unknown1f
	push  dword $
	jmp   i_u_errorhandle


i_unknown20
	push  dword $
	jmp   i_u_errorhandle
i_unknown21
	push  dword $
	jmp   i_u_errorhandle
i_unknown22
	push  dword $
	jmp   i_u_errorhandle
i_unknown23
	push  dword $
	jmp   i_u_errorhandle
i_unknown24
	push  dword $
	jmp   i_u_errorhandle
i_unknown25
	push  dword $
	jmp   i_u_errorhandle
i_unknown26
	push  dword $
	jmp   i_u_errorhandle
i_unknown27
	push  dword $
	jmp   i_u_errorhandle
i_unknown28
	push  dword $
	jmp   i_u_errorhandle
i_unknown29
	push  dword $
	jmp   i_u_errorhandle
i_unknown2a
	push  dword $
	jmp   i_u_errorhandle
i_unknown2b
	push  dword $
	jmp   i_u_errorhandle
i_unknown2c
	push  dword $
	jmp   i_u_errorhandle
i_unknown2d
	push  dword $
	jmp   i_u_errorhandle
i_unknown2e
	push  dword $
	jmp   i_u_errorhandle
i_unknown2f
	push  dword $
	jmp   i_u_errorhandle


i_unknown30
	push  dword $
	jmp   i_u_errorhandle
i_unknown31
	push  dword $
	jmp   i_u_errorhandle
i_unknown32
	push  dword $
	jmp   i_u_errorhandle
i_unknown33
	push  dword $
	jmp   i_u_errorhandle
i_unknown34
	push  dword $
	jmp   i_u_errorhandle
i_unknown35
	push  dword $
	jmp   i_u_errorhandle
i_unknown36
	push  dword $
	jmp   i_u_errorhandle
i_unknown37
	push  dword $
	jmp   i_u_errorhandle
i_unknown38
	push  dword $
	jmp   i_u_errorhandle
i_unknown39
	push  dword $
	jmp   i_u_errorhandle
i_unknown3a
	push  dword $
	jmp   i_u_errorhandle
i_unknown3b
	push  dword $
	jmp   i_u_errorhandle
i_unknown3c
	push  dword $
	jmp   i_u_errorhandle
i_unknown3d
	push  dword $
	jmp   i_u_errorhandle
i_unknown3e
	push  dword $
	jmp   i_u_errorhandle
i_unknown3f
	push  dword $
	jmp   i_u_errorhandle


i_unknown40
	push  dword $
	jmp   i_u_errorhandle
i_unknown41
	push  dword $
	jmp   i_u_errorhandle
i_unknown42
	push  dword $
	jmp   i_u_errorhandle
i_unknown43
	push  dword $
	jmp   i_u_errorhandle
i_unknown44
	push  dword $
	jmp   i_u_errorhandle
i_unknown45
	push  dword $
	jmp   i_u_errorhandle
i_unknown46
	push  dword $
	jmp   i_u_errorhandle
i_unknown47
	push  dword $
	jmp   i_u_errorhandle
i_unknown48
	push  dword $
	jmp   i_u_errorhandle
i_unknown49
	push  dword $
	jmp   i_u_errorhandle
i_unknown4a
	push  dword $
	jmp   i_u_errorhandle
i_unknown4b
	push  dword $
	jmp   i_u_errorhandle
i_unknown4c
	push  dword $
	jmp   i_u_errorhandle
i_unknown4d
	push  dword $
	jmp   i_u_errorhandle
i_unknown4e
	push  dword $
	jmp   i_u_errorhandle
i_unknown4f
	push  dword $
	jmp   i_u_errorhandle



i_unknown50
	push  dword $
	jmp   i_u_errorhandle
i_unknown51
	push  dword $
	jmp   i_u_errorhandle
i_unknown52
	push  dword $
	jmp   i_u_errorhandle
i_unknown53
	push  dword $
	jmp   i_u_errorhandle
i_unknown54
	push  dword $
	jmp   i_u_errorhandle
i_unknown55
	push  dword $
	jmp   i_u_errorhandle
i_unknown56
	push  dword $
	jmp   i_u_errorhandle
i_unknown57
	push  dword $
	jmp   i_u_errorhandle
i_unknown58
	push  dword $
	jmp   i_u_errorhandle
i_unknown59
	push  dword $
	jmp   i_u_errorhandle
i_unknown5a
	push  dword $
	jmp   i_u_errorhandle
i_unknown5b
	push  dword $
	jmp   i_u_errorhandle
i_unknown5c
	push  dword $
	jmp   i_u_errorhandle
i_unknown5d
	push  dword $
	jmp   i_u_errorhandle
i_unknown5e
	push  dword $
	jmp   i_u_errorhandle
i_unknown5f
	push  dword $
	jmp   i_u_errorhandle


i_unknown60
	push  dword $
	jmp   i_u_errorhandle
i_unknown61
	push  dword $
	jmp   i_u_errorhandle
i_unknown62
	push  dword $
	jmp   i_u_errorhandle
i_unknown63
	push  dword $
	jmp   i_u_errorhandle
i_unknown64
	push  dword $
	jmp   i_u_errorhandle
i_unknown65
	push  dword $
	jmp   i_u_errorhandle
i_unknown66
	push  dword $
	jmp   i_u_errorhandle
i_unknown67
	push  dword $
	jmp   i_u_errorhandle
i_unknown68
	push  dword $
	jmp   i_u_errorhandle
i_unknown69
	push  dword $
	jmp   i_u_errorhandle
i_unknown6a
	push  dword $
	jmp   i_u_errorhandle
i_unknown6b
	push  dword $
	jmp   i_u_errorhandle
i_unknown6c
	push  dword $
	jmp   i_u_errorhandle
i_unknown6d
	push  dword $
	jmp   i_u_errorhandle
i_unknown6e
	push  dword $
	jmp   i_u_errorhandle
i_unknown6f
	push  dword $
	jmp   i_u_errorhandle


i_unknown70
	push  dword $
	jmp   i_u_errorhandle
i_unknown71
	push  dword $
	jmp   i_u_errorhandle
i_unknown72
	push  dword $
	jmp   i_u_errorhandle
i_unknown73
	push  dword $
	jmp   i_u_errorhandle
i_unknown74
	push  dword $
	jmp   i_u_errorhandle
i_unknown75
	push  dword $
	jmp   i_u_errorhandle
i_unknown76
	push  dword $
	jmp   i_u_errorhandle
i_unknown77
	push  dword $
	jmp   i_u_errorhandle
i_unknown78
	push  dword $
	jmp   i_u_errorhandle
i_unknown79
	push  dword $
	jmp   i_u_errorhandle
i_unknown7a
	push  dword $
	jmp   i_u_errorhandle
i_unknown7b
	push  dword $
	jmp   i_u_errorhandle
i_unknown7c
	push  dword $
	jmp   i_u_errorhandle
i_unknown7d
	push  dword $
	jmp   i_u_errorhandle
i_unknown7e
	push  dword $
	jmp   i_u_errorhandle
i_unknown7f
	push  dword $
	jmp   i_u_errorhandle

; note original kernel included this even though it was unused.
%ifdef CFG_LEAVE_UNUSED_CODE CFGSYS_NO_LINUX_EMULATION
i_unknown80
	push  dword $
	jmp   i_u_errorhandle
%endif ; CFG_LEAVE_UNUSED_CODE or CFGSYS_NO_LINUX_EMULATION

i_unknown81
	push  dword $
	jmp   i_u_errorhandle
i_unknown82
	push  dword $

	jmp   i_u_errorhandle
i_unknown83
	push  dword $
	jmp   i_u_errorhandle
i_unknown84
	push  dword $
	jmp   i_u_errorhandle
i_unknown85
	push  dword $
	jmp   i_u_errorhandle
i_unknown86
	push  dword $
	jmp   i_u_errorhandle
i_unknown87
	push  dword $
	jmp   i_u_errorhandle
i_unknown88
	push  dword $
	jmp   i_u_errorhandle
i_unknown89
	push  dword $
	jmp   i_u_errorhandle
i_unknown8a
	push  dword $
	jmp   i_u_errorhandle
i_unknown8b
	push  dword $
	jmp   i_u_errorhandle
i_unknown8c
	push  dword $
	jmp   i_u_errorhandle
i_unknown8d
	push  dword $
	jmp   i_u_errorhandle
i_unknown8e
	push  dword $
	jmp   i_u_errorhandle
i_unknown8f
	push  dword $
	jmp   i_u_errorhandle


i_unknown90
	push  dword $
	jmp   i_u_errorhandle
i_unknown91
	push  dword $
	jmp   i_u_errorhandle
i_unknown92
	push  dword $
	jmp   i_u_errorhandle
i_unknown93
	push  dword $
	jmp   i_u_errorhandle
i_unknown94
	push  dword $
	jmp   i_u_errorhandle
i_unknown95
	push  dword $
	jmp   i_u_errorhandle
i_unknown96
	push  dword $
	jmp   i_u_errorhandle
i_unknown97
	push  dword $
	jmp   i_u_errorhandle
i_unknown98
	push  dword $
	jmp   i_u_errorhandle
i_unknown99
	push  dword $
	jmp   i_u_errorhandle
i_unknown9a
	push  dword $
	jmp   i_u_errorhandle
i_unknown9b
	push  dword $
	jmp   i_u_errorhandle
i_unknown9c
	push  dword $
	jmp   i_u_errorhandle
i_unknown9d
	push  dword $
	jmp   i_u_errorhandle
i_unknown9e
	push  dword $
	jmp   i_u_errorhandle
i_unknown9f
	push  dword $
	jmp   i_u_errorhandle


i_unknowna0
	push  dword $
	jmp   i_u_errorhandle
i_unknowna1
	push  dword $
	jmp   i_u_errorhandle
i_unknowna2
	push  dword $
	jmp   i_u_errorhandle
i_unknowna3
	push  dword $
	jmp   i_u_errorhandle
i_unknowna4
	push  dword $
	jmp   i_u_errorhandle
i_unknowna5
	push  dword $
	jmp   i_u_errorhandle
i_unknowna6
	push  dword $
	jmp   i_u_errorhandle
i_unknowna7
	push  dword $
	jmp   i_u_errorhandle
i_unknowna8
	push  dword $
	jmp   i_u_errorhandle
i_unknowna9
	push  dword $
	jmp   i_u_errorhandle
i_unknownaa
	push  dword $
	jmp   i_u_errorhandle
i_unknownab
	push  dword $
	jmp   i_u_errorhandle
i_unknownac
	push  dword $
	jmp   i_u_errorhandle
i_unknownad
	push  dword $
	jmp   i_u_errorhandle
i_unknownae
	push  dword $
	jmp   i_u_errorhandle
i_unknownaf
	push  dword $
	jmp   i_u_errorhandle


i_unknownb0
	push  dword $
	jmp   i_u_errorhandle
i_unknownb1
	push  dword $
	jmp   i_u_errorhandle
i_unknownb2
	push  dword $
	jmp   i_u_errorhandle
i_unknownb3
	push  dword $
	jmp   i_u_errorhandle
i_unknownb4
	push  dword $
	jmp   i_u_errorhandle
i_unknownb5
	push  dword $
	jmp   i_u_errorhandle
i_unknownb6
	push  dword $
	jmp   i_u_errorhandle
i_unknownb7
	push  dword $
	jmp   i_u_errorhandle
i_unknownb8
	push  dword $
	jmp   i_u_errorhandle
i_unknownb9
	push  dword $
	jmp   i_u_errorhandle
i_unknownba
	push  dword $
	jmp   i_u_errorhandle
i_unknownbb
	push  dword $
	jmp   i_u_errorhandle
i_unknownbc
	push  dword $
	jmp   i_u_errorhandle
i_unknownbd
	push  dword $
	jmp   i_u_errorhandle
i_unknownbe
	push  dword $
	jmp   i_u_errorhandle
i_unknownbf
	push  dword $
	jmp   i_u_errorhandle


i_unknownc0
	push  dword $
	jmp   i_u_errorhandle
i_unknownc1
	push  dword $
	jmp   i_u_errorhandle
i_unknownc2
	push  dword $
	jmp   i_u_errorhandle
i_unknownc3
	push  dword $
	jmp   i_u_errorhandle
i_unknownc4
	push  dword $
	jmp   i_u_errorhandle
i_unknownc5
	push  dword $
	jmp   i_u_errorhandle
i_unknownc6
	push  dword $
	jmp   i_u_errorhandle
i_unknownc7
	push  dword $
	jmp   i_u_errorhandle
i_unknownc8
	push  dword $
	jmp   i_u_errorhandle
i_unknownc9
	push  dword $
	jmp   i_u_errorhandle
i_unknownca
	push  dword $
	jmp   i_u_errorhandle
i_unknowncb
	push  dword $
	jmp   i_u_errorhandle
i_unknowncc
	push  dword $
	jmp   i_u_errorhandle
i_unknowncd
	push  dword $
	jmp   i_u_errorhandle
i_unknownce
	push  dword $
	jmp   i_u_errorhandle
i_unknowncf
	push  dword $
	jmp   i_u_errorhandle


i_unknownd0
	push  dword $
	jmp   i_u_errorhandle
i_unknownd1
	push  dword $
	jmp   i_u_errorhandle
i_unknownd2
	push  dword $
	jmp   i_u_errorhandle
i_unknownd3
	push  dword $
	jmp   i_u_errorhandle
i_unknownd4
	push  dword $
	jmp   i_u_errorhandle
i_unknownd5
	push  dword $
	jmp   i_u_errorhandle
i_unknownd6
	push  dword $
	jmp   i_u_errorhandle
i_unknownd7
	push  dword $
	jmp   i_u_errorhandle
i_unknownd8
	push  dword $
	jmp   i_u_errorhandle
i_unknownd9
	push  dword $
	jmp   i_u_errorhandle
i_unknownda
	push  dword $
	jmp   i_u_errorhandle
i_unknowndb
	push  dword $
	jmp   i_u_errorhandle
i_unknowndc
	push  dword $
	jmp   i_u_errorhandle
i_unknowndd
	push  dword $
	jmp   i_u_errorhandle
i_unknownde
	push  dword $
	jmp   i_u_errorhandle
i_unknowndf
	push  dword $
	jmp   i_u_errorhandle


i_unknowne0
	push  dword $
	jmp   i_u_errorhandle
i_unknowne1
	push  dword $
	jmp   i_u_errorhandle
i_unknowne2
	push  dword $
	jmp   i_u_errorhandle
i_unknowne3
	push  dword $
	jmp   i_u_errorhandle
i_unknowne4
	push  dword $
	jmp   i_u_errorhandle
i_unknowne5
	push  dword $
	jmp   i_u_errorhandle
i_unknowne6
	push  dword $
	jmp   i_u_errorhandle
i_unknowne7
	push  dword $
	jmp   i_u_errorhandle
i_unknowne8
	push  dword $
	jmp   i_u_errorhandle
i_unknowne9
	push  dword $
	jmp   i_u_errorhandle
i_unknownea
	push  dword $
	jmp   i_u_errorhandle
i_unknowneb
	push  dword $
	jmp   i_u_errorhandle
i_unknownec
	push  dword $
	jmp   i_u_errorhandle
i_unknowned
	push  dword $
	jmp   i_u_errorhandle
i_unknownee
	push  dword $
	jmp   i_u_errorhandle
i_unknownef
	push  dword $
	jmp   i_u_errorhandle



i_unknownf0
	push  dword $
	jmp   i_u_errorhandle
i_unknownf1
	push  dword $
	jmp   i_u_errorhandle
i_unknownf2
	push  dword $
	jmp   i_u_errorhandle
i_unknownf3
	push  dword $
	jmp   i_u_errorhandle
i_unknownf4
	push  dword $
	jmp   i_u_errorhandle
i_unknownf5
	push  dword $
	jmp   i_u_errorhandle
i_unknownf6
	push  dword $
	jmp   i_u_errorhandle
i_unknownf7
	push  dword $
	jmp   i_u_errorhandle
i_unknownf8
	push  dword $
	jmp   i_u_errorhandle
i_unknownf9
	push  dword $
	jmp   i_u_errorhandle
i_unknownfa
	push  dword $
	jmp   i_u_errorhandle
i_unknownfb
	push  dword $
	jmp   i_u_errorhandle
i_unknownfc
	push  dword $
	jmp   i_u_errorhandle
i_unknownfd
	push  dword $
	jmp   i_u_errorhandle
i_unknownfe
	push  dword $
	jmp   i_u_errorhandle
i_unknownff
	push  dword $
	jmp   i_u_errorhandle



i_u_errorhandle:


	cli

	pusha

	mov  eax,[printerrorat]      ; bad interrupt
	sub  eax,7*6
	shl  eax,16
	mov  ax,90
	mov  ebx,0xff0000
	mov  ecx,bad_int
	mov  edx,8
	mov  edi,1
	call dtext

	popa

	mov   [iu_eaxtemp],eax
	mov   [iu_ebxtemp],ebx
	mov   [iu_ecxtemp],ecx
	mov   [iu_edxtemp],edx

	pop   eax
	sub   eax,i_unknown10

	xor   edx,edx
	mov   ebx,10
	div   ebx

	add   eax,0x10

	push  eax

	mov   eax,[iu_eaxtemp]
	mov   ebx,[iu_ebxtemp]
	mov   ecx,[iu_ecxtemp]
	mov   edx,[iu_edxtemp]

	jmp   errorhandle

iu_eaxtemp
	dd    0x0
iu_ebxtemp
	dd    0x0
iu_ecxtemp
	dd    0x0
iu_edxtemp
	dd    0x0


bad_int db 'BAD INT.'



;    descriptors


gdts:
  
	dw  gdte-$-1
	dd  gdts
	dw  0

os_code        equ  $-gdts

	dw     0xffff                    ;  limit 0:15
	dw     0x0000                    ;  base  0:15
	db     0x00                      ;  base  23:16
	dw     11010000b *256 +10011010b 
	db     0x00                      ;  base  31:24

os_data        equ  $-gdts

	dw     0xffff                    ;  limit 0:15
	dw     0x0000                    ;  base  0:15
	db     0x00                      ;  base  23:16
	dw     11010000b *256 +10010010b 
	db     0x00                      ;  base  31:24


; for jmp in scheduler 

tss0       equ  $-gdts

	   times  64 dd 0,0           


; for application code

app_code   equ    $-gdts

	   times  64 dd 0,0           


; for application data

app_data   equ    $-gdts

	   times  64 dd 0,0   

gdte:



	
idts:

	dw  idte-$-1
	dd  idts+8
	dw  0

	times 256 dd 0,0


	idte:



sys_int:

    dd   s0,s1,s2,s3,s4,s5,s6,s7,s8,s9,sa,sb,sc,sd,se,sf

    dd   s10        ,s11        ,i_unknown12,i_unknown13,i_unknown14,i_unknown15,i_unknown16,i_unknown17
    dd   i_unknown18,i_unknown19,i_unknown1a,i_unknown1b,i_unknown1c,i_unknown1d,i_unknown1e,i_unknown1f

    dd   irq0  ,irq1  ,p_irq2 ,p_irq3 ,p_irq4 ,p_irq5,p_irq6 ,p_irq7
    dd   p_irq8,p_irq9,p_irq10,p_irq11,p_irq12,irqD  ,p_irq14,p_irq15


    dd   i_unknown30,i_unknown31,i_unknown32,i_unknown33,i_unknown34,i_unknown35,i_unknown36,i_unknown37
    dd   i_unknown38,i_unknown39,i_unknown3a,i_unknown3b,i_unknown3c,i_unknown3d,i_unknown3e,i_unknown3f

    dd   i40        ,i_unknown41,i_unknown42,i_unknown43,i_unknown44,i_unknown45,i_unknown46,i_unknown47
    dd   i_unknown48,i_unknown49,i_unknown4a,i_unknown4b,i_unknown4c,i_unknown4d,i_unknown4e,i_unknown4f

    dd   i_unknown50,i_unknown51,i_unknown52,i_unknown53,i_unknown54,i_unknown55,i_unknown56,i_unknown57
    dd   i_unknown58,i_unknown59,i_unknown5a,i_unknown5b,i_unknown5c,i_unknown5d,i_unknown5e,i_unknown5f

    dd   i_unknown60,i_unknown61,i_unknown62,i_unknown63,i_unknown64,i_unknown65,i_unknown66,i_unknown67
    dd   i_unknown68,i_unknown69,i_unknown6a,i_unknown6b,i_unknown6c,i_unknown6d,i_unknown6e,i_unknown6f

    dd   i_unknown70,i_unknown71,i_unknown72,i_unknown73,i_unknown74,i_unknown75,i_unknown76,i_unknown77
    dd   i_unknown78,i_unknown79,i_unknown7a,i_unknown7b,i_unknown7c,i_unknown7d,i_unknown7e,i_unknown7f

%ifdef CFG_LINUX_EMULATION
    dd   i_Linux_sys
%else
    dd   i_unknown80
%endif

    dd      i_unknown81,i_unknown82,i_unknown83,i_unknown84,i_unknown85,i_unknown86,i_unknown87

    dd   i_unknown88,i_unknown89,i_unknown8a,i_unknown8b,i_unknown8c,i_unknown8d,i_unknown8e,i_unknown8f

    dd   i_unknown90,i_unknown91,i_unknown92,i_unknown93,i_unknown94,i_unknown95,i_unknown96,i_unknown97
    dd   i_unknown98,i_unknown99,i_unknown9a,i_unknown9b,i_unknown9c,i_unknown9d,i_unknown9e,i_unknown9f

    dd   i_unknowna0,i_unknowna1,i_unknowna2,i_unknowna3,i_unknowna4,i_unknowna5,i_unknowna6,i_unknowna7
    dd   i_unknowna8,i_unknowna9,i_unknownaa,i_unknownab,i_unknownac,i_unknownad,i_unknownae,i_unknownaf

    dd   i_unknownb0,i_unknownb1,i_unknownb2,i_unknownb3,i_unknownb4,i_unknownb5,i_unknownb6,i_unknownb7
    dd   i_unknownb8,i_unknownb9,i_unknownba,i_unknownbb,i_unknownbc,i_unknownbd,i_unknownbe,i_unknownbf

    dd   i_unknownc0,i_unknownc1,i_unknownc2,i_unknownc3,i_unknownc4,i_unknownc5,i_unknownc6,i_unknownc7
    dd   i_unknownc8,i_unknownc9,i_unknownca,i_unknowncb,i_unknowncc,i_unknowncd,i_unknownce,i_unknowncf

    dd   i_unknownd0,i_unknownd1,i_unknownd2,i_unknownd3,i_unknownd4,i_unknownd5,i_unknownd6,i_unknownd7
    dd   i_unknownd8,i_unknownd9,i_unknownda,i_unknowndb,i_unknowndc,i_unknowndd,i_unknownde,i_unknowndf

    dd   i_unknowne0,i_unknowne1,i_unknowne2,i_unknowne3,i_unknowne4,i_unknowne5,i_unknowne6,i_unknowne7
    dd   i_unknowne8,i_unknowne9,i_unknownea,i_unknowneb,i_unknownec,i_unknowned,i_unknownee,i_unknownef

    dd   i_unknownf0,i_unknownf1,i_unknownf2,i_unknownf3,i_unknownf4,i_unknownf5,i_unknownf6,i_unknownf7
    dd   i_unknownf8,i_unknownf9,i_unknownfa,i_unknownfb,i_unknownfc,i_unknownfd,i_unknownfe,i_unknownff
		  




  
keymap:
   
     db   '6',27
     db   '1234567890-=',8,9
     db   'qwertyuiop[]',13
     db   '~asdfghjkl;',39,96,0,'\zxcvbnm,./',0,'45 '
     db   '@234567890123',180,178,184,'6',176,'7'
     db   179,'8',181,177,183,185,182
     db   'AB<D',255,'FGHIJKLMNOPQRSTUVWXYZ'
     db   'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
     db   'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
     db   'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
   
   
keymap_shift:
   
     db   '6',27
     db   '!@#$%^&*()_+',8,9
     db   'QWERTYUIOP{}',13
     db   '~ASDFGHJKL:"~',0,'|ZXCVBNM<>?',0,'45 '
     db   '@234567890123',180,178,184,'6',176,'7'
     db   179,'8',181,177,183,185,182
     db   'AB>D',255,'FGHIJKLMNOPQRSTUVWXYZ'
     db   'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
     db   'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
     db   'ABCDEFGHIJKLMNOPQRSTUVWXYZ'


keymap_alt:

     db   ' ',27
     db   ' @ $  {[]}\ ',8,9
     db   '            ',13
     db   '             ',0,'           ',0,'4',0,' '
     db   '             ',180,178,184,'6',176,'7'
     db   179,'8',181,177,183,185,182
     db   'ABCD',255,'FGHIJKLMNOPQRSTUVWXYZ'
     db   'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
     db   'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
     db   'ABCDEFGHIJKLMNOPQRSTUVWXYZ'



; device irq owners


irq_owner       ; process id

     dd   0x0
     dd   0x0
     dd   0x0
     dd   0x0
     dd   0x0
     dd   0x0
     dd   0x0
     dd   0x0
     dd   0x0
     dd   0x0
     dd   0x0
     dd   0x0
     dd   0x0
     dd   0x0
     dd   0x0
     dd   0x0


; on irq read ports


irq00read  dd  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
irq01read  dd  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
irq02read  dd  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
irq03read  dd  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
irq04read  dd  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
irq05read  dd  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
irq06read  dd  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
irq07read  dd  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
irq08read  dd  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
irq09read  dd  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
irq10read  dd  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
irq11read  dd  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
irq12read  dd  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
irq13read  dd  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
irq14read  dd  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
irq15read  dd  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0



; device addresses

mididp     dd 0x0
midisp     dd 0x0

cdbase     dd 0x0
cdid       dd 0x0

hdbase     dd 0x0
hdid       dd 0x0
hdpos      dd 0x0

fat32part  dd 0x0
fat32label db '           '

keyboard   dd 0x1

sb16       dd 0x0
wss        dd 0x0

syslang    dd CFG_DEFAULT_LANGUAGE

; display

buttontype             dd 0x0 
buttontypechanged      dd 0x0 


tss_sceleton:

l.back   resw   2
l.esp0   resd   1
l.ss0    resw   2
l.esp1   resd   1
l.ss1    resw   2
l.esp2   resd   1
l.ss2    resw   2
l.cr3    resd   1
l.eip    resd   1
l.eflags resd   1
l.eax    resd   1
l.ecx    resd   1
l.edx    resd   1
l.ebx    resd   1
l.esp    resd   1
l.ebp    resd   1
l.esi    resd   1
l.edi    resd   1
l.es     resw   2
l.cs     resw   2
l.ss     resw   2
l.ds     resw   2
l.fs     resw   2
l.gs     resw   2
l.ldt    resw   2
l.trap   resw   1
l.io     resw   1


mouseunder
	resb  16*16*4


mousepointer

       db 2,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0
       db 2,1,2,0,0,0,0,0,0,0,0,0,0,0,0,0
       db 2,1,1,2,0,0,0,0,0,0,0,0,0,0,0,0
       db 2,1,1,1,2,0,0,0,0,0,0,0,0,0,0,0
       db 2,1,1,1,1,2,0,0,0,0,0,0,0,0,0,0
       db 2,1,1,1,1,1,2,0,0,0,0,0,0,0,0,0
       db 2,1,1,1,1,1,1,2,0,0,0,0,0,0,0,0
       db 2,1,1,1,1,1,1,1,2,0,0,0,0,0,0,0
       db 2,1,1,1,1,1,1,1,1,2,0,0,0,0,0,0
       db 2,1,1,1,1,1,2,2,2,2,2,0,0,0,0,0
       db 2,1,1,2,1,1,2,0,0,0,0,0,0,0,0,0
       db 2,1,2,0,2,1,1,2,0,0,0,0,0,0,0,0
       db 2,2,0,0,2,1,1,2,0,0,0,0,0,0,0,0
       db 2,0,0,0,0,2,1,1,2,0,0,0,0,0,0,0
       db 0,0,0,0,0,2,1,1,2,0,0,0,0,0,0,0
       db 0,0,0,0,0,0,2,2,2,0,0,0,0,0,0,0
       db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
       db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0



endofcode      ; end of code  -VT



